/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

#include "config.h"		/* Invoke configuration header */
#include <stdlib.h>		/* Access malloc etc */
#include <memory.h>		/* access memcpy etc. */
#include <string.h>		/* String manipulation functions */
#include <assert.h>

#include "icmP.h"		/* Access private interface */
#include "icmW.h"		/* Access queue manager */
#include "icmIoP.h"		/* Access private I/O interface */
#include "icmM.h"

#if DMALLOC
#include <dmalloc.h>
#endif

/* Higher level address management */
/* Use a hash table of handles to access addresses from handles */

#ifndef MAXHANDLE
#define MAXHANDLE 256
#endif

#ifndef MAXLOC
#define MAXLOC 20
#endif

#ifdef HANDLETRACE
icmTruth traceHandle = icmNo;
#endif

static icmPoolBase hpool = NULL;
static icmPoolBase cpool = NULL;

static icmHashPo htble=NULL;	/* table of handles */

static pthread_mutex_t hdl_mutex;
static pthread_mutex_t monitor_mutex;

static char *copyStringEl(char *src,char **buffer,long *len);

/* The real definition of a handle */
typedef struct _icmHandle_ {
  char *tgt;
  char *name;
  char *home;
  int locCount;
  char **location;		/* List of locations -- may be IP quartets */
  char buffer[64];		/* We try to store the data here ... */
  char *data;			/* But if not, ... we use this */
  int refCount;			/* The number of times this handle has been `kept' */
} icmHandleRec;

typedef struct _connList_ *connPo;
typedef struct _connList_ {
  icmMsgProc callback;	/* callback function */
  icmConn conn;
  void *client;			/* Any client data to use */
  connPo next;
} connRec;
  
/* Handles may have callbacks associated with them ... */
typedef struct _handle_info_ *hpo;

typedef struct _handle_info_ {
  icmHandleRec handle;		/* the handle itself */
  connPo callbacks;
} h_rec;

/* compute the hash value of a handle .. */
static long handleHash(icmHandle h)
{
  register long hash = 0;
  char *tgt = h->tgt;
  char *name = h->name;
  char *home = h->home;
  char **location = h->location;
  int i;

  if(tgt!=NULL)
    while(*tgt)
      hash = (hash<<4) + (hash^*tgt++);

  if(name!=NULL)
    while(*name)
      hash = (hash<<4) + (hash^*name++);

  if(home!=NULL)
    while(*home)
      hash = (hash<<4) + (hash^*home++);

  for(i=0;i<h->locCount;i++){
    char *loc = location[i];
    while(*loc)
      hash = (hash<<4) + (hash^*loc++);
  }

  return hash;
}

/* Compare two handles */
static long handleCmp(icmHandle h1,icmHandle h2)
{
  char *tgt1 = (h1->tgt?h1->tgt:"");
  char *tgt2 = (h2->tgt?h2->tgt:"");
  char *name1 = (h1->name?h1->name:"");
  char *name2= (h2->name?h2->name:"");
  char *home1 = (h1->home?h1->home:"");
  char *home2 = (h2->home?h2->home:"");
  int ret = strcmp(tgt1,tgt2);
  int i,j;
  icmTruth found = h1->locCount==0 && h1->locCount==0;

  if(ret!=0)
    return ret;

  ret = strcmp(name1,name2);
  if(ret!=0)
    return ret;

  ret = strcmp(home1,home2);
  if(ret!=0)
    return ret;

  for(i=0;!found && i<h1->locCount;i++){
    char *loc1 = h1->location[i];

    for(j=0;j<h2->locCount;j++)
      if(strcmp(loc1,h2->location[j])==0){
	found=icmYes;
	break;
      }
  }

  if(found)
    return 0;			/* We found a matching location */
  else
    return 1;
}

static icmStatus dispHdl(filePo f,void *h,int width,int precision,icmTruth alt);

static pthread_once_t hdlInit = PTHREAD_ONCE_INIT;

static void init_handles(void)
{
  /* initialize address table itself */
  hpool = icmInitPool(sizeof(h_rec),MAXHANDLE);
  cpool = icmInitPool(sizeof(connRec),MAXHANDLE);
  htble = icmNewHash(MAXHANDLE,(icmHashFun)handleHash,(icmCompFun)handleCmp,NULL);
  _installMsgProc('h',dispHdl);	/* extend outMsg to cope with handles */
  icmCreateMutex(&hdl_mutex);
  icmCreateMutex(&monitor_mutex);
}

icmHandle icmMakeHandle(char *target,char *name,char *home,int count,char **location)
{
  icmHandleRec handle = {target,name,home,count,location};
  hpo hp;
  int i;
  char *loc;

  pthread_once(&hdlInit,init_handles); /* initialize handles DB just once */

  if(home==NULL){
    home = handle.home = icmMachineName();
    loc = icmMachineIP();
    handle.location = location = &loc;
    handle.locCount = count = 1;
  }

  icmLockMutex(&hdl_mutex);

  hp = (hpo)icmSearch(&handle,htble); /* we dont want to be installing twice */

  if(hp!=NULL)
    hp->handle.refCount++;	/* keep these balanced */
  else{
    long len = 0;
    char *data;

    hp = (hpo)icmAllocPool(hpool);

    /* count how long we need ... */
    if(target!=NULL)
      len+=strlen(target)+1;

    if(name!=NULL)
      len+=strlen(name)+1;

    if(home!=NULL)
      len+=strlen(home)+1;

    for(i=0;i<count;i++)
      len+=strlen(location[i])+1;

    if(len>sizeof(hp->handle.buffer))
      data = hp->handle.data = (char*)malloc(len);
    else
      data = hp->handle.data = hp->handle.buffer;

    if(target!=NULL)
      hp->handle.tgt = copyStringEl(target,&data,&len);
    else
      hp->handle.tgt = NULL;
    if(name!=NULL)
      hp->handle.name = copyStringEl(name,&data,&len);
    else
      hp->handle.name = NULL;
    if(home!=NULL)
      hp->handle.home = copyStringEl(home,&data,&len);
    else
      hp->handle.home = NULL;

    hp->handle.locCount = count;
    if(count>0)
      hp->handle.location = (char**)icmMalloc(sizeof(char*)*count);
    else
      hp->handle.location = NULL;

    for(i=0;i<count;i++)
      hp->handle.location[i]=copyStringEl(location[i],&data,&len);

    hp->callbacks = NULL;
    hp->handle.refCount = 1;

    icmInstall(&hp->handle,hp,htble);
  }

  icmUnlockMutex(&hdl_mutex);
  
#ifdef HANDLETRACE_
  if(traceHandle)
    icmLogMsg(BOLD_ESC_ON"%d: make handle %h"BOLD_ESC_OFF,pthread_self(),&hp->handle);
#endif

  return &hp->handle;
}

icmHandle icmCommServerHandle(char *host)
{
  char *locs[MAXLOC];
  char locations[MAXLOC][MAXSTRING];
  int count=0;

  while(count<MAXLOC){
    if(icmGetNthHostIP(host,count,locations[count],MAXSTRING)==NULL)
      break;
    else{
      locs[count]=locations[count];
      count++;
    }
  }

  return icmMakeHandle("",ICM_COMMSERVER,icmGetHostname(host),count,(char **)locs);
}

icmHandle icmDummyHandle()
{
  return icmParseHandle("<dummy>");
}


icmHandle icmParseHandle(char *buffer)
{
  char copy[MAXHANDLE*2];
  char *target;
  char *name;
  char *home;
  char *loc;
  char *location[MAXLOC];
  char locs[MAXLOC][MAXSTRING];
  int count = 0;
  char *p;

  icmStrNCpy(copy,buffer,2*MAXHANDLE);

  /* Separate out the possible bits of the handle */

  /* Extract the locations field */
  if((p=strchr(copy,'/'))!=NULL){
    *p++ = '\0';
    loc = p;
    
    if(*loc++=='['){
      char *comma;

      while((comma=strchr(loc,','))!=NULL){
	while(strchr(" \t",*loc)!=NULL && loc!=comma)
	  loc++;
	location[count++]=loc;
	*comma++='\0';
	loc = comma;
      }
      if((comma=strchr(loc,']'))!=NULL){
	while(strchr(" \t",*loc)!=NULL && loc!=comma)
	  loc++;
	location[count++]=loc;
	*comma='\0';
      }
    }
    else if(strlen(loc)>0){	/* support the dec:agent@home/location form */
      while(strchr(" \t",*loc)!=NULL && *loc!='\0')
	loc++;
      if(*loc!='\0')
	location[count++]=loc;
    }
  } /* We have no explicit locations in this handle */

  /* extract the home base of the agent */
  if((p=strrchr(copy,'@'))!=NULL){
    *p++='\0';
    home = icmGetHostname(p);

    if(count==0){
      while(count<MAXLOC){
	if(icmGetNthHostIP(home,count,locs[count],MAXSTRING)==NULL)
	  break;
	else{
	  location[count]=locs[count];
	  count++;
	}
      }
    }
  }
  else{
    home = icmMachineName();

    if(count==0)
      location[count++]=icmMachineIP();
  }

  /* extract the name of the agent */
  if((p=strchr(copy,':'))!=NULL){
    target = copy;
    *p++='\0';
    name = p;
  }
  else{
    target="";
    name=copy;
  }

  return icmMakeHandle(target,name,home,count,location);
}

icmStatus icmReleaseHandle(icmHandle h)
{
  if(h!=NULL){
    icmLockMutex(&hdl_mutex);

    if(--h->refCount<=0){
      hpo hp = (hpo)icmSearch(h,htble); /* we dont want to be removing twice */

      if(hp!=NULL && hp==(hpo)h && h->refCount<=0){ /* check reference count again */

#ifdef HANDLETRACE_
	if(traceHandle)
	  icmLogMsg(BOLD_ESC_ON"%d: Free handle %h"BOLD_ESC_OFF,
		    pthread_self(),h);
#endif

	icmUninstall(h,htble);

	if(h->data!=h->buffer)
	  icmFree(h->data);

	if(h->location!=NULL)
	  icmFree(h->location);

	icmFreePool(hpool,hp);
      }
    }

    icmUnlockMutex(&hdl_mutex);
  }
  return icmOk;
}

void icmCheckHandle(icmHandle h)
{
  icmLockMutex(&hdl_mutex);

  if(h!=NULL){
    hpo hrec=icmSearch(h,htble);

    if(hrec==NULL)
      icmLogMsg(BOLD_ESC_ON"Handle %h not in table"BOLD_ESC_OFF,h);
    else if(h->refCount<=0)
      icmLogMsg(BOLD_ESC_ON"Handle %h w/empty refcount"BOLD_ESC_OFF,h);
  }
  
  icmUnlockMutex(&hdl_mutex);
}

icmStatus icmKeepHandle(icmHandle h)
{
  if(h!=NULL){
    icmLockMutex(&hdl_mutex);
    h->refCount++;
    icmUnlockMutex(&hdl_mutex);
#ifdef HANDLETRACE_
    if(traceHandle)
      icmLogMsg(BOLD_ESC_ON"%d: Keep handle %h"BOLD_ESC_OFF,pthread_self(),h);
#endif
  }
  return icmOk;
}

icmStatus icmSameAgentHandle(icmHandle h1,icmHandle h2)
{
  if(h1!=NULL && h2!=NULL){
    if(strncmp(h1->name,h2->name,MAXSTRING)==0 &&
       strncmp(h1->home,h2->home,MAXSTRING)==0)
      return icmOk;
    else
      return icmFailed;
  }
  else
    return icmError;
}

icmStatus icmSameHandle(icmHandle h1,icmHandle h2)
{
  if(h1!=NULL && h2!=NULL){
    if(strncmp(h1->tgt,h2->tgt,MAXSTRING)==0 &&
       strncmp(h1->name,h2->name,MAXSTRING)==0 &&
       strncmp(h1->home,h2->home,MAXSTRING)==0)
      return icmOk;
    else
      return icmFailed;
  }
  else
    return icmError;
}

/*
 * Analyse a handle into constituent components 
 */
icmStatus icmAnalyseHandle(icmHandle handle,
			   char **tgt,char **name,char **home,
			   char ***location, int *locCount)
{
  if(handle==NULL)
    return icmError;
  else{
    icmStatus ret = icmOk;

    if(tgt!=NULL)
      *tgt = handle->tgt;
    if(name!=NULL)
      *name = handle->name;
    if(home!=NULL)
      *home = handle->home;
    if(location!=NULL)
      *location = handle->location;
    if(locCount!=NULL)
      *locCount = handle->locCount;
    return ret;
  }
}

icmStatus icmHandleName(icmHandle handle,char *buffer,int buffsize)
{
  icmStatus ret = icmOk;

  if(handle==NULL)
    icmStrNCpy(buffer,"<null handle>",buffsize);
  else{
    int i;
    char *sep = "/[";
    *buffer = '\0';

    if(handle->tgt!=NULL && strlen(handle->tgt)>0){
      icmStrNCpy(buffer,handle->tgt,buffsize);
      icmStrNCat(buffer,":",buffsize);
    }

    if(handle->name!=NULL){
      icmStrNCat(buffer,handle->name,buffsize);
      icmStrNCat(buffer,"@",buffsize);
    }
    else
      icmStrNCat(buffer,"<noname>@",buffsize);

    if(handle->home!=NULL)
      icmStrNCat(buffer,handle->home,buffsize);

    if(handle->locCount>0){
      for(i=0;i<handle->locCount;i++){
	icmStrNCat(buffer,sep,buffsize);
	icmStrNCat(buffer,handle->location[i],buffsize);
	sep = ",";
      }

      icmStrNCat(buffer,"]",buffsize);
    }
    else
      icmStrNCat(buffer,"/[]",buffsize);

#ifdef HANDLETRACE_
    if(traceHandle){
      char buff[32];
      icmStrMsg(buff,32,"-%d",handle->refCount);
      icmStrNCat(buffer,buff,buffsize);
    }
#endif
  }

  return ret;
}

static icmStatus dispHdl(filePo f,void *h,int width,int precision,icmTruth alt)
{
  char buff[MAXSTRING];

  icmHandleName((icmHandle)h,buff,MAXSTRING);

  return outString(f,buff,strlen(buff),width,precision,' ',alt,"");
}


/* Support for handle monitoring ... */

icmStatus icmMonitorHandle(icmConn conn,icmHandle handle,icmMsgProc proc,void *cl)
{
  hpo hrec;

  if(handle==NULL||(hrec=icmSearch(handle,htble))==NULL)
    return icmError;
  else{
    connPo c;

    icmLockMutex(&monitor_mutex);

    c=hrec->callbacks;

    while(c!=NULL){
      if(c->conn==conn){
	c->callback = proc;	/* just override existing callback */
	c->client = cl;

	icmUnlockMutex(&monitor_mutex);
	return icmOk;
      }
      else
	c = c->next;
    }

    c = (connPo)icmAllocPool(cpool);

    c->next = hrec->callbacks;
    hrec->callbacks = c;
    c->conn = conn;
    c->callback = proc;
    c->client = cl;

    icmUnlockMutex(&monitor_mutex);
    return icmOk;
  }
}

icmStatus icmUnmonitorHandle(icmConn conn,icmHandle handle)
{
  hpo hrec;
  icmLockMutex(&monitor_mutex);

  if(handle==NULL||(hrec=icmSearch(handle,htble))==NULL){
    icmUnlockMutex(&monitor_mutex);
    return icmError;
  }
  else{
    connPo *c=&hrec->callbacks;

    while(*c!=NULL){
      if((*c)->conn==conn){
	connPo n = (*c)->next;
	icmFreePool(cpool,*c);
	*c = n;
	icmUnlockMutex(&monitor_mutex);
	return icmOk;
      }
      else
	c = &(*c)->next;
    }
    icmUnlockMutex(&monitor_mutex);
    return icmFailed;
  }
}


/* Call the callback associated with a handle */
/* It returns true if there was a call back */
icmStatus _CallCallBack_(icmConn conn,icmHandle t,icmHandle s,icmOption o,icmMsg m)
{
  hpo hp = (hpo)icmSearch(t,htble);

  if(hp==NULL)
    return icmError;
  else{
    connPo c = hp->callbacks;

    icmLockMutex(&monitor_mutex);

    while(c!=NULL)
      if(c->conn==conn){
	icmUnlockMutex(&monitor_mutex);
	return (c->callback)(conn,t,s,o,m,c->client);
      }
      else
	c = c->next;

    icmUnlockMutex(&monitor_mutex);

    return icmFailed;		/* No callback succeeded */
  }
}

static char *copyStringEl(char *src,char **buffer,long *len)
{
  char *dest = *buffer;
  char *p = dest;
  long ln = *len;

  while(*src!='\0' && ln>0){
    *p++ = *src++;
    ln--;
  }

  if(ln-->0)
    *p++='\0';

  *buffer = p;
  *len = ln;
  return dest;
}

static icmStatus dAdd(void *n,void *r,void *c)
{
  dispHdl((filePo)c,r,0,0,icmNo);
  outCh((filePo)c,'\n');
  return icmOk;
}

void icmDumpAddresses(void)
{
  icmProcessTable(dAdd,htble,(void*)icmLogFile);
}
