/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/* 
 * set up connection to the ICM communications server
 */
#include "config.h"		/* Invoke configuration header */
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#ifdef WIN32
#include <io.h>
#include <time.h>
#else
#include <unistd.h>
#include <sys/time.h>
#include <unistd.h>
#endif
#include <sys/types.h>



#include "icmP.h"		/* Real message functions */
#include "icmIo.h"		/* April I/O interface */
#include "icmIoP.h"		/* Access private I/O interface */
#include "icmSock.h"		/* April socket library interface */

static icmPoolBase cpool=NULL;
static pthread_once_t commInit = PTHREAD_ONCE_INIT;
static icmStatus dispConn(filePo f,void *h,int width,int precision,icmTruth alt);

static void initCommPool(void)
{
  cpool = icmInitPool(sizeof(icmConnRec),16);

  _installMsgProc('Q',dispConn);	/* extend outMsg to cope with connections */
}

icmConn icmMakeConn(filePo rem,int port,char *host)
{
  pthread_once(&commInit,initCommPool);

  {
    icmConn conn = (icmConn)icmAllocPool(cpool);

    if(conn!=NULL){
      conn->conn = rem;
      if(rem!=NULL)
	conn->socket = (int)fileData(rem);
      else
	conn->socket = -1;
      conn->port = port;
      conn->pending = NULL;
      conn->server = NULL;
      conn->idler = NULL;
      conn->dummyH = icmDummyHandle();
      conn->seqNo = 0;
      conn->loopGoing = icmNo;
      icmStrNCpy(conn->host,host,sizeof(conn->host));

      pthread_mutex_init(&conn->mutex,NULL);
    }
    return conn;
  }
}

/* Called at startup to initialise connection to outside world. */
icmStatus icmInitComms(int port,char *host,icmConn *conn)
{
  icmStatus res;
  filePo remote;
  icmTruth usingLocalHost = icmNo;
  icmTruth explicitHost = host!=NULL;
  char msg[MAXSTRING];

  if(port==0){
    char *portstr = getenv("ICM_COMMS_PORT");

    if(portstr==NULL || (port=atoi(portstr))!=0)
      port = ICMCOMMSPORT;
  }

  if(host==NULL){
    char *locstr = getenv("ICM_COMMSERVER");

    if(locstr!=NULL && strcmp(locstr,"")!=0){
      host = locstr;
      explicitHost = icmYes;
    }
    else
      host = icmMachineName();
  }

  /* connect to the communications server */
  remote = icmConnectRemote(host,port,&res,msg,NumberOf(msg));

  switch(res){
  case icmOk:{
    icmConn cx = *conn = icmMakeConn(remote,port,host);

    cx->server = icmCommServerHandle(host);

    configureIo(remote,turnOffBlocking);
    return icmOk;
  }
  case icmError:
    if(!explicitHost && !usingLocalHost){
      remote = connectRemote("localhost",port,&res);

      if(res==icmOk){
	icmConn cx = *conn = icmMakeConn(remote,port,host);

	cx->server = icmCommServerHandle(host);

	configureIo(remote,turnOffBlocking);
	return icmOk;
      }
      else{
	icmLogMsg("cannot establish connection with %s, or with localhost",host);
	return icmError;
      }
    }
    else{
      icmLogMsg("cannot establish connection with %s",host);
      return icmError;
    }
  default:
    return res;
  }
}

int icmCommSocket(icmConn conn)
{
  return conn->socket;
}

icmHandle icmCommserver(icmConn conn)
{
  return conn->server;
}

static icmStatus dispConn(filePo f,void *h,int width,int precision,icmTruth alt)
{
  icmConn conn = (icmConn) h;

  if(conn!=NULL)
    return outMsg(f,"[%s:%d(%d)-%d]",conn->host,conn->port,conn->socket,conn->pending);
  else
    return outMsg(f,"(null conn)");
}


icmStatus icmCloseComms(icmConn conn)
{
  icmLockMutex(&conn->mutex);

  if(conn->socket>=0){
    closeFile(conn->conn);	/* Close the comms socket down */
    icmReleaseHandle(conn->dummyH); /* clear down the dummy handle */
    conn->socket = -1;		/* And allow for re-initialization */
    icmDestroyMutex(&conn->mutex);

    icmFreePool(cpool,conn);

    if(icmLogFile!=NULL)
      flushFile(icmLogFile);	/* flush but dont close logging file */
#ifdef _WINSOCKAPI_
    WSACleanup();		/* Who knows how many times it was opened? */
#endif
    return icmOk;
  }
  else
    return icmError;
}

/*
 * Handle comm server responses
 */
icmStatus icmStandardPingHandler(icmMsg msg,icmHandle to,icmHandle from,
			   icmHandle sender,void *client)
{
  icmHandle who;

  /* We parse responses from the name server */
  if(icmScanMsg(msg,"%(icmRegisterAgent%,icmOk%h%)",&who))
    return icmOk;
  else if(icmScanMsg(msg,"%(icmRegisterAgent%s%h%)",NULL,&who)){
    char name[MAXSTRING];
    icmHandleName(who,name,MAXSTRING);
    icmLogMsg("Duplicate agent: %s",name);
    return icmError;
  }
  else
    return icmFailed;
}

icmStatus icmSetIdleProc(icmConn conn,icmIdleProc proc,void *client)
{
  conn->idler = proc;
  conn->idle_client = client;
  return icmOk;
}

icmStatus icmEventLoop(icmConn conn,icmMsgProc hook,void *client)
{
  conn->loopGoing = icmYes;

  while(conn->loopGoing){
    fd_set fdset, copy;
    int status;

    if(conn->idler!=NULL)
      conn->idler(conn,conn->idle_client);

    FD_ZERO(&fdset);
    if(conn->socket>0)
      FD_SET(conn->socket, &fdset);

    copy = fdset;
    status = select(conn->socket+1, &copy, NULL, NULL, NULL);

    if(status < 0){
#ifdef _WINSOCKAPI_
      switch(WSAGetLastError()){
      case WSAENETDOWN:		/* Network has gone down */
	icmLogMsg("Network has gone down");
	icmCloseComms(conn);
	return icmEndFile;
      case WSAEINTR:		/* Interrupted -- try again */
      case WSAEINPROGRESS:
	continue;
      default:
	icmLogMsg("Problem in the network");
	return icmError;
      }
#else
      switch(errno){			/* Unix based error hanling */
      case EINTR:
	errno = 0;			/* Loop around for another try */
	continue;
      default:
	icmLogMsg("error %s(%d) in select", strerror(errno),errno);
	return icmError;
      }
#endif
    }
    else{
      icmHandle to,sender;
      icmOption opts;
      icmStatus stat=icmEndFile;
      icmMsg msg;

      while(conn->loopGoing && 
	    (stat=icmGetMsg(conn,&to,&sender,&opts,&msg,NULL))==icmOk){
	switch(hook(conn,to,sender,opts,msg,client)){
	default:
	  icmReleaseMsg(msg);
	  icmReleaseHandle(to);
	  icmReleaseHandle(sender);
	  icmReleaseOptions(opts);
	  continue;
	case icmError:
	  icmReleaseMsg(msg);
	  return icmError;
	}
      }
      if(stat==icmEndFile)
	conn->loopGoing = icmNo;
    }
  }
  return icmOk;		/* We'll probably never get here */
}

void icmTerminateEventLoop(icmConn conn)
{
  conn->loopGoing = icmNo;
}

/*
 * Agent registration functions
 */
icmStatus icmRegisterAgent(icmConn conn,icmHandle agent,icmOption opts,icmHandle *real)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmRegisterAgent%h%O%)",agent,opts);
    icmStatus ret = icmSendMsg(conn,conn->server,agent,NULL,msg);
    long seqNo = -LONG_MAX;
    icmHandle to,sndr;
    icmOption mopts;

    icmReleaseMsg(msg);
    icmReleaseOptions(opts);

    while((ret=icmGetMsg(conn,&to,&sndr,&mopts,&msg,&seqNo))==icmOk){
      char res[MAXSTRING];
      icmHandle hh;

      if(icmSameAgentHandle(sndr,conn->server)==icmOk && 
	 icmScanMsg(msg,"%(icmRegisterAgent%s%h%)",&res,&hh)==icmOk &&
	 icmSameAgentHandle(hh,agent)==icmOk){

	icmReleaseHandle(conn->server);
	icmKeepHandle(sndr);
	conn->server = sndr;

	icmReleaseMsg(msg);	/* release this message */
	icmReleaseOptions(mopts);

	if(strcmp(res,"icmOk")==0){
	  if(real!=NULL)
	    *real = hh;
	  return icmOk;
	}
	else if(strcmp(res,"icmFailed")==0)
	  return icmFailed;
	else
	  return icmError;
      }
      else
	icmReplaceMsg(conn,to,sndr,opts,msg,seqNo);
    }
    return icmError;		/* Something went wrong here */
  }
  else
    return icmError;
}

/*
 *  de-register with communications server and other acquaintances
 */
icmStatus icmDeregisterAgent(icmConn conn,icmHandle agent)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmDeregisterAgent%h%)",agent);
    icmStatus ret = icmSendMsg(conn,conn->server,agent,NULL,msg);
    long seqNo = -LONG_MAX;
    icmHandle to,sndr;
    icmOption opts;

    icmReleaseMsg(msg);

    while((ret=icmGetMsg(conn,&to,&sndr,&opts,&msg,&seqNo))==icmOk){
      char res[MAXSTRING];
      icmHandle hh;

      if(sndr==conn->server && 
	 icmScanMsg(msg,"%(icmDeregisterAgent%s%h%)",&res,&hh)==icmOk &&
	 icmSameAgentHandle(hh,agent)==icmOk){
	icmReleaseMsg(msg);	/* release this message */
	icmReleaseOptions(opts);
	if(strcmp(res,"icmOk")==0)
	  return icmOk;
	else if(strcmp(res,"icmFailed")==0)
	  return icmFailed;
	else
	  return icmError;
      }
      else
	icmReplaceMsg(conn,to,sndr,opts,msg,seqNo);
    }
    return ret;			/* Something went wrong here */
  }
  else
    return icmError;
}

icmStatus icmDetachAgent(icmConn conn,icmHandle agent,icmOption opts)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmDetachAgent%h%O%)",agent,opts);
    icmStatus ret = icmSendMsg(conn,conn->server,agent,NULL,msg);
    long seqNo = -LONG_MAX;
    icmHandle to,sndr;
    icmOption ropts;

    icmReleaseMsg(msg);
    icmReleaseOptions(opts);

    while((ret=icmGetMsg(conn,&to,&sndr,&ropts,&msg,&seqNo))==icmOk){
      char res[MAXSTRING];
      icmHandle hh;

      if(sndr==conn->server && 
	 icmScanMsg(msg,"%(icmDetachAgent%s%h%)",&res,&hh)==icmOk &&
	 icmSameAgentHandle(hh,agent)==icmOk){
	icmReleaseMsg(msg);	/* release this message */
	icmReleaseOptions(ropts);

	if(strcmp(res,"icmOk")==0)
	  return icmOk;
	else if(strcmp(res,"icmFailed")==0)
	  return icmFailed;
	else
	  return icmError;
      }
      else
	icmReplaceMsg(conn,to,sndr,ropts,msg,seqNo);
    }
    return icmError;		/* Something went wrong here */
  }
  else
    return icmError;
}


icmStatus icmAttachAgent(icmConn conn,icmHandle agent, icmOption opts)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmAttachAgent%h%O%)",agent,opts);
    icmStatus ret = icmSendMsg(conn,conn->server,agent,NULL,msg);
    long seqNo = -LONG_MAX;
    icmHandle to,sndr;
    icmOption ropts;

    icmReleaseMsg(msg);
    icmReleaseOptions(opts);

    while((ret=icmGetMsg(conn,&to,&sndr,&ropts,&msg,&seqNo))==icmOk){
      char res[MAXSTRING];
      icmHandle hh;

      if(sndr==conn->server && 
	 icmScanMsg(msg,"%(icmAttachAgent%s%h%)",&res,&hh)==icmOk &&
	 icmSameAgentHandle(hh,agent)==icmOk){
	icmReleaseMsg(msg);	/* release this message */
	icmReleaseOptions(ropts);

	if(strcmp(res,"icmOk")==0)
	  return icmOk;
	else if(strcmp(res,"icmFailed")==0)
	  return icmFailed;
	else
	  return icmError;
      }
      else
	icmReplaceMsg(conn,to,sndr,ropts,msg,seqNo);
    }
    return icmError;		/* Something went wrong here */
  }
  else
    return icmError;
}

icmStatus icmMonitorAgent(icmConn conn,icmHandle ourH,icmHandle agent)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmMonitorAgent%h%)",agent);
    icmStatus ret = icmSendMsg(conn,conn->server,ourH,NULL,msg);
    long seqNo = -LONG_MAX;
    icmHandle to,sndr;
    icmOption opts;

    icmReleaseMsg(msg);

    while((ret=icmGetMsg(conn,&to,&sndr,&opts,&msg,&seqNo))==icmOk){
      icmHandle hh;

      if(sndr==conn->server){
	if(icmScanMsg(msg,"%(icmMonitorAgent%,icmOk%h%)",&hh)==icmOk &&
	   icmSameAgentHandle(hh,agent)==icmOk){
	  icmReleaseMsg(msg);	/* release this message */
	  icmReleaseOptions(opts);

	  return icmOk;
	}
	else if(icmScanMsg(msg,"%(icmMonitorAgent%,icmFailed%h%)",&hh)==icmOk &&
		icmSameAgentHandle(hh,agent)==icmOk){
	  icmReleaseMsg(msg);	/* release this message */
	  icmReleaseOptions(opts);

	  return icmFailed;
	}
	else if(icmScanMsg(msg,"%(icmMonitorAgent%,icmError%h%)",&hh)==icmOk &&
		icmSameAgentHandle(hh,agent)==icmOk){
	  icmReleaseMsg(msg);	/* release this message */
	  icmReleaseOptions(opts);

	  return icmError;
	}
	else
	  icmReplaceMsg(conn,to,sndr,opts,msg,seqNo);
      }
      else
	icmReplaceMsg(conn,to,sndr,opts,msg,seqNo);
    }
    return icmError;		/* Something went wrong here */
  }
  else
    return icmError;
}

icmStatus icmUnMonitorAgent(icmConn conn,icmHandle ourH,icmHandle agent)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmUnMonitorAgent%h%)",agent);
    icmStatus ret = icmSendMsg(conn,conn->server,ourH,NULL,msg);
    char Buff[MAXSTRING];
    icmHandle a,to,sndr;
    icmOption opts;

    if((ret=icmWaitMsg(conn,&to,&sndr,&opts,"%(icmUnMonitorAgent%s%h%)",Buff,&a))==icmOk){
      if(icmSameAgentHandle(a,agent)==icmOk){
	icmReleaseHandle(a);
	icmReleaseOptions(opts);

	if(strcmp(Buff,"icmOk")==0)
	   return icmOk;
	else if(strcmp(Buff,"icmFailed")==0)
	  return icmFailed;
	else
	  return icmError;
      }
      else
	return icmError;
    }
    else
      return ret;
  }
  else
    return icmError;
}


/*
 * Test to see if an agent is known to the network
 */
icmStatus icmPingAgent(icmConn conn,icmHandle us,icmHandle agent,icmHandle *real)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmPingAgent%h%)",agent);
    icmStatus ret = icmSendMsg(conn,conn->server,us,NULL,msg);
    long seqNo = -LONG_MAX;
    icmHandle to,sndr;
    icmOption opts;

    icmReleaseMsg(msg);

    while((ret=icmGetMsg(conn,&to,&sndr,&opts,&msg,&seqNo))==icmOk){
      char res[MAXSTRING];
      icmHandle hh;

      if(icmSameAgentHandle(sndr,conn->server)==icmOk && 
	 icmScanMsg(msg,"%(icmPingAgent%s%h%)",&res,&hh)==icmOk &&
	 icmSameAgentHandle(hh,agent)==icmOk){

	icmReleaseHandle(conn->server);
	icmKeepHandle(sndr);
	conn->server = sndr;

	icmReleaseMsg(msg);	/* release this message */
	icmReleaseOptions(opts);

	if(strcmp(res,"icmOk")==0){
	  if(real!=NULL)
	    *real = hh;
	  return icmOk;
	}
	else if(strcmp(res,"icmFailed")==0)
	  return icmFailed;
	else
	  return icmError;
      }
      else
	icmReplaceMsg(conn,to,sndr,opts,msg,seqNo);
    }
    return icmError;		/* Something went wrong here */
  }
  else
    return icmError;
}

/*
 *  Request a list of all registered agents
 */
icmStatus icmListAgents(icmConn conn,icmHandle agent,icmHandle *agents,long *max)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"icmListAgents");
    icmStatus ret = icmSendMsg(conn,conn->server,agent,NULL,msg);
    icmHandle to,sndr;
    icmOption opts;

    icmReleaseMsg(msg);

    if((ret=icmWaitMsg(conn,&to,&sndr,&opts,"%(icmListAgents%#%)",&msg))==icmOk){
      if(icmScanMsg(msg,"icmFailed")==icmOk){
	icmReleaseMsg(msg);	/* release this message */
	return icmFailed;
      }
      else if(icmScanMsg(msg,"%*h",max,agents)==icmOk){
	icmReleaseMsg(msg);	/* release this message */
	return icmOk;
      }
    }
    return ret;			/* Something went wrong here */
  }
  else
    return icmError;
}

/*
 * Disconnect from the communications server.
 * All registered agents become detached.
 */
icmStatus icmDisconnect(icmConn conn)
{
  if(conn!=NULL && conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"icmDisconnect");
    icmStatus ret = icmSendMsg(conn,conn->server,conn->dummyH,NULL,msg);
    char Buff[MAXSTRING];
    icmHandle to,sndr;
    icmOption opts;
    long tag;

    icmReleaseMsg(msg);

    if((ret=icmWaitMsg(conn,&to,&sndr,&opts,"%(icmDisconnect%s%d%)",Buff,&tag))==icmOk){
      icmLockMutex(&conn->mutex);

      icmReleaseHandle(to);
      icmReleaseHandle(sndr);
      icmReleaseOptions(opts);

      if(strcmp(Buff,"icmOk")==0){
	conn->seqNo = tag;
	closeFile(conn->conn);	/* Close the comms socket down */
	conn->socket = -1;	/* And allow for re-initialization */
	conn->conn = NULL;
#ifdef _WINSOCKAPI_
	WSACleanup();		/* Who knows how many times it was opened? */
#endif
      }
      else if(strcmp(Buff,"icmFailed")==0)
	ret = icmFailed;
      else
	ret = icmError;
      icmUnlockMutex(&conn->mutex);
    }
    return ret;
  }
  else
    return icmError;
}

/*
 * Reconnect to the communications server.
 * All registered agents become re-attached.
 */
icmStatus icmReconnect(icmConn conn)
{
  icmStatus res;
  long tag = conn->seqNo;

  icmLockMutex(&conn->mutex);

  if(conn->conn!=NULL){
    icmUnlockMutex(&conn->mutex);
    return icmError;
  }

  conn->server = icmCommServerHandle(conn->host);
  /* connect to the communications server */
  conn->conn = connectRemote(conn->host,conn->port,&res);

  switch(res){
  case icmOk:{
    configureIo(conn->conn,turnOffBlocking);
    conn->socket = (int)fileData(conn->conn);
    conn->pending = NULL;
    conn->seqNo = 0;
    break;
  }
  case icmError:
    icmLogMsg("cannot establish connection with %h",conn->server);
    icmUnlockMutex(&conn->mutex);
    return icmError;
  default:
    icmUnlockMutex(&conn->mutex);
    return res;
  }

  icmUnlockMutex(&conn->mutex);

  if(conn->socket>0 && conn->server!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%(icmReconnect%d%)",tag);
    icmStatus ret = icmSendMsg(conn,conn->server,conn->dummyH,NULL,msg);
    char Buff[MAXSTRING];
    icmHandle to,sndr;
    icmOption opts;

    icmReleaseMsg(msg);
    
    if((ret=icmWaitMsg(conn,&to,&sndr,&opts,"%(icmReconnect%s%)",Buff))==icmOk){
      icmReleaseHandle(to);
      icmReleaseHandle(sndr);
      icmReleaseOptions(opts);

      if(strcmp(Buff,"icmOk")==0)
	return icmOk;
      else if(strcmp(Buff,"icmFailed")==0)
	return icmFailed;
      else
	return icmError;
    }
    else
      return ret;
  }
  else
    return icmError;
}


