/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*  ICM file handling system  */ 
#include "config.h"		/* Invoke configuration header */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <sys/types.h>
#ifdef WIN32
#include <windows.h>
#include <io.h>
#include <signal.h>
#include <fcntl.h>
#else
#include <sys/wait.h>
#include <unistd.h>
#endif
#include <fcntl.h>
#include <limits.h>		/* Where OPEN_MAX is defined */

#ifdef HAVE_RLIMIT
#include <sys/time.h>
#include <sys/resource.h>
#endif

#ifdef SYSV
#include <stropts.h>
#endif

#include "icmP.h"
#include "icmIoP.h"		/* Get our own definitions */

#if DMALLOC
#include <dmalloc.h>
#endif

/* Regular file functions */

filePo icmLogFile=NULL;	/* File to write the log to */
filePo stdIn = NULL;
filePo stdOut = NULL;
filePo stdErr = NULL;

/* Standard functions for re-filling the file buffer, flushing it out and closing the
   file */

static icmStatus _refillFile(filePo f,int fno)
{
  if(f->in_pos==f->in_len){	/* nead to read more input? */
    int max = MAXLINE;
    int len = read(fno,f->in_line,max); /* Use system call for reading */

    if(len<0){			/* something wrong? */
      if(errno==EAGAIN||errno==EINTR)	/* No data available */
	return icmFailed;	/* We couldnt refill yet */
      else{			/* Pretend its EOF */
	f->in_len = 0;
	f->in_pos = 0;
	return f->status = icmEndFile;	/* We have reached end of file */
      }
    }
    else{			/* Reset the buffer pointers  */
      f->in_len = len;
      f->in_pos = 0;
      if(f->in_len==0)
	f->status = icmEndFile;
      else
	f->status = icmOk;
      return f->status;
    }
  }
  else
    return icmOk;			/* Already got stuff in there */
}

static icmStatus _flushFile(filePo f,int fno)
{
  char *cp = f->out_line;
  long len = f->out_pos;
  long written;

  while(len>0 && (written=write(fno,cp,len))!=len)
    if(written==-1){
      switch(errno){
      case 0:
	continue;
      case EINTR:
      case EAGAIN:
	return icmFailed;
	
      default:
	icmLogMsg("Problem %s (%d) in writing %d bytes to %s[%d]\n",
		      strerror(errno),errno,len,f->filename,f->client);
	return icmError;
      }
    }
    else{
      cp+=written;
      len-=written;
    }
    f->out_pos = 0;
    return icmOk;
}

/* We can asssume that flush has already been called */
static icmStatus _closeFile(filePo f,int fno)
{
  if(close(fno)==-1)
    return icmError;
  else
    return icmOk;
}

static icmStatus _configureFile(filePo f,icmIoConfigOpt mode,int fno)
{
#ifdef WIN32
  long flag = 0;
#else
  long flag = fcntl(fno,F_GETFL);
#endif

  switch(mode){
  case turnOffBlocking:
#ifndef WIN32
	  flag |= _O_NONBLOCK;		/* make file non-blocking */
#endif
    break;
  case turnOnBlocking:
#ifndef WIN32
    flag &= ~_O_NONBLOCK;	/* make file blocking again */
#endif
    break;

  case enableAsynch:		/* Enable asynchronous I/O */
#ifndef WIN32
#ifdef SYSV
    if(ioctl(fno,I_SETSIG,S_INPUT)>=0)
      return icmOk;
    else
      return icmError;
#else
    if(fcntl(fno,F_SETOWN,getpid())<0)
      icmLogMsg("Warning: couldnt set ownership of fd %d",fno);
    flag |= O_NONBLOCK;
#endif
    break;
  case disableAsynch:
#ifdef SYSV
    if(ioctl(fno,I_SETSIG,0)>=0)
      return icmOk;
    else
      return icmError;
#else
    flag &= ~O_NONBLOCK;	/* We no longer want to receive interrupts */
#endif
#endif
    break;
  default:			/* nothing */
    ;
  }
#ifdef WIN32
  if(ioctlsocket(fno,FIONBIO,&flag)==-1)
	return icmError;
#else
  if(fcntl(fno,F_SETFL,flag)==-1)
    return icmError;
#endif
  else
    return icmOk;
}

/* Check to see if a file is ready */
icmStatus _fileInReady(filePo f,int fno)
{
  if((f->mode&ioREAD)!=0){
    fd_set fdin;
    struct timeval period;

    FD_ZERO(&fdin);

    FD_SET(fno,&fdin);

    period.tv_sec = 0;
    period.tv_usec = 0;

    if(select(fno+1,&fdin,NULL,NULL,&period)>0)
      return icmOk;
    else
      return icmFailed;
  }
  else
    return icmError;
}

/* Check to see if a file is ready for output */
icmStatus _fileOutReady(filePo f,int fno)
{
  if((f->mode&ioWRITE)!=0){
    fd_set fdout;
    struct timeval period;

    FD_ZERO(&fdout);

    FD_SET(fno,&fdout);

    period.tv_sec = 0;
    period.tv_usec = 0;

    if(select(fno+1,NULL,&fdout,NULL,&period)>0)
      return icmOk;
    else
      return icmFailed;
  }
  else
    return icmError;
}

static icmStatus regularFile(char *name)
{
#ifndef __QNX__
  struct stat buf;
#else
  struct _fsys_stat buf;
#endif

#ifndef __QNX__
  if(stat(name, &buf) == -1)
#else
  if(fsys_stat(name, &buf) == -1)
#endif
    return icmFailed;		/* File not found */
#ifndef WIN32
  else if(S_ISDIR(buf.st_mode))
    return icmError;
#endif
  else
    return icmOk;
}

/* Open a file for input only */
filePo icmOpenInFile(char *file)
{
  initICMIo();

  if(regularFile(file)==icmOk){
    int inFileRefNum = open(file,O_RDONLY);

    if(inFileRefNum==-1)
      return NULL;
    else
      return _new_io(file,ioFile,ioREAD,inFileRefNum,_refillFile,NULL,
		     _closeFile,_fileInReady,_fileOutReady,_configureFile);
  }
  else
    return NULL;
}

/* This is a fake ... */
filePo icmLocateFile(char *name)
{
  return openInFile(name);
}

/* These only apply to Unix */
filePo icmOpenStdin(void)
{
  if(stdIn==NULL)
    stdIn = _new_io("stdin",ioFile,ioREAD,0,_refillFile,NULL,_closeFile,
		    _fileInReady,NULL,_configureFile);
  return stdIn;
}

filePo icmOpenStdout(void)
{
  if(stdOut==NULL)
    stdOut=_new_io("stdout",ioFile,ioWRITE,1,NULL,_flushFile,_closeFile,
		   NULL,_fileOutReady,_configureFile);
  return stdOut;
}

filePo icmOpenStderr(void)
{
  if(stdErr==NULL)
    stdErr=_new_io("stderr",ioFile,ioWRITE,2,NULL,_flushFile,_closeFile,
		   NULL,_fileOutReady,_configureFile);
  return stdErr;
}

icmStatus icmInitLogfile(char *name)
{
  if(strcmp(name,"-")==0)
    icmLogFile = OpenStdout();
  else
    icmLogFile = openAppendFile(name);
  if(icmLogFile!=NULL)
    return icmOk;
  else
    return icmError;
}

icmStatus icmSeekFile(filePo f,long where,long mode)
{
  long new_pos;
  if(f->type!=ioFile)
    return icmError;			/* can only seek regular files for now */

  icmLockMutex(&f->i_mutex);
  f->in_pos = f->in_len = 0;
  flushFile(f);

  new_pos = lseek((int)f->client,where,mode);

  if(new_pos==-1){
    icmUnlockMutex(&f->i_mutex);
    return icmError;
  }
  else{
    icmUnlockMutex(&f->i_mutex);
    f->f_pos = new_pos;
    return icmOk;
  }
}
  
filePo icmOpenOutFile(char *file)
{
  initICMIo();

  switch(regularFile(file)){
  case icmOk:
  case icmFailed:{			/* File not previously found */
#ifdef WIN32
	  int outFileRefNum = open(file,O_WRONLY|O_CREAT);
#else
	  int outFileRefNum = open(file,O_WRONLY|O_CREAT,STD_PERMISSIONS);
#endif

    if(outFileRefNum==-1)
      return NULL;
    else
      return _new_io(file,ioFile,ioREAD|ioWRITE,outFileRefNum,
		     _refillFile,_flushFile,_closeFile,
		     _fileInReady,_fileOutReady,_configureFile);
  }
  default:
    return NULL;
  }
}

/* Open a file for input and output */
filePo icmOpenInOutFile(char *file)
{
  initICMIo();

  switch(regularFile(file)){
  case icmOk:
  case icmFailed:{			/* File not previously found */
#ifdef WIN32
    int fileRefNum = open(file,O_RDWR|O_CREAT);
#else
	int fileRefNum = open(file,O_RDWR|O_CREAT,STD_PERMISSIONS);
#endif

    if(fileRefNum==-1)
      return NULL;
    else
      return _new_io(file,ioFile,ioREAD|ioWRITE,fileRefNum,
		     _refillFile,_flushFile,_closeFile,
		     _fileInReady,_fileOutReady,_configureFile);
  }
  default:
    return NULL;
  }
}

/* create and/or truncate existing file */
filePo icmNewOutFile(char *file)
{
  initICMIo();

  switch(regularFile(file)){
  case icmOk:
  case icmFailed:{			/* File not previously found */
#ifdef WIN32
    int outFileRefNum = open(file,O_WRONLY|O_TRUNC|O_CREAT);
#else
	int outFileRefNum = open(file,O_WRONLY|O_TRUNC|O_CREAT,STD_PERMISSIONS);
#endif

    if(outFileRefNum==-1)
      return NULL;
    else
      return _new_io(file,ioFile,ioWRITE,outFileRefNum,NULL,_flushFile,
		     _closeFile,_fileInReady,_fileOutReady,_configureFile);
  }
  default:
    return NULL;
  }
}

filePo icmOpenAppendFile(char *file)
{
  initICMIo();

  switch(regularFile(file)){
  case icmOk:
  case icmFailed:{			/* File not previously found */
#ifdef WIN32
	int outFileRefNum = open(file,O_CREAT|O_WRONLY|O_APPEND);
#else
	int outFileRefNum = open(file,O_CREAT|O_WRONLY|O_APPEND,STD_PERMISSIONS);
#endif

    if(outFileRefNum==-1)
      return NULL;
    else
      return _new_io(file,ioFile,ioWRITE,outFileRefNum,NULL,_flushFile,
		     _closeFile,_fileInReady,_fileOutReady,_configureFile);
  }
  default:
    return NULL;		/* directories cant be written to */
  }
}

filePo icmOpenInOutAppendFile(char *file)
{
  initICMIo();

  switch(regularFile(file)){
  case icmOk:
  case icmFailed:{			/* File not previously found */
#ifdef WIN32
    int outFileRefNum = open(file,O_RDWR|O_CREAT|O_APPEND);
#else
    int outFileRefNum = open(file,O_RDWR|O_CREAT|O_APPEND,STD_PERMISSIONS);
#endif

    if(outFileRefNum==-1)
      return NULL;
    else
      return _new_io(file,ioFile,ioWRITE|ioREAD,outFileRefNum,NULL,
		     _flushFile,_closeFile,_fileInReady,_fileOutReady,
		     _configureFile);
  }
  default:
    return NULL;		/* directories cant be written to */
  }
}

/* Open a pipe */

/* pipe data structure to allow the sub-process to be terminated correctly */
typedef struct _pipe_info_{
  int child;
  struct _pipe_info_ *other;
} pipeInfo;
#ifdef WIN32
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
#endif

static icmStatus _closePipe(filePo f,int cl);

static int _maxOpenFds()
{
#if HAVE_GETDTABLESIZE
  return getdtablesize();
#elif HAVE_GETRLIMIT
  return getrlimit(RLIMIT_NOFILE);
#elif HAVE_OPEN_MAX
  return OPEN_MAX
#endif
#ifdef WIN32
	  // We use an arbitrary value for ease
	  // Surely no more than 200 fds will be needed
	  return 200;
#endif
}

icmStatus icmOpenPipe(char *exec,char **argv,char **envv,filePo *inpipe,filePo *outpipe)
{
  int child,pipe1[2],pipe2[2];

#ifdef WIN32
  if(_pipe(pipe1, 256, _O_BINARY)<0 || _pipe(pipe2, 256, _O_BINARY)<0 || exec==NULL)
#else
  if(_pipe(pipe1)<0 || _pipe(pipe2)<0 || exec==NULL)
#endif
    return icmError;			/* Cant open the pipes properly */
#ifdef WIN32
	//Here we exec the program in one step using CreateProcess
  //execve(exec,argv,envv)

  GetStartupInfo(&si);

  CreateProcess(NULL, exec, //name
				NULL, //process security attributes
				NULL, //thread security attributes
				TRUE, //inherit handles from parent
				0,    //normal priority
				NULL, //same environment as parent
				NULL, //launch in current dir
				&si,  //startup info
				&pi); // process info stored upon return
#else
  else if((child=fork())<0)
    return icmError;			/* We can fork ... */
  else if(child>0){		/* We are the parent ... */
    pipeInfo *in = (pipeInfo*)malloc(sizeof(pipeInfo));
    pipeInfo *out = (pipeInfo*)malloc(sizeof(pipeInfo));

    in->child = child;
    in->other = out;

    out->child = child;
    out->other = in;

    *inpipe = _new_io(exec,ioFile,ioREAD,pipe2[0],
		      _refillFile,NULL,_closePipe,_fileInReady,_fileOutReady,
		      _configureFile);
    *outpipe = _new_io(exec,ioFile,ioWRITE,pipe1[1],
		       NULL,_flushFile,_closePipe,_fileInReady,_fileOutReady,
		       _configureFile);
    close(pipe1[0]);
    close(pipe2[1]);

    setFileClient(*inpipe,in);
    setFileClient(*outpipe,out);

    return icmOk;
  }
  else{				/* We are the child */
    close(pipe1[1]);
    close(pipe2[0]);		/* Close the unused portions */

    if(dup2(pipe2[1],1)<0 || dup2(pipe2[1],2)<0 ||
       dup2(pipe1[0],0)<0){
      exit(-29);
    }
    else{
      int i,maxFds = _maxOpenFds();

      for(i=3;i<maxFds;i++)
        close(i);               /* Close all other files ... */

      if(execve(exec,argv,envv)<0)	/* Execute the program */
	perror("Problem in executing");
    }
    return icmOk;		/* We wont get here */
  }
#endif
}

static icmStatus _closePipe(filePo f,int fno)
{
  pipeInfo *pipe = (pipeInfo*)fileClient(f);

  if(close(fno)==-1)
    return icmError;
  else{
    if(pipe->other==NULL){
      int stat;

#ifndef WIN32
	  //irrelevant as we're not reaping under windows - processes work differently.
	  if(waitpid(pipe->child,&stat,0)!=pipe->child)
	icmLogMsg("Problem when waiting for sub-process %d",pipe->child);
#endif

      free(pipe);
    }
    else{
      pipeInfo *other = pipe->other;

      other->other = NULL;
      free(pipe);
    }
      
    return icmOk;
  }
}


/* Legacy interfaces */

filePo openInFile(char *file)
{
  return icmOpenInFile(file);
}

filePo locateFile(char *name)
{
  return icmLocateFile(name);
}

filePo OpenStdin(void)
{
  return icmOpenStdin();
}

filePo OpenStdout(void)
{
  return icmOpenStdout();
}

filePo OpenStderr(void)
{
  return icmOpenStderr();
}

icmStatus seekFile(filePo f,long where,long mode)
{
  return icmSeekFile(f,where,mode);
}

filePo openOutfile(char *file)
{
  return icmOpenOutFile(file);
}

filePo openInOutFile(char *file)
{
  return icmOpenInOutFile(file);
}

filePo newOutfile(char *file)
{
  return icmNewOutFile(file);
}

filePo openAppendFile(char *file)
{
  return icmOpenAppendFile(file);
}

filePo openInOutAppendFile(char *file)
{
  return icmOpenInOutAppendFile(file);
}

icmStatus openPipe(char *exec,char **argv,char **envv,filePo *inpipe,filePo *outpipe)
{
  return icmOpenPipe(exec,argv,envv,inpipe,outpipe);
}
