/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/* Hash table functions 
 *
 * A generic collection of functions to build hash tables, search them and
 * add entries to them.
 */
#include "config.h"		/* Invoke configuration header */
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include "icmP.h"
#if DMALLOC
#include <dmalloc.h>		/* debugging malloc library */
#endif


/* The structure of a hash table ... */
typedef struct bucket *bucketPo;

typedef struct bucket{
  void *nme;			/* The symbol in the hash table */
  void *r;			/* 'value' of the hashed record */
  bucketPo link;		/* Link in case of hash-crash */
} BucketRec;

typedef struct _hashtable_{
  long size;			/* The number of entries in the table */
  long entries;			/* how many entries do we have? */
  bucketPo *table;		/* The table of entries */
  icmHashFun hash;		/* The hashing function */
  icmCompFun compare;		/* The comparison function */
  icmDestFun destroy;		/* Entry destruction function */
  pthread_mutex_t mutex;	/* Mutex controls access to the hash table */
} HashTableRec;

static void rehash(icmHashPo tbl);	/* forward declarations */
static long nextPrime(long n);

static icmPoolBase bpool = NULL;

/* Create a new table */
icmHashPo icmNewHash(long size,icmHashFun hash,icmCompFun compare,icmDestFun dest)
{
  long i;
  icmHashPo hp;

  hp = (icmHashPo)malloc(sizeof(HashTableRec));

  size = nextPrime(size);

  if(bpool==NULL)
    bpool = icmInitPool(sizeof(BucketRec),1024);

  hp->size = size;
  hp->entries = 0;		/* this will be used to determine rehashing */
  hp->table = (bucketPo*)malloc(size*sizeof(bucketPo));
  hp->hash = hash;
  hp->compare = compare;
  hp->destroy = dest;
  icmCreateMutex(&hp->mutex);

  for(i=0;i<size;i++)
    hp->table[i]=NULL;

  return hp;
}

static icmStatus delBuckets(icmHashPo ht,bucketPo *b)
{
  bucketPo old;
  icmStatus stat=icmOk;

  while(stat==icmOk){
    icmLockMutex(&ht->mutex);

    if(*b==NULL){
      icmUnlockMutex(&ht->mutex); /* we can release the lock now */
      return icmOk;
    }
    else{
      old = *b;			/* unlink the bucket from the chain */
      *b = old->link;

      icmUnlockMutex(&ht->mutex); /* we can release the lock now */

      if(ht->destroy!=NULL && (stat=(ht->destroy)(old->nme,old->r))!=icmOk){
	icmLockMutex(&ht->mutex);

	old->link = *b;
	*b = old;
	icmUnlockMutex(&ht->mutex); /* we can release the lock now */
	return stat;
      }

      icmFreePool(bpool,old);
    }
  }
  return stat;
}

icmStatus icmDelHash(icmHashPo ht)
{
  icmStatus stat = icmOk;
  register bucketPo *pp = ht->table;
  register int i=0;
  register int size = ht->size;

  for(i=0;stat==icmOk&&i<size;i++)
    if(pp[i]!=NULL)
      stat=delBuckets(ht,&pp[i]);

  if(stat!=icmOk)
    return stat;

  pthread_mutex_destroy(&ht->mutex);
  free(ht->table);
  free(ht);
  return icmOk;
}

/* Search the hash table */
void *icmSearch(void *name,register icmHashPo htbl)
{
  register bucketPo b = htbl->table[(htbl->hash)(name)%htbl->size]; 
  
  icmLockMutex(&htbl->mutex);

  while(b!=NULL){
    if((htbl->compare)(b->nme,name)==0){ /* we have found the entry */
      icmUnlockMutex(&htbl->mutex);
      return b->r;
    }
    else
      b = b->link;		/* down the chain */
  }

  icmUnlockMutex(&htbl->mutex);
  return NULL;			/* nothing there */
}

/* Install in the hash table */
icmStatus icmInstall(void *name, void *r,icmHashPo htbl)
{
  register long offset = (htbl->hash)(name)%htbl->size;
  register bucketPo b = htbl->table[offset];

  icmLockMutex(&htbl->mutex);

  while(b!=NULL){
    if((htbl->compare)(b->nme,name)==0){ /* we have found the entry */
      b->nme = name;
      b->r = r;
      icmUnlockMutex(&htbl->mutex);
      
      return icmOk;		/* we found an old entry */
    }
    else
      b = b->link;		/* down the chain */
  }

  if(htbl->entries>=htbl->size){ /* table is full */
    rehash(htbl);
    icmUnlockMutex(&htbl->mutex);
    return icmInstall(name,r,htbl);
  }
  else{
    b = icmAllocPool(bpool);	/* create a new entry in the chain */
    b->nme = name;
    b->r = r;
    b->link = htbl->table[offset];
    htbl->table[offset]=b;
    htbl->entries++;

    icmUnlockMutex(&htbl->mutex);
    return icmOk;
  }
}

/* remove an entry from the hash table */
icmStatus icmUninstall(void *name, icmHashPo htbl)
{
  register long offset = (htbl->hash)(name)%htbl->size;
  register bucketPo *b = &htbl->table[offset];

  icmLockMutex(&htbl->mutex);

  while(*b!=NULL){
    if((htbl->compare)((*b)->nme,name)==0){ /* we have found the entry */
      bucketPo old = *b;
      icmStatus stat=icmOk;

      *b = old->link;		/* remove the bucket from the table */

      icmUnlockMutex(&htbl->mutex);

      /* the destroy function is run in unlocked mode .. to allow further removals */
      if(htbl->destroy!=NULL && (stat=(htbl->destroy)(old->nme,old->r))!=icmOk){
	icmLockMutex(&htbl->mutex); /* put the entry back in */
	*b = old;
	icmUnlockMutex(&htbl->mutex);
	return stat;
      }

      icmFreePool(bpool,old);
      htbl->entries--;		/* decrement number of entries */
      return icmOk;
    }
    else
      b = &(*b)->link;
  }

  icmUnlockMutex(&htbl->mutex);
  return icmFailed;		/* not in the table... */
}

/* Rebuild the hash table when it has filled up */
/* Runs when the table mutex is locked */
static void rehash(icmHashPo tbl)
{
  long old_size = tbl->size;
  register long new_size = nextPrime(old_size*2);
  register long i;
  register bucketPo *old = tbl->table;

  tbl->size = new_size;

  if((tbl->table=(bucketPo*)calloc(new_size,sizeof(bucketPo)))==NULL){
    icmLogMsg("no space for hash table"); /* If not abort */
    exit(-1);
  }

  for(i=0;i<old_size;i++){
    if(old[i]!=NULL){
      bucketPo b = old[i];

      while(b!=NULL){
	register long offset = (tbl->hash)(b->nme)%tbl->size;
	register bucketPo bb = b; /* switch bucket from old table to new table */

	b = b->link;
	bb->link = tbl->table[offset];
	tbl->table[offset]=bb;
      }
    }
  }
  free(old);			/* release the old table */
}


/* Process the whole table
 * Additions and removals are permitted during this process
 * But a rehash is unsafe ... so insertions should be avoided
 */

icmStatus icmProcessTable(icmProcFun pr,icmHashPo htbl,void *c)
{
  register int i;
  register int size = htbl->size;
  icmStatus stat = icmOk;

  icmLockMutex(&htbl->mutex);

  for(i=0;i<size;i++){
    if(htbl->table[i]!=NULL){
      bucketPo b = htbl->table[i];
      while(b!=NULL){
	BucketRec B = *b;

	icmUnlockMutex(&htbl->mutex); /* run the proc while unlocked! */

	if((stat=pr(B.nme, B.r, c))!=icmOk)
	  return stat;

	icmLockMutex(&htbl->mutex); /* lock it up again */
	b = b->link;
      }
    }
  }

  icmUnlockMutex(&htbl->mutex); /* all done -- we can unlock now! */
  return icmOk;
}

/* this is not thread safe! */
icmTruth verifyHash(icmHashPo htbl)
{
  register int i;
  register int size = htbl->size;

  for(i=0;i<size;i++){
    if(htbl->table[i]!=NULL){
      bucketPo b = htbl->table[i];
      while(b!=NULL){
	if(icmSearch(b->nme,htbl)!=b->r)
	  return icmNo;

	b = b->link;
      }
    }
  }

  return icmYes;
}


/* Use the sieve of Erastosthenes to find the next prime that is larger than a given 
   number
   Not the fastest algorithm, but it isnt called very often
*/

typedef struct {
  long pr;
  long next;
} PrimeRec, *primePo;

static long nextPrime(long n)
{
  long i;
  long len = n/2+1;
  primePo table = (primePo)malloc(len*sizeof(PrimeRec));
  long pos = 0;
  long pr = 3;

  while(pr<=n){
    assert(pos<len);

    table[pos].pr=pr;
    table[pos++].next=2*pr;

    pr+=2;

    i=0;
    while(i<pos){
      if(pr>table[i].next)
	table[i].next+=table[i].pr;

      if(pr==table[i].next){
	table[i].next+=table[i].pr;
	pr+=2;
	i=0;			/* start again */
      }
      else
	i++;
    }
  }

  free(table);
  return pr;
}

