/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * Host name cacheing for ICM communications 
 *
 */
#include "config.h"		/* Invoke configuration header */
#include "icmP.h"
#include "icmIoP.h"
#ifdef WIN32
#include <winsock.h>
#include <windows.h>
#include <time.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>		/* Basic network database stuff */
#include <sys/time.h>		/* used for retiring host names */
#endif
#include <string.h>
#include <ctype.h>
#ifdef WIN32
#include <io.h>
#include <time.h>
#include <windows.h>
#else
#include <unistd.h>
#endif
#include <stdlib.h>

#define MAXALIAS 16
#define MAXHOST 16
#define MAXIP 10		/* Maximum number of IP addresses for one host */

#if 0
#define SECSINDAY 8		/* Use this version to test host-name caching */
#else
#define SECSINDAY 86400
#endif

#define RECHECK 60		/* We re-check unavailable hosts after 60 seconds */

typedef struct _host_record_ *hostPo;

typedef struct _host_record_ {
  char host[MAXSTRING];		/* Canonical name of this host */
  char *aliases[MAXALIAS];	/* list of aliases for this host */
  struct in_addr ip[MAXIP];	/* ip address list for this host */
  int ip_count;			/* How many addresses are there? */
  icmTruth avail;		/* True if this is a `positive' entry */
  long when;			/* Used to retire a host address */
} HostRecord;

static icmPoolBase hpool=NULL;
static hostPo ourHost = NULL;	/* This is our machine ... */
static icmHashPo htble = NULL;

static hostPo locateHost(char *name);
static long strhash(register char *name);
static icmStatus dispHost(filePo f,void *h,int width,int precision,icmTruth alt);

#ifdef WIN32
//Windows is dumb and doesn't support gettimeofday
int gettimeofday(struct timeval* tp, void* tzp) {
    DWORD t;
    t = timeGetTime();
    tp->tv_sec = t / 1000;
    tp->tv_usec = t % 1000;
    /* 0 indicates that the call succeeded. */
    return 0;
}
#endif

static void _initHosts(void)
{
  char mbuff[MAXSTRING];

  htble = icmNewHash(MAXHOST,(icmHashFun)strhash,(icmCompFun)strcmp,NULL);
  hpool = icmInitPool(sizeof(HostRecord),32);
  
  if(gethostname(mbuff,NumberOf(mbuff)))
    icmStrNCpy(mbuff,"localhost",NumberOf(mbuff));
        
  if((ourHost=locateHost(mbuff))==NULL)
    ourHost = locateHost("localhost");
  else if(ourHost->ip_count<NumberOf(ourHost->ip)){
    hostPo localHost = locateHost("localhost");
    ourHost->ip[ourHost->ip_count++] = localHost->ip[0];
  }

  _installMsgProc('H',dispHost); /* extend outMsg to cope with host names */
}

static void initHosts(void)
{
  static pthread_once_t hostInit = PTHREAD_ONCE_INIT;

  pthread_once(&hostInit,_initHosts); /* initialize the host DB just once */
}

/* compute the hash value of a string .. */
static long strhash(register char *name)
{
  register long hash = 0;

  if(name!=NULL)
    while(*name){
      hash = (hash<<4) + (hash^tolower(*name));
      name++;
    }
  return hash;
}

static void retireHost(hostPo h)
{
  int i;

  for(i=0;i<MAXALIAS && h->aliases[i]!=NULL;i++){
    icmUninstall(h->aliases[i],htble);
    free(h->aliases[i]);
  }

  icmUninstall(h->host,htble);
  icmFreePool(hpool,h);
}

/* This will have to be upgraded to IPv6 */
static icmTruth ipAddr(const char *addr,struct in_addr* ip)
{
  short quads[4];
  char *p = (char*)addr;
  int i;

  for(i=0;i<4;i++){
    quads[i]=strtol(p,&p,10);	/* look for a quad number */

    if(*p=='.')
      p++;
    else if(*p!='\0' && i!=3)
      return icmNo;
  }

  ip->s_addr = htonl(((quads[0]&0xff)<<24)|((quads[1]&0xff)<<16) | ((quads[2]&0xff)<<8) |
    (quads[3]&0xff));
  return icmYes;
}

void icmIPToStr(const struct in_addr *ip, char *addr,long len)
{
  unsigned long h = ntohl(ip->s_addr);

  icmStrMsg(addr,len,"%d.%d.%d.%d",(int)((h >> 24)&0xFF),
                        (int)((h >> 16)&0xFF),(int)((h >> 8)&0xFF),(int)(h&0xFF));
}

static hostPo locateHost(char *nme)
{
  static pthread_mutex_t host_mutex = PTHREAD_MUTEX_INITIALIZER;

  /* We need this because we might two overlapping searches */
  pthread_mutex_lock(&host_mutex);

  {
    char nameBuff[MAXSTRING],*name=icmLowerStr(nameBuff,nme,MAXSTRING);
    hostPo h = (hostPo)icmSearch(name,htble); /* We try our own database first */
    struct hostent *he;
    struct timeval now;
    struct in_addr ip;

    gettimeofday(&now,NULL);

    if(h!=NULL){
      if(now.tv_sec>h->when)	/* should be be retiring this host name? */
	retireHost(h);		/* we have to re-compute this host */
      else{
	icmTruth avail = h->avail;

	pthread_mutex_unlock(&host_mutex);

	if(avail)
	  return h;
	else{
#ifdef TRACEHOST  
	  icmLogMsg("%s marked as unavailable",nme);
#endif
	  return NULL;		/* It was a negative entry ....!!!! */
	}
      }
    }

    if(ipAddr(name,&ip) &&	/* We have a name expressed in numbers and dots */
       (he=gethostbyaddr((char*)&ip,sizeof(struct in_addr),AF_INET))!=NULL){
      int a = 0;
      char **al;

      h = (hostPo)icmAllocPool(hpool);

      h->ip_count = 0;
      h->when = now.tv_sec+SECSINDAY;
      h->avail = icmYes;

      icmLowerStr(h->host,(char*)he->h_name,MAXSTRING);

      icmInstall(h->host,h,htble);

      he = gethostbyname(h->host); /* make sure that we get the proper name */

      for(al=he->h_aliases;*al!=NULL;al++,a++){
	char alias[MAXSTRING];

	icmLowerStr(alias,*al,MAXSTRING);
	h->aliases[a]=icmDupStr(alias);
	icmInstall(h->aliases[a],h,htble);
      }

      for(al=he->h_addr_list;*al!=NULL;al++){ /* extract the IP numbers */
	char ip[MAXSTRING];
	icmIPToStr((struct in_addr*)(*al),ip,NumberOf(ip));
	h->ip[h->ip_count++]=*(struct in_addr*)(*al);
	h->aliases[a]=icmDupStr(ip);
	icmInstall(h->aliases[a++],h,htble); /* put the IP in as an alias */
      }

      h->aliases[a]=NULL;
      pthread_mutex_unlock(&host_mutex);
      return h;
    }
    else{
      /* We must now use the DNS to get the name */
      he = gethostbyname(name);	/* first stab ... */

      if(he!=NULL && he->h_addr_list[0]!=NULL){
	int a = 0,i;
	char **al;

	h = (hostPo)icmAllocPool(hpool);

	h->ip_count = 0;
	h->when = now.tv_sec+SECSINDAY;
	h->avail = icmYes;

	icmLowerStr(h->host,(char*)he->h_name,MAXSTRING);
      
	icmInstall(h->host,h,htble);

	for(al=he->h_aliases;*al!=NULL;al++,a++){
	  char alias[MAXSTRING];

	  icmLowerStr(alias,*al,MAXSTRING);
	  h->aliases[a]=icmDupStr(alias);
	  icmInstall(h->aliases[a],h,htble);
	}

	for(al=he->h_addr_list;*al!=NULL;al++){ /* extract the IP numbers */
	  char ip[MAXSTRING];
	  icmIPToStr((struct in_addr*)(*al),ip,NumberOf(ip));
	  h->ip[h->ip_count++]=*(struct in_addr*)(*al);
	  h->aliases[a]=icmDupStr(ip);
	  icmInstall(h->aliases[a++],h,htble); /* put the IP in as an alias */
	}

	if(strchr(h->host,'.')==NULL){ /* Use IP to get canonical name */
	  for(i=0;i<h->ip_count;i++){ /* We check each IP address ... */
	    he = gethostbyaddr((char*)&h->ip[0],sizeof(struct in_addr),AF_INET);

	    if(he!=NULL){
	      int j;

	      /* check out the new set of aliases */
	      for(al = he->h_aliases; *al!=NULL; al++){
		icmTruth found=icmNo;	

		for(j=0;j<a && !found;j++){
		  char alias[MAXSTRING];

		  icmLowerStr(alias,*al,MAXSTRING);
		  if(strcmp(alias,h->aliases[j]))
		    found=icmYes;
		}

		if(!found && a<MAXALIAS){
		  char alias[MAXSTRING];

		  icmLowerStr(alias,*al,MAXSTRING);
		  h->aliases[a]=icmDupStr(alias);
		  icmInstall(h->aliases[a++],h,htble);
		}
	      }
	    }
	  }
	}
    
	h->aliases[a]=NULL;
	pthread_mutex_unlock(&host_mutex);
	return h;
      }
      else{
	pthread_mutex_unlock(&host_mutex);
	return NULL;
      }
    }
  }
}

void icmMarkHostUnavail(char *name)
{
  hostPo h;

  initHosts();

  if((h=locateHost(name))!=NULL){
    struct timeval now;

    gettimeofday(&now,NULL);

    h->avail = icmNo;		/* host is off-line */
    h->when = now.tv_sec+RECHECK; /* when do we re-check the host */
  }
}
  

char *icmGetHostname(char *name)
{
  hostPo h;

  initHosts();

  if((h=locateHost(name))==NULL)
    return NULL;
  else
    return h->host;
}

struct in_addr *icmGetHostIP(char *name)
{
  hostPo h;

  initHosts();

  if((h=locateHost(name))==NULL)
    return NULL;
  else
    return &h->ip[0];
}  

char *icmGetNthHostIP(char *name,long i,char *buffer,long len)
{
  hostPo h;

  initHosts();

  if((h=locateHost(name))==NULL)
    return NULL;
  else if(i<h->ip_count){
    char ip[MAXSTRING];
    icmIPToStr(&h->ip[i],ip,NumberOf(ip));
    icmStrNCpy(buffer,ip,len);

    return buffer;
  }
  else
    return NULL;
}

icmStatus icmIsIPofHost(char *name,unsigned long ip)
{
  hostPo h;

  if((h=locateHost(name))==NULL)
    return icmError;
  else{
    int i;

    for(i=0;i<h->ip_count;i++)
      if(ip==h->ip[i].s_addr)
	return icmOk;
    return icmFailed;
  }
}

/* What is our name? */
char *icmMachineName(void)
{
  initHosts();

  if(ourHost==NULL){
    icmLogMsg("failed to get machine name");
    return NULL;
  }
  else
    return ourHost->host;
}

/* What is our first IP number? */
char *icmMachineIP(void)
{
  initHosts();

  if(ourHost==NULL){
    icmLogMsg("failed to get machine name");
    return NULL;
  }
  else{
    static char IP[MAXSTRING];
    icmIPToStr(&ourHost->ip[0],IP,NumberOf(IP));
    return IP;
  }
}

static icmStatus dispHost(filePo f,void *h,int width,int precision,icmTruth alt)
{
  hostPo host = (hostPo)h;
  int i;
  char *sep = "(";
  char buff[MAXSTRING] = "";

  icmStrNCpy(buff,host->host,MAXSTRING);

  if(host->aliases[0]!=NULL){
    for(i=0;host->aliases[i]!=NULL;i++){
      icmStrNCat(buff,sep,MAXSTRING);
      icmStrNCat(buff,host->aliases[i],MAXSTRING);
      sep = ",";
    }
    sep = ")@";
  }
  else
    sep = "@";

  for(i=0;i<host->ip_count;i++){
    char ip[MAXSTRING];
    icmIPToStr(&host->ip[i],ip,NumberOf(ip));
    icmStrNCat(buff,sep,MAXSTRING);
    icmStrNCat(buff,ip,MAXSTRING);
    sep = ",@";
  }
  return outString(f,buff,strlen(buff),width,precision,' ',alt,"");
}


#ifdef TRACEHOST  
static void dH(void *n,void *r,void *c)
{
  char *name = (char*)n;
  hostPo host = (hostPo)r;
  filePo f = (filePo)c;
  int i;

  outMsg(f,"%s:\n",name);
  for(i=0;host->aliases[i]!=NULL;i++)
    outMsg(f," -> %s\n",host->aliases[i]);

  for(i=0;i<host->ip_count;i++) {
    char ip[MAXSTRING];
    icmIPToStr(&host->ip[i],ip,NumberOf(ip));
    outMsg(f," @ %s\n",ip);
  }
}

void dumpHosts()
{
  icmProcessHash(dH,htble,icmLogFile);
  flushFile(icmLogFile);
}
#endif
