/* 
 * I/O handling for the ICM library
 * Imperial College
 * (c) 1994-1998 Imperial College and F.G. McCabe
 * All rights reserved
 * This library should be thread-safe ... (famous last words)
 */ 

#include "config.h"		/* Invoke configuration header */
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include "icmP.h"
#include "icmIoP.h"
#include <assert.h>
#if DMALLOC
#include <dmalloc.h>		/* debugging malloc library */
#endif

#ifndef IOMAX
#define IOMAX	FD_SETSIZE
#endif

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static IoEntry io_table[IOMAX];

static icmStatus nullRefiller(filePo f,int cl);
static icmStatus nullFlusher(filePo f,int cl);
static icmStatus nullCloser(filePo f,int cl);
static icmStatus nullConfigureIo(filePo io,icmIoConfigOpt mode,int client);

static void initIo(void);

void initICMIo(void)
{
  static pthread_once_t initOnce = PTHREAD_ONCE_INIT;

  pthread_once(&initOnce,initIo);
}

static void initIo(void)
{
  int i;

  for (i=0; i<IOMAX; i++){
    io_table[i].mode = ioNULL;
    io_table[i].in_line = NULL;
    io_table[i].out_line = NULL;
    icmCreateMutex(&io_table[i].i_mutex);
    icmCreateMutex(&io_table[i].o_mutex);
  }

  _initMsgProcs_();

  OpenStdin();			/* open standard files */
  OpenStdout();
  OpenStderr();

}

void closeICMIo(void)
{
  int i;
  
  icmLockMutex(&mutex);
  for(i=0;i<IOMAX;i++)
    if(io_table[i].mode!=ioNULL)
      closeFile(&io_table[i]);
  icmUnlockMutex(&mutex);
}
  
/* File descriptor management */
/*
 * new_io()
 * 
 * find a free descriptor in the table
 */

filePo _new_io(char *name,ioType type,int mode,int client,
	       fileProc refill,fileProc flusher,fileProc clser,
	       fileProc readyin,fileProc readyout,fileConfProc configure)
{
  int i;

  icmLockMutex(&mutex);
  for (i=0; i<IOMAX; i++)
    if (io_table[i].mode == ioNULL) {
      icmStrNCpy(io_table[i].filename,name,MAXFILELEN);
      io_table[i].mode = mode;
      io_table[i].type = type;
      
      io_table[i].line_number=1;
      io_table[i].in_pos=0;
      io_table[i].in_len=0;
      io_table[i].in_line = (char*)malloc(sizeof(char)*MAXLINE);
      io_table[i].status = icmOk;
      io_table[i].readyin = readyin;
      io_table[i].f_pos = 0;
      io_table[i].out_pos = 0;
      io_table[i].out_line = (char*)malloc(sizeof(char)*MAXLINE);
      io_table[i].depth = 32767;
      io_table[i].readyout = readyout;

      io_table[i].client = client;
      io_table[i].client2 = NULL;

      if(refill!=NULL)
	io_table[i].refill = refill;
      else
	io_table[i].refill = nullRefiller;

      if(flusher!=NULL)
	io_table[i].flusher = flusher;
      else
	io_table[i].flusher = nullFlusher;

      if(clser!=NULL)
	io_table[i].closer = clser;
      else
	io_table[i].closer = nullCloser;

      if(configure!=NULL)
	io_table[i].configure = configure;
      else
	io_table[i].configure = nullConfigureIo;

      io_table[i].refCount = 1;

      icmUnlockMutex(&mutex);
      return &io_table[i];
    }
  icmUnlockMutex(&mutex);
  return NULL;
}

long icmIoRefCount(filePo f)
{
  return f->refCount;
}

void icmIncRefCount(filePo f)
{
  f->refCount++;
}

void icmDecRefCount(filePo f)
{
  f->refCount--;
}



/* Access macros & functions */
long fileNo(filePo f)
{
  return f-io_table;
}

filePo filePtr(long n)
{
  return &io_table[n];
}

char *icmFileName(filePo f)
{
  return f->filename;
}

int icmFileData(filePo f)
{
  return f->client;
}

void *icmFileClient(filePo f)
{
  return f->client2;
}

void icmSetFileClient(filePo f,void *c)
{
  f->client2 = c;
}

ioType icmFileType(filePo f)
{
  return f->type;
}

long icmFilePos(filePo f)
{  
  icmLockMutex(&f->i_mutex);

  {
    long pos = f->f_pos;

    icmUnlockMutex(&f->i_mutex);
    return pos;
  }
}

long icmFileOutPos(filePo f)
{
  icmLockMutex(&f->o_mutex);

  {
    long out_pos = f->out_pos;

    icmUnlockMutex(&f->o_mutex);
    return out_pos;
  }
}

long icmFileLineNo(filePo f)
{
  icmLockMutex(&f->i_mutex);

  {
    long line_number = f->line_number;

    icmUnlockMutex(&f->i_mutex);
    return line_number;
  }
}

ioState icmFileMode(filePo f)
{
  return f->mode;
}

icmStatus icmIsFileOpen(filePo f)
{
  if(f==NULL)
    return icmError;
  else{
    ioState mode = fileMode(f);

    if(mode!=ioNULL)
      return icmOk;
    else
      return icmFailed;
  }
}

icmStatus icmIsReadingFile(filePo f)
{
  if(f==NULL)
    return icmError;
  else if((fileMode(f)&ioREAD)!=0)
    return icmOk;
  else
    return icmFailed;
}

icmStatus icmIsWritingFile(filePo f)
{
  if(f==NULL)
    return icmError;
  else if((fileMode(f)&ioWRITE)!=0)
    return icmOk;
  else
    return icmFailed;
}

int filePrintDepth(filePo f)
{
  icmLockMutex(&f->o_mutex);

  {
    long depth = f->depth;

    icmUnlockMutex(&f->o_mutex);
    return depth;
  }
}

int setFilePrintDepth(filePo f,int depth) /* Current print depth */
{
  icmLockMutex(&f->o_mutex);

  {
    int current = f->depth;
    f->depth = depth;

    icmUnlockMutex(&f->o_mutex);
    return current;
  }
}

/* File opening is specific to the type of file being opened, 
 * but all files can be closed using the same function
 */

icmStatus icmCloseFile(filePo f)
{
  if(f==NULL)
    return icmError;
  else if(--f->refCount<=0){
    icmStatus ret=icmError;

    icmLockMutex(&f->o_mutex);

    if((f->mode & ioWRITE)!=0){
      (f->configure)(f,turnOnBlocking,f->client);

	while((ret=(f->flusher)(f,f->client))==icmFailed)
	  ;
    
    }

    ret = (f->closer)(f,f->client);

    f->mode = ioNULL;		/* no one else can look in here ... */
    if(f->in_line!=NULL){
      free(f->in_line);
      f->in_line = NULL;
    }
    if(f->out_line!=NULL){
      free(f->out_line);
      f->out_line = NULL;
    }
    icmUnlockMutex(&f->o_mutex);
    return ret;			/* closing a file twice raises an error */
  }
  else
    return icmOk;		/* ignore close attempt */
}

/* test that a file is ready without actually reading anything */
icmStatus icmIsInReady(filePo fe)
{
  if(fe==NULL)
    return icmError;
  else{
    icmStatus stat;
    icmLockMutex(&fe->i_mutex);

    if((fe->mode&ioREAD)!=0){
      if(fe->status==icmEndFile)
	stat = icmOk;		/* End of file is a `char' */
      else if(fe->in_pos!=fe->in_len)
	stat = icmOk;
      else if(fe->readyin!=NULL)
	stat = fe->readyin(fe,fe->client);
      else
	stat = icmFailed;	/* Dont try to refill */
    }
    else
      stat = icmError;

    icmUnlockMutex(&fe->i_mutex);
    return stat;
  }
}

icmStatus icmIsOutReady(filePo f)
{
  if(f==NULL)
    return icmError;
  else{
    icmStatus stat = icmOk;

    icmLockMutex(&f->o_mutex);

    if((f->mode&ioWRITE)!=0){
      if(f->out_pos==MAXLINE){		/* need to flush output? */
	if(f->readyout!=NULL)
	  stat = f->readyout(f,f->client);
	else
	  stat = f->flusher(f,f->client);
      }
    }
    else
      stat = icmError;

    icmUnlockMutex(&f->o_mutex);
    return stat;
  }
}

/* Character reading and unreading functions */
int icmInCh(filePo f)
{
  char ch;

  switch(icmInBlock(f,&ch,1)){
  case icmEndFile:
    return EOF;
  case icmOk:
    return ch&0xff;
  default:
    return EOF;
  }
}

icmStatus icmInChar(filePo f,char *chP)
{
  long count;
  return icmInBlck(f,chP,1,&count);
}

/* Blocked version of read a block of chars */
icmStatus icmInBlock(filePo f,char *buffer,long len)
{
  icmStatus stat = icmOk;

  assert(f!=NULL);

  icmLockMutex(&f->i_mutex);

  if((f->mode&ioREAD)!=0){
    while(stat==icmOk && len-->0){
      if(f->in_pos==f->in_len && f->status!=icmEndFile){
	if((stat=(f->refill)(f,f->client))==icmFailed){
				/* Only happen if file is non-blocking */
	  (f->configure)(f,turnOnBlocking,f->client);
	  icmUnlockMutex(&f->i_mutex);
	  flushOutChannels();	/* Flush out any prompts */
	  icmLockMutex(&f->i_mutex);
	  while((stat=(f->refill)(f,f->client))==icmFailed)
	    ;			/* Normally this will only be executed once */
	  (f->configure)(f,turnOffBlocking,f->client);
	}
      }

      if(f->status==icmEndFile || f->in_pos==f->in_len)
	stat=icmEndFile;	/* no point in further loop processing */
      else{
	unsigned int ch = f->in_line[f->in_pos++]&0xff;
	
	if(ch==NEW_LINE)
	  f->line_number++;
	f->f_pos++;
	*buffer++=ch;
      }
    }
  }
  else
    stat = icmError;

  icmUnlockMutex(&f->i_mutex);
  return stat;
}

/* Un-blocked version of read a block of chars */
icmStatus icmInBlck(filePo f,char *buffer,long len,long *ret)
{
  icmStatus stat = icmError;

  icmLockMutex(&f->i_mutex);

  if((f->mode&ioREAD)!=0){
    int count=0;

    stat = icmOk;		/* We are ok now */

    while(stat==icmOk && len-->0){
      if(f->in_pos==f->in_len && f->status!=icmEndFile){
	if((stat=(f->refill)(f,f->client))==icmFailed)
	  break;		/* Only happen if file is non-blocking */
      }

      if(f->status==icmEndFile || f->in_pos==f->in_len){
	stat=icmEndFile;
	break;			/* no point in further loop processing */
      }
      else{
	unsigned int ch = f->in_line[f->in_pos++]&0xff;
	
	if(ch==NEW_LINE)
	  f->line_number++;
	f->f_pos++;
	*buffer++=ch;
	count++;		/* increment number of character read */
	continue;
      }
    }

    if(ret!=NULL)
      *ret = count;		/* determine actual number of bytes read */

    switch(stat){
    case icmFailed:
    case icmEndFile:
      if(count!=0)
	stat = icmOk;
    case icmOk:
    case icmError:
      ;
    }
  }

  icmUnlockMutex(&f->i_mutex);
  return stat;
}

/*
 * unblocked read a line ... read up to a terminating character 
 * len should be at least 2 ... one for the final NULL byte
 */
icmStatus icmInLine(filePo f,char *buffer,long len,char *term)
{
  icmStatus stat = icmOk;

  icmLockMutex(&f->i_mutex);

  if((f->mode&ioREAD)!=0){
    long count = 0;

    while(stat==icmOk && --len>0){ /* we need at least one char for the NULL */
      if(f->in_pos==f->in_len && f->status!=icmEndFile){
	if((stat=(f->refill)(f,f->client))==icmFailed)
	  break;		/* Only happen if file is non-blocking */
      }

      if(f->status==icmEndFile && f->in_pos==f->in_len){
	stat=icmEndFile;
	break;			/* no point in further loop processing */
      }
      else{
	unsigned int ch = f->in_line[f->in_pos++]&0xff;
	
	if(ch==NEW_LINE)
	  f->line_number++;
	f->f_pos++;
	count++;

	if(strchr(term,ch)==NULL) /* have we found a terminating byte? */
	  *buffer++=ch;
	else
	  break;
      }
    }

    *buffer = '\0';		/* terminate the buffer */
  }
  else
    stat = icmError;

  icmUnlockMutex(&f->i_mutex);
  return stat;
}

icmStatus icmIsFileAtEof(filePo f)		/* icmYes if at end of file */
{
  if(f==NULL)
    return icmError;
  else
    return f->status;
}

icmStatus icmUnGetCh(filePo f, int ch)
{
  if(f!=NULL && ch!=EOF){
    icmStatus ret = icmOk;

    icmLockMutex(&f->i_mutex);

    if((f->mode&ioREAD)!=0){
      if(f->in_pos!=0 && f->in_line[f->in_pos-1]==ch){
	f->in_pos--;
	if(ch==NEW_LINE)
	  f->line_number--;
	f->f_pos --;		/* decrement file position also */
      }
      else
	ret = icmFailed;	/* we dont allow backtracking past buffer */
    }
    else
      ret = icmError;		/* tried to backtrack on a non-input file */

    icmUnlockMutex(&f->i_mutex);
    return ret;
  }
  else
    return icmEndFile;
}

icmStatus icmIsCharsWaiting(filePo fe)
{
  if(fe!=NULL){
    icmStatus stat = icmOk;

    icmLockMutex(&fe->i_mutex);

    if((fe->mode&ioREAD)!=0){
      if(fe->status!=icmEndFile && fe->in_pos==fe->in_len)
	switch((fe->refill)(fe,fe->client)){	/* try to re-fill the buffer */
	case icmOk:
	  if(fe->in_pos>=fe->in_len)
	    stat = icmFailed;	/* no data waiting */
	  break;
	case icmEndFile:	/* EOF is allowed here */
	  break;
	case icmFailed:
	  stat = icmFailed;
	  break;
	case icmError:
	  stat = icmError;
	}
    }
    else
      stat = icmError;

    icmUnlockMutex(&fe->i_mutex);
    return stat;
  }
  else
    return icmError;		/* We have a problem */
}

icmStatus icmSkipBlanks(filePo f)
{
  if(f!=NULL){
    icmStatus stat = icmOk;

    icmLockMutex(&f->i_mutex);

    if((f->mode&ioREAD)!=0){

      if(f->status!=icmOk)
	stat = f->status;	/* End of file is a `char' */
      else{
	while(f->in_pos<f->in_len && iscntrl((unsigned char)f->in_line[f->in_pos]))
	  f->in_pos++;
	if(f->in_pos!=f->in_len)
	  stat = icmOk;
	else{
	  icmUnlockMutex(&f->i_mutex);
	  return icmIsCharsWaiting(f); /* special exit here */
	}
      }
    }
    else
      stat = icmError;		/* file not open for reading */
    icmUnlockMutex(&f->i_mutex);
    return stat;
  }
  else
    return icmError;
}

icmStatus icmFlushFile(filePo f)
{
  if(f!=NULL){
    icmStatus stat;

    icmLockMutex(&f->o_mutex);

    if((f->mode&ioWRITE)!=0)
      stat = (f->flusher)(f,f->client);
    else
      stat = icmError;

    icmUnlockMutex(&f->o_mutex);

    return stat;
  }
  else
    return icmError;
}

void icmFlushOutChannels(void)
{
  int i;

  icmLockMutex(&mutex);
  for (i=0; i<IOMAX; i++)
    if((io_table[i].mode&ioWRITE)!=0)
      flushFile(&io_table[i]);
  icmUnlockMutex(&mutex);
}

/* Blocking version of write a character */
icmStatus icmOutCh(filePo f,int ch)
{
  icmStatus ret = outChar(f,ch);

  switch(ret){
  case icmFailed:{
    icmLockMutex(&f->o_mutex);

    (f->configure)(f,turnOnBlocking,f->client);

    while((f->flusher)(f,f->client)==icmFailed)	/* Just keep on going */
      ;				/* Normally executed once */
    (f->configure)(f,turnOffBlocking,f->client);

    if(f->out_pos<MAXLINE){	/* Just in case ... */
      f->out_line[f->out_pos++]=ch;
      f->f_pos++;		/* increment seek pointer */
      ret = icmOk;
    }
    else
      ret = icmError;		/* shouldnt happen */

    icmUnlockMutex(&f->o_mutex);

    return ret;
  }
  default:
    return ret;
  }
}

icmStatus icmOutChar(filePo f,int ch)
{
  if(f!=NULL){			/* writes to NULL are discarded */
    icmStatus ret = icmOk;

    icmLockMutex(&f->o_mutex);

    if((f->mode&ioWRITE)!=0){
      if(f->out_pos==MAXLINE)	/* need to flush output? */
	ret=(f->flusher)(f,f->client);

      if(ret==icmOk && f->out_pos<MAXLINE){ /* Just in case ... */
	f->out_line[f->out_pos++]=ch;
	f->f_pos++;		/* increment seek pointer */
      }
      else
	ret = icmFailed;
    }
    else
      ret = icmError;

    icmUnlockMutex(&f->o_mutex);
    return ret;
  }
  else
    return icmError;			/* funny thing happened here */
}

icmStatus icmUnOutCh(filePo f,long i)
{
  if(f!=NULL){
    icmStatus ret = icmOk;

    icmLockMutex(&f->o_mutex);

    if((f->mode&ioWRITE)!=0){
      if(f->out_pos>=i){
	f->out_pos-=i;
	f->f_pos-=i;
      }
      else
	ret = icmFailed;
    }
    else
      ret = icmError;
    icmUnlockMutex(&f->o_mutex);
    return ret;
  }
  else
    return icmError;
}

/*
 * Non-blocking version of outBlock ... write as many characters as possible 
 * returning the actual number written 
 */

icmStatus icmOutBlck(filePo f,char *data,long len,long *actual)
{
  int i;
  icmStatus ret = icmOk;

  if(len>0){
    icmLockMutex(&f->o_mutex);

    if((f->mode&ioWRITE)!=0){
      for(i=0;ret==icmOk && i<len;i++){
	if(f->out_pos==MAXLINE){ /* need to flush output? */
	  ret=(f->flusher)(f,f->client);

	  switch(ret){
	  case icmOk:
	    break;
	  case icmFailed:
	    if(i!=0){
	      ret = icmOk;
	      len = i;
	    }
	    break;
	  case icmError:
	  case icmEndFile:
	    len = i;		/* length is cut short on error */
	  }
	}
	if(ret==icmOk && i<len && f->out_pos<MAXLINE){	/* Just in case ... */
	  f->out_line[f->out_pos++]=data[i];
	  f->f_pos++;		/* increment seek pointer */
	}
      }
    }
    else
      ret = icmError;

    icmUnlockMutex(&f->o_mutex);

    if(actual!=NULL)
      *actual = len;
  }
  return ret;
}

/* Blocking version of outBlock 
 * even if the file is set to nonblocking mode, this will force the file to be
 * written out
 */

icmStatus icmOutBlock(filePo f,char *data,long len)
{
  int i;
  icmStatus ret = icmOk;

  if(len>0){
    icmLockMutex(&f->o_mutex);

    if((f->mode&ioWRITE)!=0){
      for(i=0;ret==icmOk && i<len;i++){
	if(f->out_pos==MAXLINE){	/* need to flush output? */
	  if((ret=(f->flusher)(f,f->client))==icmFailed){
	    (f->configure)(f,turnOnBlocking,f->client);
	    
	    while((f->flusher)(f,f->client)==icmFailed)	/* Just keep on going */
	      ;				/* Normally executed once */
	    (f->configure)(f,turnOffBlocking,f->client);
	    ret = icmOk;
	  }
	}
	if(ret==icmOk && f->out_pos<MAXLINE){	/* Just in case ... */
	  f->out_line[f->out_pos++]=data[i];
	  f->f_pos++;		/* increment seek pointer */
	}
      }
    }

    icmUnlockMutex(&f->o_mutex);
  }

  return ret;
}

void icmDumpIO(filePo f)
{
  if(f!=NULL){
    outMsg(icmLogFile,"Dump of file: %s\n",f->filename);
    outMsg(icmLogFile,"Mode = %s%s ",(((f->mode)&ioREAD)?"read ":""),
	   (((f->mode)&ioWRITE)?"write":""));
    outMsg(icmLogFile,"Type = %s\n",f->type==ioFile?"file":
	   f->type==ioChar?"char":f->type==ioBlock?"block":
	   f->type==ioDir?"dir":f->type==ioPipe?"pipe":
	   f->type==ioString?"string":"unknown");

    if(f->mode&ioREAD){
      int i;
      outMsg(icmLogFile,"Input buffer: `");
      for(i=f->in_pos;i<f->in_len;i++)
	outCh(icmLogFile,f->in_line[i]);
      outMsg(icmLogFile,"'@%d/%d\n",f->line_number,f->f_pos);
    }

    if(f->mode&ioWRITE){
      int i;
      outMsg(icmLogFile,"Output buffer: `");
      for(i=0;i<f->out_pos;i++)
	outCh(icmLogFile,f->out_line[i]);
      outMsg(icmLogFile,"'@%d\n",f->f_pos);
    }

  }
  else
    outMsg(icmLogFile,"Null-io");
  flushFile(icmLogFile);
}


void icmTriggerIo(fileProc filter,ioType type,int cl)
{
  int i;

  icmLockMutex(&mutex);
  for(i=0;i<IOMAX;i++)
    if(io_table[i].mode!=ioNULL && io_table[i].type==type)
      filter(&io_table[i],cl);
  icmUnlockMutex(&mutex);
}

filePo icmFindIo(ioType type,int cl)
{
  int i;

  icmLockMutex(&mutex);
  for(i=0;i<IOMAX;i++)
    if(io_table[i].mode!=ioNULL && io_table[i].type==type && io_table[i].client==cl)
      return &io_table[i];
  icmUnlockMutex(&mutex);

  return NULL;
}

icmStatus icmConfigureIo(filePo f,icmIoConfigOpt mode)
{
  if(f!=NULL){
    icmLockMutex(&f->i_mutex);
    icmLockMutex(&f->o_mutex);

    {
      icmStatus ret = (f->configure)(f,mode,f->client);

      icmUnlockMutex(&f->i_mutex);
      icmUnlockMutex(&f->o_mutex);

      return ret;
    }
  }
  else
    return icmError;
}

static icmStatus nullConfigureIo(filePo io,icmIoConfigOpt mode,int cl)
{
  return icmError;
}

static icmStatus nullRefiller(filePo f,int cl)
{
  return icmEndFile;
}

static icmStatus nullFlusher(filePo f,int cl)
{
  return icmOk;
}

static icmStatus nullCloser(filePo f,int cl)
{
  return icmOk;
}

/* Legacy interface implementation */

icmStatus fileOpen(filePo f)
{
  return icmIsFileOpen(f);
}

icmStatus readingFile(filePo f)
{
  return icmIsReadingFile(f);
}

icmStatus writingFile(filePo f)
{
  return icmIsWritingFile(f);
}

icmStatus inReady(filePo fe)
{
  return icmIsInReady(fe);
}

icmStatus outReady(filePo f)
{
  return icmIsOutReady(f);
}

icmStatus fileAtEof(filePo f)		/* icmYes if at end of file */
{
  return icmIsFileAtEof(f);
}

icmStatus closeFile(filePo f)
{
  return icmCloseFile(f);
}

int inCh(filePo f)
{
  return icmInCh(f);
}

icmStatus inChar(filePo f,char *chP)
{
  return icmInChar(f,chP);
}

icmStatus inBlock(filePo f,char *buffer,long len)
{
  return icmInBlock(f,buffer,len);
}

icmStatus inBlck(filePo f,char *buffer,long len,long *ret)
{
  return icmInBlck(f,buffer,len,ret);
}

icmStatus inLine(filePo f,char *buffer,long len,char *term)
{
  return icmInLine(f,buffer,len,term);
}

int unGetCh(filePo f, int ch)
{
  return icmUnGetCh(f,ch);
}

icmStatus charsWaiting(filePo fe)
{
  return icmIsCharsWaiting(fe);
}

icmStatus skipBlanks(filePo f)
{
  return icmSkipBlanks(f);
}

icmStatus flushFile(filePo f)
{
  return icmFlushFile(f);
}

void flushOutChannels(void)
{
  icmFlushOutChannels();
}

icmStatus outCh(filePo f,int ch)
{
  return icmOutCh(f,ch);
}

icmStatus outChar(filePo f,int ch)
{
  return icmOutChar(f,ch);
}

icmStatus unOutCh(filePo f,long i)
{
  return icmUnOutCh(f,i);
}

icmStatus outBlck(filePo f,char *data,long len,long *actual)
{
  return icmOutBlck(f,data,len,actual);
}

icmStatus outBlock(filePo f,char *data,long len)
{
  return icmOutBlock(f,data,len);
}

void triggerIo(fileProc filter,ioType type,int cl)
{
  icmTriggerIo(filter,type,cl);
}

filePo findIo(ioType type,int cl)
{
  return icmFindIo(type,cl);
}

icmStatus configureIo(filePo f,icmIoConfigOpt mode)
{
  return icmConfigureIo(f,mode);
}

char *fileName(filePo f)
{
  return f->filename;
}

int fileData(filePo f)
{
  return f->client;
}

void *fileClient(filePo f)
{
  return f->client2;
}

void setFileClient(filePo f,void *c)
{
  f->client2 = c;
}

ioType fileType(filePo f)
{
  return f->type;
}

long filePos(filePo f)
{  
  return icmFilePos(f);
}

long fileOutPos(filePo f)
{
  return icmFileOutPos(f);
}

long fileLineNo(filePo f)
{
  return icmFileLineNo(f);
}

ioState fileMode(filePo f)
{
  return icmFileMode(f);
}

