/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * Message buffer handling functions for the ICM system
 * Part of the ICM message handling sub-system
 *
 */
#include "config.h"		/* Invoke configuration header */
#include <assert.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>

#include "icmSig.h"		/* Access the signature definitions */
#include "icmP.h"		/* Private message handling interface */
#include "icmIoP.h"
#if DMALLOC
#include <dmalloc.h>		/* debugging malloc library */
#endif

#define ICM_TAG_MASK 0xF0
#define ICM_VAL_MASK 0x0F

#define icm_tag(c) ((c)&ICM_TAG_MASK)
#define icm_val(c) ((c)&ICM_VAL_MASK)

/* For testing purposes */
#if 1
#define PKTSIZE 64
#endif

#ifndef PKTSIZE
#define PKTSIZE MAXSTRING
#endif

#ifdef COMMTRACE
icmTruth traceMsg = icmNo;
#endif

static icmPoolBase mpool=NULL;
static icmPoolBase opool = NULL;

static icmStatus extendMsg(icmMsg msg,long amount);
static icmStatus displayMsg(filePo f,icmMsg msg,int depth,icmTruth alt,long x,long *nx);
static icmStatus dispMsg(filePo f,void *msg,int width,int precision,icmTruth alt);
static icmStatus dispOpts(filePo f,void *msg,int width,int precision,icmTruth alt);
static char *outSig(filePo f,char *sig);

static void initMsgs(void)
{
  mpool = icmInitPool(sizeof(icmMsgRec),256);
  opool = icmInitPool(sizeof(icmOptRecord),32);
  _installMsgProc('M',dispMsg);	/* extend outMsg to cope with msg buffers */
  _installMsgProc('O',dispOpts); /* extend outMsg to cope with options */
}

/* Create a new message structure */
icmStatus icmNewMsg(icmMsg *p)
{
  static pthread_once_t msgInit = PTHREAD_ONCE_INIT;

  pthread_once(&msgInit,initMsgs); /* initialize the message pool just once */

  {
    icmMsg m = (icmMsg)icmAllocPool(mpool);

    if(m!=NULL){
      m->used = 0;
      m->size = MAXSTRING;
      m->data = &m->buffer[0];

      *p = m;

      return icmOk;
    }
    else{
      *p = NULL;
      return icmError;
    }
  }
}

void icmResetMsg(icmMsg m)
{
  m->used = 0;
  if(m->data!=m->buffer)
    free(m->data);
  m->data = m->buffer;
  m->size = MAXSTRING;
}

void icmClearMsg(icmMsg m)
{
  m->used = 0;
  m->data = m->buffer;
  m->size = MAXSTRING;
}

void icmSetMsg(icmMsg m,Uchar *data,long size)
{
  m->used = size;
  if(size<=MAXSTRING){
    m->data = m->buffer;
    m->size = MAXSTRING;
  }
  else{
    m->data = (Uchar*)malloc(size*sizeof(Uchar));
    m->size = size;
  }
  memmove(m->data,data,size*sizeof(Uchar));
}

icmStatus icmCmpMsg(icmMsg m1,icmMsg m2)
{
  if(m1==m2)
    return icmOk;
  else if(m1!=NULL && m2!=NULL){
    if(m1->used == m2->used && memcmp(m1->data,m2->data,m1->used)==0)
      return icmOk;
    else
      return icmFailed;
  }
  else
    return icmError;
}

icmMsg icmDupMsg(icmMsg m)
{
  icmMsg new;

  icmNewMsg(&new);

  extendMsg(new,m->used);

  memmove(new->data,m->data,m->used);

  return new;
}

static icmStatus extendMsg(icmMsg msg,long amount)
{
  if(msg->used+amount>=msg->size){
    long newSize = msg->size+(msg->size>>1)+amount;
    Uchar *new;

    if(msg->data==&msg->buffer[0]){
      new = malloc(newSize*sizeof(Uchar));
      memmove(new,msg->data,msg->used*sizeof(Uchar));
    }
    else
      new = (Uchar*)realloc(msg->data,newSize*sizeof(Uchar));

    if(new==NULL)
      return icmError;
    else{
      msg->data = new;
      msg->size = newSize;
    }
  }
  return icmOk;
}

icmStatus icmReleaseMsg(icmMsg msg)
{
  if(msg->data!=&msg->buffer[0])
    free(msg->data);

  icmFreePool(mpool,msg);
  return icmOk;
}

long icmMsgLength(icmMsg msg)
{
  return msg->used;
}

Uchar *icmMsgData(icmMsg msg)
{
  return msg->data;
}

static icmStatus putByte(Uchar byte,icmMsg msg)
{
  icmStatus res = extendMsg(msg,1);

  if(res!=icmOk)
    return res;

  msg->data[msg->used++]=byte;
  return icmOk;
}

icmStatus icmPutBlock(Uchar *bytes,long len,icmMsg msg)
{
  icmStatus res = extendMsg(msg,len);
  
  if(res!=icmOk)
    return res;

  while(len-->0)
    msg->data[msg->used++]=*bytes++;
  return icmOk;
}

static icmStatus putIntVal(llong val,int tag,icmMsg msg)
{
  int len=0;
  Uchar bytes[32];
  unsigned long v = val;

  if(v==0)			/* special case for 0 to allow 128 bit numbers */
    bytes[len++]=0;
  else
    while(v>0){
      bytes[len++]=v&0xff;
      v >>= 8;
    }

  if(val>0 && (bytes[len-1]&0x80)) /* correct issues to do with signs */
    bytes[len++]=0;

  putByte(tag|(len&0xf),msg);

  while(len-->0)
    putByte(bytes[len],msg);

  return icmOk;
}


icmStatus icmPlaceInt(icmMsg msg,llong i)
{
  return putIntVal(i,icmInt,msg);
}

static icmStatus putFltVal(double f,icmMsg msg)
{
  Uchar buff[16], *bf = buff;
  int exp,len,sign=icmFlt;

  /* Convert floating point into buffer first -- to get the length */

  if(f<0.0){
    sign = icmNegFlt;
    f = -f;
  }

  f = frexp(f, &exp);

  for(len=0;f!=0.0;len++){
    double ip;
    f = modf(f*256,&ip);

    *bf++ = (int)ip;
  }

  putByte(sign|len,msg);	/* the lead character of the number */
  putIntVal(exp,icmInt,msg);	/* the exponent */

  return icmPutBlock(buff,len,msg);
}

icmStatus icmPlaceFlt(icmMsg msg,double f)
{
  return putFltVal(f,msg);
}

static icmStatus putSym(icmMsg msg,char *sym)
{
  int len = strlen(sym);

  putIntVal(len,icmSym,msg);
  return icmPutBlock((Uchar*)sym,len,msg);
}

icmStatus icmPlaceSym(icmMsg msg,char *sym)
{
  return putSym(msg,sym);
}

static icmStatus putHandle(icmMsg msg,icmHandle handle)
{
  char *tgt;
  char *name;
  char *home;
  char **location;
  int count;
  icmStatus ret = icmAnalyseHandle(handle,&tgt,&name,&home,&location,&count);

  if(ret!=icmOk)
    return ret;
  else{
    putByte(icmHdl,msg);

    if(tgt==NULL || strcmp(tgt,"")==0)
      putSym(msg,"");
    else
      putSym(msg,tgt);


    if(strcmp(name,"")==0)
      putSym(msg,"");
    else
      putSym(msg,name);


    if(strcmp(home,"")==0)
      putSym(msg,"");
    else
      putSym(msg,home);

    while(count--){
      putByte(icmList,msg);
      putSym(msg,*location++);
    }

    return putByte(icmNil,msg);
  }
}

icmStatus icmPlaceHandle(icmMsg msg,icmHandle handle)
{
  return putHandle(msg,handle);
}

icmStatus icmPlaceData(icmMsg msg,void *data, long size)
{
  putIntVal(size,icmData,msg);

  return icmPutBlock((Uchar*)data,size,msg);
}

icmStatus icmPlaceCode(icmMsg msg,void *code,long size)
{
  putIntVal(size,icmCode,msg);

  return icmPutBlock((Uchar*)code,size,msg);
}

icmStatus icmPlaceNil(icmMsg msg)
{
  return putByte(icmNil,msg);
}

icmStatus icmPlaceList(icmMsg msg)
{
  return putByte(icmList,msg);
}

icmStatus icmPlaceApply(icmMsg msg)
{
  return putByte(icmApply,msg);
}

icmStatus icmPlaceTuple(icmMsg msg,long arity)
{
  return putIntVal(arity,icmTuple,msg);
}

icmStatus icmPlaceOpaque(icmMsg msg)
{
  return putByte(icmOpaque,msg);
}

icmStatus icmPlaceSigned(icmMsg msg,char *sig,long sigLen)
{
  putByte(icmSigned,msg);
  return icmPlaceData(msg,sig,sigLen);
}

icmStatus icmPlaceTag(icmMsg msg,long tag)
{
  return putIntVal(tag,icmTag,msg);
}

icmStatus icmPlaceRef(icmMsg msg,long tag)
{
  return putIntVal(tag,icmRef,msg);
}

icmStatus icmPlaceShort(icmMsg msg,long tag)
{
  return putIntVal(tag,icmShort,msg);
}

icmStatus icmPlaceMsg(icmMsg msg,icmMsg sub)
{
  return icmPutBlock(sub->data,sub->used,msg);
}

icmStatus icmCopyMsg(icmMsg msg,icmMsg from,long x,long *nx)
{
  icmStatus ret=icmSkipElement(from,x,nx);

  if(ret==icmOk)
    ret = icmPutBlock(&from->data[x],*nx-x,msg);
  return ret;
}

static char *skipFmtEl(char *fmt)
{
  switch(*fmt) {
  case '%':			/* A special command */
    switch(*++fmt) {

    case 'd':
    case 'i':
    case 'f':
    case 'h':
    case 's':
    case 'S':
    case 'O':
    case '#':
    case '@':
      return fmt+1;
      
    case '"':			/* Embedded string */
      fmt++;
      while(*fmt!='\0' && strncmp(fmt,"%\"",2)!=0)
	fmt++;
      return fmt+2;
      
    case '(':			/* Skip a sub-tuple */
      fmt++;
      while(*fmt!='\0' && strncmp(fmt,"%)",2)!=0)
	fmt = skipFmtEl(fmt);
      return fmt+2;

    case '}':			/* Skip a sub-tuple */
      fmt++;
      while(*fmt!='\0' && strncmp(fmt,"%}",2)!=0)
	fmt = skipFmtEl(fmt);
      return fmt+2;

    case '[':			/* Skip a list */
      fmt++;
      while(*fmt!='\0' && strncmp(fmt,"%]",2)!=0)
	fmt = skipFmtEl(fmt);

      return fmt+2;

    case 'R':{			/* Skip an apply */
      fmt++;
      while(*fmt!='\0' && *fmt!='%')
	fmt++;
      
      return skipFmtEl(fmt);
    }

    case '*':
      return skipFmtEl(fmt+1);
    default:
      return fmt+1;
    }

  default:			/* write out a string */
    while(*fmt!='\0' && *fmt!='%')
      fmt++;
    if(strncmp(fmt,"%,",2)==0)
      return fmt+2;
    return fmt;
  }
}

static int varCount = 0;

/* Generate a type signature string from a format specification */
static char *genFSig(char *fmt,char **sig);

static char *genFmtSig(char *fmt,char **sig)
{
  switch(*fmt) {
  case '%':			/* A special command */
    return genFSig(fmt+1,sig);
  default:			/* write out a string */
    while(*fmt!='\0' && *fmt!='%')
      fmt++;
    *(*sig)++=symbol_sig;	/* We have a symbol */
    if(strncmp(fmt,"%,",2)==0)
      return fmt+2;
    return fmt;
  }
}

static char *genFSig(char *fmt,char **sig)
{
  switch(*fmt) {
  case 'd':
  case 'i':
  case 'f':
    *(*sig)++=number_sig;	/* We have a number */
    return fmt+1;
  case 'h':
    *(*sig)++=handle_sig;	/* We have a handle */
    return fmt+1;
  case 's':
    *(*sig)++=symbol_sig;	/* We have a symbol */
    return fmt+1;
  case 'S':
    *(*sig)++=string_sig;	/* We have a string value */
    return fmt+1;
  case 'x':
    *(*sig)++=opaque_sig;	/* We have an opaque value */
    return fmt+1;

  case 'O':{
    char s[] = ICM_MSG_OPTS;
    long len = NumberOf(s)-1;
    long i;

    for(i=0;i<len;i++)
      *(*sig)++=s[i];

    return fmt+1;
  }
  case '#':
  case '@':
    *(*sig)++=unknown_sig;	/* We have an unknown value */
    return fmt+1;
      
  case '"':			/* Embedded string */
    fmt++;
    while(*fmt!='\0' && strncmp(fmt,"%\"",2)!=0)
      fmt++;
    *(*sig)++=string_sig;	/* We have a string value */
    return fmt+2;
      
  case '(':{			/* Skip a sub-tuple */
    long count = 0;
    char *f = fmt+1;
    fmt++;
    while(*fmt!='\0' && strncmp(fmt,"%)",2)!=0){
      count++;
      fmt = skipFmtEl(fmt);
    }
    *(*sig)++=tuple_sig;	/* put in a tuple signature */
    *(*sig)++=count;		/* length of the tuple */

    while(count-->0)
      f = genFmtSig(f,sig);	/* put in the elements of the tuple */

    return fmt+2;
  }
  case '{':			/* A signed sub-value */
    fmt++;
    while(*fmt!='\0' && strncmp(fmt,"%}",2)!=0)
      fmt = skipFmtEl(fmt);
    *(*sig)++=unknown_sig;	/* We have an unknown value */
    return fmt+2;

  case '[':{			/* Skip a list */
    icmTruth elfound = icmNo;

    *(*sig)++=list_sig;		/* We have a list value */
    fmt++;
    while(*fmt!='\0' && strncmp(fmt,"%]",2)!=0){
      if(!elfound){		/* pick up the type of the first element */
	char *sg = *sig;
	fmt = genFmtSig(fmt,sig);
	elfound = sg!=*sig;
      }
      else
	fmt = skipFmtEl(fmt);

      if(!elfound){
	*(*sig)++=var_sig;	/* We have an unknown value */
	*(*sig)++=varCount++;
      }
    }

    return fmt+2;
  }

  case 'R':{			/* Generate an apply structure */
    char *delims = "'\"/#$\\A0"; /* Potential delimiters */
    char *p = ++fmt;

    while(*p!='\0'&&*p!='%'&&*delims!='\0'){
      if(strchr(delims,*p)!=NULL)
	delims++;		/* find a suitable delimiter */
      p++;
    }

    if(*p!='\0'&&*delims!='\0'){
      *(*sig)++=poly_sig;	/* construct a record signature */
      *(*sig)++=*delims;
      while(fmt!=p)
	*(*sig)++=*fmt++;
      *(*sig)++=*delims;
      return genFSig(fmt,sig);	/* stick in the argument signature */
    }
  }

  case '*':{
    *(*sig)++=list_sig;	/* We have a list value */
    return genFSig(fmt+1,sig);
  }
  default:
    return "";			/* not a legal format string */
  }
}


/* This generates and encoded buffer from a format control string */

#define QT_FMT "%\""
#define ETPL_FMT "%)"
#define ELST_FMT "%]"
#define ESIG_FMT "%}"

static char *format_msg(icmMsg msg, char *fmt, va_list *args)
{
  switch(*fmt) {
  case '%':{			/* A special command */
    switch(*++fmt) {

    case 'd':
      putIntVal(va_arg(*args, long),icmInt,msg); /* integer */
      return fmt+1;
      
    case 'i':
      putIntVal(va_arg(*args, llong),icmInt,msg); /* long integer */
      return fmt+1;
      
    case 'f':
      putFltVal(va_arg(*args,double),msg); /* a floating point */
      return fmt+1;

    case 's':{
      char *sym = va_arg(*args,char *);

      putSym(msg,sym);

      return fmt+1;
    }

    case 'x':{			/* An opaque value, we lead with a name */
      icmOpaquePo ptr = va_arg(*args,icmOpaquePo);
      
      icmPlaceOpaque(msg);
      icmPlaceSym(msg,ptr->name);
      icmPlaceData(msg,ptr->val->data,ptr->val->size);
      return fmt+1;
    }

    case 'S':{
      icmDataPo ptr = va_arg(*args,icmDataPo);
      long size = ptr->size;

      putIntVal(size,icmData,msg);
      icmPutBlock((Uchar*)ptr->data,size,msg);

      return fmt+1;
    }

    case 'h':{
      icmHandle hdl = va_arg(*args,icmHandle);

      if(putHandle(msg,hdl)!=icmOk)
	return NULL;
      else
	return fmt+1;
    }

    case '"':{			/* Embedded string */
      char strdata[1024];
      char *s = strdata;
      long size;

      fmt++;
      while(*fmt!='\0' && strncmp(fmt,"%\"",2)!=0)
	*s++=*fmt++;
      size = s-strdata;

      putIntVal(size,icmData,msg);
      icmPutBlock((Uchar*)strdata,size,msg);

      return fmt+2;		/* skip over the closing % */
    }
      
    case '#':{			/* we have a complete encoded message to add */
      icmMsg sub = va_arg(*args,icmMsg);

      icmPutBlock(sub->data,sub->used,msg);

      return fmt+1;
    }

    case '@':{			/* insert raw data */
      icmDataPo ptr = va_arg(*args,icmDataPo);

      icmPutBlock(ptr->data,ptr->size,msg);

      return fmt+1;
    }

    case 'O':{			/* list of message or command options */
      icmOption opts = va_arg(*args,icmOption);

      while(opts!=NULL){
	icmPlaceList(msg);
	switch(opts->opt){
	case icmLeaseTime:{
	  char sigBuff[] = {number_sig};

	  icmPlaceTuple(msg,2);
	  icmPlaceSym(msg,"icmLeaseTime");
	  icmPlaceSigned(msg,sigBuff,NumberOf(sigBuff));

	  icmPlaceInt(msg,opts->o.when);
	  break;
	}
	case icmReplyto:{
	  char sigBuff[] = {handle_sig};
	  
	  icmPlaceTuple(msg,2);
	  icmPlaceSym(msg,"icmReplyto");
	  icmPlaceSigned(msg,sigBuff,NumberOf(sigBuff));
	  icmPlaceHandle(msg,opts->o.reply);
	  break;
	}
	case icmReceiptRequest:{
	  char sigBuff[] = {number_sig};
	  
	  icmPlaceTuple(msg,2);
	  icmPlaceSym(msg,"icmReceiptRequest");
	  icmPlaceSigned(msg,sigBuff,NumberOf(sigBuff));
	  icmPlaceInt(msg,opts->o.code);
	  break;
	}
	case icmAuditTrail:{
	  char sigBuff[] = {list_sig,handle_sig};
	  icmHandle *trail = opts->o.trail;

	  icmPlaceTuple(msg,2);
	  icmPlaceSym(msg,"icmAuditTrail");
	  icmPlaceSigned(msg,sigBuff,NumberOf(sigBuff));
	  while(*trail!=NULL){
	    icmPlaceList(msg);
	    icmPlaceHandle(msg,*trail++);
	  }
	  icmPlaceNil(msg);
	  break;
	}
	  
	}
	opts = opts->next;
      }
      icmPlaceNil(msg);
    }

    case ',':			/* %, acts as a string separator */
      return fmt+1;

    case '(':{			/* A tuple to write */
      long count = 0;
      char *nfmt = fmt+1;

      while(*nfmt!='\0' && strncmp(nfmt,ETPL_FMT,2)!=0){
	count++;
	nfmt = skipFmtEl(nfmt);
      }

      putIntVal(count,icmTuple,msg); /* write the tuple header */

      fmt++;			/* Step over the lead format char */

      while(fmt!=NULL && *fmt!='\0' && strncmp(fmt,ETPL_FMT,2)!=0){
	fmt=format_msg(msg,fmt,args);
	count--;
      }

      if(count>0){
	while(count-->0)
	  putIntVal(-1,icmInt,msg);
	return NULL;		/* This is an error indication */
      }
      else if(fmt!=NULL && strncmp(fmt,ETPL_FMT,2)==0)
	return fmt+2;
      else
	return fmt;
    }

    case '[':{			/* A list to write */
      fmt++;
      while(fmt!=NULL && *fmt!='\0' && strncmp(fmt,ELST_FMT,2)!=0){
	putByte(icmList,msg);
	fmt = format_msg(msg,fmt,args);
      }

      putByte(icmNil,msg);

      if(fmt!=NULL && strncmp(fmt,ELST_FMT,2)==0)
	return fmt+2;
      else
	return fmt;
    }

    case 'R':{			/* An apply record structure */
      char fn[MAXSTRING];
      int i=0;

      fmt++;
      putByte(icmApply,msg);

      while(*fmt!='\0'&&*fmt!='%')
	fn[i++]=*fmt++;
      fn[i]='\0';

      putSym(msg,fn);
      return format_msg(msg,fmt,args);
    }

    case '{':{			/* A signed value */
      char sigBuff[MAXSTRING];
      char *sig=sigBuff;
      long sigSize;

      genFmtSig(fmt+1,&sig);

      fmt++;
      sigSize = sig-sigBuff;

      icmPlaceSigned(msg,sigBuff,sigSize);
      
      while(fmt!=NULL && *fmt!='\0' && strncmp(fmt,ESIG_FMT,2)!=0)
	fmt=format_msg(msg,fmt,args);

      if(fmt!=NULL && strncmp(fmt,ESIG_FMT,2)==0)
	return fmt+2;
      else
	return fmt;
    }

    case '*':{			/* An array of things - output as a list */
      long count = va_arg(*args,long); /* count of the number of elements */
      void **sub = va_arg(*args,void **);
      long i;

      fmt++;

      for(i=0;i<count;i++,sub++){
	putByte(icmList,msg);

	switch(*fmt){
	case 'd':
	  putIntVal((long)(*sub),icmInt,msg); /* integer */
	  break;
      
	case 'i':
	  putIntVal(*(llong*)sub,icmInt,msg); /* long integer */
	  break;
      
	case 'f':
	  putFltVal(*(double*)(*sub),msg); /* a floating point */
	  break;

	case 's':
	  putSym(msg,(char*)(*sub));
	  break;

	case 'S':{
	  icmDataPo ptr = (icmDataPo)(*sub);
	  long size = ptr->size;

	  putIntVal(size,icmData,msg);
	  icmPutBlock((Uchar*)ptr->data,size,msg);
	  break;
	}

	case 'x':{		/* We expect a pair - name and value */
	  icmOpaquePo ptr = (icmOpaquePo)(*sub);

	  icmPlaceOpaque(msg);
	  icmPlaceSym(msg,ptr->name);
	  icmPlaceData(msg,ptr->val->data,ptr->val->size);
	  break;
	}

	case 'h':
	  putHandle(msg,(icmHandle)(*sub));
	  break;
      
	case '#':{		/* we have a complete encoded message to add */
	  icmMsg subMsg = (icmMsg)(*sub);

	  icmPutBlock(subMsg->data,subMsg->used,msg);
	}

	case '@':{		/* we have raw data to add */
	  icmDataPo ptr = (icmDataPo)(*sub);

	  icmPutBlock(ptr->data,ptr->size,msg);
	}

	default:
	  outMsg(icmLogFile,"Illegal format character [%c] ignored\n",*fmt);
	  putByte(icmNil,msg);
	  return fmt+1;
	}
      }

      putByte(icmNil,msg);
      return fmt+1;
    }

    default:
      outMsg(icmLogFile,"Illegal format character [%c] ignored\n",*fmt);
      return fmt+1;
    }
  }

  default:{			/* write out a string */
    char buff[1024],*bp=buff;
    while(*fmt!='\0' && *fmt!='%')
      *bp++=*fmt++;
    *bp='\0';

    putSym(msg,buff);
    
    if(*fmt!='\0' && *(fmt+1)==',')
      return fmt+2;		/* step over comma separator */
    else
      return fmt;
  }
  }
}

icmMsg icmFormatMsg(icmStatus *ret,char *fmt,...)
{
  va_list args;
  icmMsg msg;
  char *efmt;
  icmStatus retn;
  icmDataPo ptr;

  icmNewMsg(&msg);
  
  /* Now look at the format string to determine the expected message */
  va_start(args,fmt);		/* start accessing the arguments */

  efmt = format_msg(msg,fmt,&args);

  va_end(args);			/* exit */

  if(efmt!=NULL){
    if(*efmt=='\0')
      retn = icmOk;
    else
      retn = icmFailed;
  }
  else
    retn = icmError;

#ifdef COMMTRACE
  if(retn!=icmOk){
    icmLogMsg("%d Problem in formatting message: %s",pthread_self(),fmt);
    exit(100);
  }
  else if(icmCheckMsg(msg)!=icmOk){
    icmLogMsg("%d Bad message: %s->%M",pthread_self(),fmt,msg);
    retn = icmError;
    exit(100);
  }
#endif

  if(ret!=NULL)
    *ret = retn;

  return msg;
}

icmStatus icmFmtSendMsg(icmConn conn,icmHandle recip,
			icmHandle sender,icmOption opts,char *fmt,...)
{
  va_list args;
  icmMsg msg;
  icmStatus ret;

  icmNewMsg(&msg);
  
  /* Now look at the format string to determine the expected message */
  va_start(args,fmt);		/* start accessing the arguments */

  fmt = format_msg(msg,fmt,&args);

  va_end(args);			/* exit */

  if(fmt!=NULL){
    if(*fmt=='\0')
      ret = icmOk;
    else
      ret = icmFailed;
  }
  else
    ret = icmError;

  if(ret==icmOk)
    ret = icmSendMsg(conn,recip,sender,opts,msg);

  if(msg!=NULL)
    icmReleaseMsg(msg);

  return ret;
}


#if 0
icmMsg icmPacketizeMsg(icmMsg msg,icmHandle recip,icmHandle owner)
{
  icmMsgRec from;
  long mark;
  icmMsg new=NULL;
  icmMsg p = msg;
  icmMsg n;
  long pNo = 0;
  static long msgSequenceNo = 0;
  static pthread_mutex_t atomic = PTHREAD_MUTEX_INITIALIZER;
  long msgNo;
  long off = 0;
  long size = icmMsgLength(msg);

  icmLockMutex(&atomic);
  msgNo = msgSequenceNo++;
  icmUnlockMutex(&atomic);

  from.pNo = 0;
  from.used = 0;
  from.next = NULL;

  icmPlaceTuple(&from,6);	/* We construct a header buffer */
  icmPlaceHandle(&from,recip);
  icmPlaceHandle(&from,owner);
  icmPlaceInt(&from,msgNo);
  icmPlaceInt(&from,size);

  mark = from.used;		/* the size of the fixed header part of the packet  */

  while(p!=NULL){

    from.used = mark;		/* reset header packet */
    icmPlaceInt(&from,off);	/* construct the nth packet */

    if(new==NULL){
      icmNewMsg(&new);
      n = new;
    }
    else
      n = extendMsg(new,);

    memmove(&n->data[0],&from.data[0],from.used);	/* copy packet header */
    n->used = from.used;
    off+=n->used;

    putIntVal(p->used,icmData,n,LONG_MAX);
    
    memmove(&n->data[n->used],&p->data[0],p->used); /* OK since MAXPACKET>MAXSTRING */
    n->used += p->used;

    p = p->next;
  }
  return new;
}
#endif


/* Parsing messages */
static icmStatus icmDecInt(icmMsg msg,long index,long *newx,long *result);

static int getByte(icmMsg msg,long index)
{
  if(msg==NULL){
    icmLogMsg("Tried to access illegal message");
    return EOF;
  }
  else if(index>=msg->used)
    return EOF;
  else
    return msg->data[index];
}

static icmStatus getBlock(icmMsg msg,long index,long *nx,Uchar *buffer,long len)
{
  if(msg==NULL){
    icmLogMsg("Tried to access illegal message");
    return EOF;
  }
  else{
    while(len-->0){
      if(index>=msg->used)
	return icmEndFile;
      else
	*buffer++=msg->data[index++];
    }
    if(nx!=NULL)
      *nx = index;
    return icmOk;
  }
}

static icmStatus skipBlock(icmMsg msg,long index,long len,long *nx)
{
  if(msg==NULL){
    icmLogMsg("Tried to access illegal message");
    return EOF;
  }
  else if(index+len>msg->used)
    return icmEndFile;
  else{
    if(nx!=NULL)
      *nx = index+len;
    return icmOk;
  }
}

/* Low level token functions */
icmStatus icmMsgElType(icmMsg msg,long index,icmElTag *tag)
{
  if(msg!=NULL){
    int ch = getByte(msg,index);

    if(ch==EOF)
      return icmEndFile;
    else{
      if((*tag=icm_tag(ch))==icmList){
	if(ch==icmNil)
	  *tag = icmNil;
	else
	  *tag = icmList;
      }
      return icmOk;
    }
  }
  else
    return icmError;
}

icmStatus icmMsgElSize(icmMsg msg,long index,long *size)
{
  if(msg!=NULL){
    int ch = getByte(msg,index);

    if(ch==EOF)
      return icmEndFile;
    else{
      switch(icm_tag(ch)){
      case icmInt:{
	int len = icm_val(ch);

	if(len<=(sizeof(long)/8))
	  *size = sizeof(long)/8;
	else
	  *size = (len+7)/8;	/* try to guess the true size of the integer */
	return icmOk;
      }

      case icmFlt:
      case icmNegFlt:
	*size = sizeof(double);
	return icmOk;

      case icmSym:
      case icmData:
      case icmCode:
	return icmDecInt(msg,index,&index,size);

      case icmNil:
      case icmList:
	if(ch==icmNil){
	  *size = 0;
	  return icmOk;
	}
	else if(ch==icmList){
	  long count = 0;

	  while((ch=getByte(msg,index++))!=EOF && ch==icmList){
	    icmStatus ret = icmSkipElement(msg,index,&index);

	    count++;

	    if(ret!=icmOk)
	      return ret;
	  }

	  if(ch==EOF)
	    return icmEndFile;
	  else if(ch!=icmNil)
	    return icmFailed;
	  else{
	    *size = count;
	    return icmOk;
	  }
	}
	else if(ch==icmApply){
	  *size = 0;		/* fake */
	  return icmFailed;
	}
	else{
	  *size = 0;
	  return icmOk;
	}

      case icmTuple:
	return icmDecInt(msg,index,&index,size);
	
      case icmTag:
      case icmRef:
      case icmShort:
        ;
      }
      return icmFailed;
    }
  }
  else
    return icmError;
}

icmStatus icmExpectInt(icmMsg msg,long index,long *newx,llong *result)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmInt)
    return icmFailed;
  else{
    Uchar buff[32];
    int len = icm_val(ch);
    icmStatus ret = getBlock(msg,index+1,newx,buff,len);
    long x = 0;
    int i;

    if(ret!=icmOk)
      return ret;

    for(i=0;i<len;i++)
      x = (x<<8)|buff[i];

    if(result!=NULL)
      *result = x;
    return icmOk;
  }
}

static long iLen(icmMsg msg,long index,long *newx)
{
  int ch = getByte(msg,index++);

  if(icm_val(ch)==0){		/* enter a recursive length cycle */
    long len;
    icmDecInt(msg,index,newx,&len);
    return len;
  }
  else{
    *newx = index;
    return icm_val(ch);
  }
}

static icmStatus icmDecInt(icmMsg msg,long index,long *newx,long *result)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else{
    Uchar buff[256];
    long len = iLen(msg,index,&index);
    icmStatus ret = getBlock(msg,index,newx,buff,len);
    long x = 0;
    int i;

    if(ret!=icmOk)
      return ret;

    for(i=0;i<len;i++)
      x = (x<<8)|buff[i];

    if(result!=NULL)
      *result = x;
    return icmOk;
  }
}

icmStatus icmExpectFlt(icmMsg msg,long index,long *newx,double *result)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmFlt && icm_tag(ch)!=icmNegFlt)
    return icmFailed;
  else{
    long exp;
    icmStatus ret=icmDecInt(msg,index+1,&index,&exp);
    Uchar buff[16];
    int len = icm_val(ch);
    double f = 0.0;

    if(ret!=icmOk || (ret=getBlock(msg,index,newx,buff,len))!=icmOk)
      return ret;

    while(len-->0)
      f = (f+buff[len])/256;

    f = ldexp(f,exp);

    if(icm_tag(ch)==icmNegFlt)
      f = -f;

    if(result!=NULL)
      *result = f;

    return icmOk;
  }
}

icmStatus icmExpectSym(icmMsg msg,long index,long *newx,char *buffer,long max)
{
  int ch = getByte(msg,index);
  long len;

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmSym)
    return icmFailed;
  else if(icmDecInt(msg,index,&index,&len)!=icmOk || (buffer!=NULL && max<=len))
    return icmError;
  else if(buffer!=NULL){
    icmStatus ret = getBlock(msg,index,newx,(Uchar*)buffer,len);
    buffer[len]='\0';
    return ret;
  }
  else
    return skipBlock(msg,index,len,newx);
}

icmStatus icmExpectHandle(icmMsg msg,long x,long *nx,icmHandle *handle)
{
  int ch = getByte(msg,x++);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmHdl)
    return icmFailed;
  else{
    char tgt[MAXSTRING];
    char name[MAXSTRING];
    char home[MAXSTRING];
    int locCount = 0;
    char l[10][MAXSTRING];
    char *loc[10]={l[0],l[1],l[2],l[3],l[4],l[5],l[6],l[7],l[8],l[9]};
    icmStatus ret;

    if(icmExpectSym(msg,x,&x,tgt,MAXSTRING)!=icmOk){
      if((ret=icmExpectNil(msg,x,&x))!=icmOk)
	return ret;
      else
	tgt[0]='\0';
    }

    if(icmExpectSym(msg,x,&x,name,MAXSTRING)!=icmOk){
      if((ret=icmExpectNil(msg,x,&x))!=icmOk)
	return ret;
      else
	name[0]='\0';
    }

    if(icmExpectSym(msg,x,&x,home,MAXSTRING)!=icmOk){
      if((ret=icmExpectNil(msg,x,&x))!=icmOk)
	return ret;
      else
	home[0]='\0';
    }

    while(icmExpectList(msg,x,&x)==icmOk){
      if(locCount<NumberOf(loc)){
	
	if((ret=icmExpectSym(msg,x,&x,loc[locCount++],MAXSTRING))!=icmOk)
	  return ret;
      }
      else if((ret=icmSkipElement(msg,x,&x))!=icmOk)
	return ret;
    }

    if((ret=icmExpectNil(msg,x,nx))!=icmOk)
      return ret;

    if(handle!=NULL && (*handle=icmMakeHandle(tgt,name,home,locCount,loc))==NULL)
      return icmError;
    else
      return icmOk;
  }
}

icmStatus icmExpectData(icmMsg msg,long index,long *newx,void *buffer,long *len,long max)
{
  int ch = getByte(msg,index);
  long length;

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmData)
    return icmFailed;
  else if(icmDecInt(msg,index,&index,&length)!=icmOk || (buffer!=NULL && max<length))
    return icmError;
  else{
    if(len!=NULL)
      *len = length;

    if(buffer!=NULL)
      return getBlock(msg,index,newx,(Uchar*)buffer,length);
    else
      return skipBlock(msg,index,length,newx);
  }
}

icmStatus icmExpectCode(icmMsg msg,long index,long *newx,void *buffer,long *len,long max)
{
  int ch = getByte(msg,index);
  long length;

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmCode)
    return icmFailed;
  else if(icmDecInt(msg,index,&index,&length)!=icmOk || (buffer!=NULL && max<length))
    return icmError;
  else{
    if(len!=NULL)
      *len = length;

    if(buffer!=NULL)
      return getBlock(msg,index,newx,(Uchar*)buffer,length);
    else
      return skipBlock(msg,index,length,newx);
  }
}

icmStatus icmExpectNil(icmMsg msg,long index,long *newx)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(ch!=icmNil)
    return icmFailed;
  else{
    if(newx!=NULL)
      *newx=index+1;
    return icmOk;
  }
}

icmStatus icmExpectList(icmMsg msg,long index,long *newx)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(ch!=icmList)
    return icmFailed;
  else{
    if(newx!=NULL)
      *newx=index+1;
    return icmOk;
  }
}

icmStatus icmExpectApply(icmMsg msg,long index,long *newx)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(ch!=icmApply)
    return icmFailed;
  else{
    if(newx!=NULL)
      *newx=index+1;
    return icmOk;
  }
}

icmStatus icmExpectTuple(icmMsg msg,long index,long *newx,long *arity)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmTuple)
    return icmFailed;
  else
    return icmDecInt(msg,index,newx,arity);
}


icmStatus icmExpectSigned(icmMsg msg,long index,long *newx,char *sig,long *max)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmSigned)
    return icmFailed;
  else if(icmExpectData(msg,index+1,newx,sig,max,*max)!=icmOk)
    return icmFailed;
  else
    return icmOk;
}


icmStatus icmExpectOpaque(icmMsg msg,long index,long *newx)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(ch!=icmOpaque)
    return icmFailed;
  else{
    if(newx!=NULL)
      *newx=index+1;
    return icmOk;
  }
}

icmStatus icmExpectTag(icmMsg msg,long index,long *newx,long *tag)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmTag)
    return icmFailed;
  else
    return icmDecInt(msg,index,newx,tag);
}

icmStatus icmExpectRef(icmMsg msg,long index,long *newx,long *tag)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmRef)
    return icmFailed;
  else
    return icmDecInt(msg,index,newx,tag);
}

icmStatus icmExpectShort(icmMsg msg,long index,long *newx,long *tag)
{
  int ch = getByte(msg,index);

  if(ch==EOF)
    return icmEndFile;
  else if(icm_tag(ch)!=icmShort)
    return icmFailed;
  else
    return icmDecInt(msg,index,newx,tag);
}

icmStatus icmExpectAny(icmMsg msg,long x,long *nx,icmMsg *m)
{
  long xx;
  icmStatus ret;

  if((ret=icmSkipElement(msg,x,&xx))!=icmOk || m==NULL){
    if(nx!=NULL)
      *nx=xx;
    return ret;
  }
  else{
    icmStatus ret=icmNewMsg(m);
    long ix=x;
    long len=xx-x;
    icmMsg sub=*m;

    while(len--)
      putByte(getByte(msg,ix++),sub);

    if(nx!=NULL)
      *nx=xx;
    return ret;
  }
}

icmStatus icmSkipElement(icmMsg msg,long index,long *newx)
{
  register int skip = 1;	/* number of terms to skip */

  while(skip-->0){
    int ch=getByte(msg,index);

    if(ch==EOF)
      return icmEndFile;

    switch(icm_tag(ch)){
    case icmInt:
      icmDecInt(msg,index,&index,NULL);
      break;

    case icmFlt:
    case icmNegFlt:{
      icmDecInt(msg,index+1,&index,NULL);

      index+=icm_val(ch);
      break;
    }

    case icmHdl:
      index++;
      skip+=4;			/* There are 4 components to a handle */
      break;
	
    case icmSym:
    case icmData:
    case icmCode:{
      long size=0;

      icmDecInt(msg,index,&index,&size);
      index+=size;
      break;
    }

    case icmNil:
    case icmList:
    case icmApply:
      index++;
      if(ch==icmList||ch==icmApply)
	skip+=2;
      break;

    case icmTuple:{
      long arity=0;

      icmDecInt(msg,index,&index,&arity);

      skip+=arity;
      break;
    }

    case icmOpaque:
      index++;
      skip+=2;
      break;

    case icmTag:
      icmDecInt(msg,index,&index,NULL);

      skip++;
      break;

    case icmRef:
      icmDecInt(msg,index,&index,NULL);
      break;

    case icmShort:
      icmDecInt(msg,index,&index,NULL);
      skip+=2;			/* we skip 2 elements for a short */
      break;

    case icmSigned:
      skip+=2;			/* skip two more elements */
      index++;
      break;
    }

    /* Test for the end of the message */
    if(skip>0 && getByte(msg,index)==EOF)
      return icmEndFile;

  }
  if(newx!=NULL)
    *newx=index;
  return icmOk;
}

icmStatus icmCheckMsg(icmMsg msg)
{
  long len;

  if(icmSkipElement(msg,0,&len)!=icmOk)
    return icmError;
  else if(len!=msg->used)
    return icmError;
  else
    return icmOk;
}

static icmStatus scan_msg(icmMsg msg,long x,long *nx,char **fmt,va_list *args)
{
  char c=*(*fmt)++;
  icmStatus ret = icmOk;

  switch(c){
  case '%':{
    switch((c=*(*fmt)++)){

    case 'd':{
      long *k = va_arg(*args,long *); /* pick up a integer pointer */
      llong kk;
      icmStatus ret = icmExpectInt(msg,x,nx,&kk);

      if(ret==icmOk && k!=NULL)
	*k = (long)kk;		/* cast result down to 32 bits */
      return ret;
    }

    case 'i':{
      llong *k = va_arg(*args,llong *); /* pick up a integer pointer */

      return icmExpectInt(msg,x,nx,k);
    }

    case 'f':{
      double *g = va_arg(*args, double*);

      return icmExpectFlt(msg,x,nx,g);
    }

    case 's':{			/* expect a symbol */
      char *s = va_arg(*args, char *); /* pick up a char * pointer */

      return icmExpectSym(msg,x,nx,s,MAXSTRING);
    }

    case 'h':{
      icmHandle *handle = va_arg(*args,icmHandle*);

      return icmExpectHandle(msg,x,nx,handle);
    }

    case 'S':{
      icmDataPo p = va_arg(*args,icmDataPo);
      long size;

      icmMsgElSize(msg,x,&size);

      if(p!=NULL){
	p->data=(Uchar*)malloc(sizeof(Uchar)*size);
	p->size=size;
	return icmExpectData(msg,x,nx,p->data,&size,size);
      }
      else
	return icmExpectData(msg,x,nx,NULL,NULL,size);
    }

    case 'x':{			/* extract an opaque value */
      if((ret=icmExpectOpaque(msg,x,&x))==icmOk){
	icmOpaquePo p = va_arg(*args,icmOpaquePo);
	long nsize,vsize;
	char *buff;

	icmMsgElSize(msg,x,&nsize);

	buff = (char*)malloc(nsize+1);
	if((ret=icmExpectSym(msg,x,&x,buff,nsize))==icmOk){
	  p->name = buff;;

	  icmMsgElSize(msg,x,&vsize);

	  p->val->data=(Uchar*)malloc(sizeof(Uchar)*vsize);
	  p->val->size=vsize;
	  return icmExpectData(msg,x,nx,p->val->data,&p->val->size,
			       p->val->size);
	}
      }
      return ret;
    }

    case 'C':{
      icmDataPo p = va_arg(*args,icmDataPo);
      long size;

      icmMsgElSize(msg,x,&size);

      if(p!=NULL){
	p->data=(Uchar*)malloc(sizeof(Uchar)*size);
	p->size=size;

	return icmExpectCode(msg,x,nx,p->data,&size,size);
      }
      else
	return icmExpectCode(msg,x,nx,NULL,NULL,size);
    }

    case 'O':{
      icmOption *opts = va_arg(*args,icmOption*);
      long ar;

      if(opts!=NULL){
	*opts = NULL;

	while(icmExpectList(msg,x,&x)==icmOk){
	  char sig[MAXSTRING];
	  long sigLen = NumberOf(sig);

	  if(icmExpectTuple(msg,x,&x,&ar)==icmOk && ar==2){
	    char key[MAXSTRING];

	    if(icmExpectSym(msg,x,&x,key,NumberOf(key))==icmOk){
	      llong when,code;
	      icmHandle reply;

	      if(strcmp(key,"icmLeaseTime")==0 && 
		 icmExpectSigned(msg,x,&x,sig,&sigLen)==icmOk && 
		 icmExpectInt(msg,x,&x,&when)==icmOk)
		*opts = icmNewOpt(*opts,icmLeaseTime,when);
	      else if(strcmp(key,"icmReplyto")==0 &&
		      icmExpectSigned(msg,x,&x,sig,&sigLen)==icmOk && 
		      icmExpectHandle(msg,x,&x,&reply)==icmOk)
		*opts = icmNewOpt(*opts,icmReplyto,reply);
	      else if(strcmp(key,"icmReceiptRequest")==0 &&
		      icmExpectSigned(msg,x,&x,sig,&sigLen)==icmOk && 
		      icmExpectInt(msg,x,&x,&code)==0)
		*opts = icmNewOpt(*opts,icmReceiptRequest,code);
	      else if(strcmp(key,"icmAuditTrail")==0 &&
		 icmExpectSigned(msg,x,&x,sig,&sigLen)==icmOk){
		icmHandle hdls[256];
		int t = 0;

		while(icmExpectList(msg,x,&x)==icmOk){
		  if(icmExpectHandle(msg,x,&x,&hdls[t++])!=icmOk)
		    return icmFailed;
		}

		if(icmExpectNil(msg,x,&x)==icmOk){
		  icmHandle *trail = (icmHandle*)malloc(sizeof(icmHandle)*(t+1));
		  int i;

		  for(i=0;i<t;i++)
		    trail[i]=hdls[i];
		  trail[i]=NULL;
		  *opts = icmNewOpt(*opts,icmAuditTrail,trail);
		}
		else
		  return icmFailed;
	      }
	      else
		return icmFailed;
	    }
	    else
	      return icmFailed;
	  }
	  else
	    return icmFailed;
	}

	return icmExpectNil(msg,x,nx);
      }
      else
	return icmSkipElement(msg,x,nx);
    }

    case '@':{		/* we simply return the index to the term */
      long *mark = va_arg(*args,long *);

      if(mark!=NULL)
	*mark = x;

      return icmSkipElement(msg,x,nx);
    }

    case '#':{			/* one `term' as an encoded block */
      icmMsg *m = va_arg(*args,icmMsg*);

      if(m!=NULL)
	return icmExpectAny(msg,x,nx,m);
      else
	return icmSkipElement(msg,x,nx);
    }

    case '"':{			/* Literal string in the format spec */
      Uchar data[MAXSTRING];
      long size;
      char strdata[MAXSTRING];
      char *s = strdata;
      fmt++;

      while(fmt!=NULL && *fmt!=NULL && **fmt!='\0' && strncmp(*fmt,"%\"",2)!=0)
	*s++=*(*fmt)++;

      ret = icmExpectData(msg,x,nx,data,&size,MAXSTRING);

      if(ret!=icmOk)
	return ret;

      (*fmt)+=2;		/* move over the closing quote */

      if(memcmp(strdata,data,size)==0)
	return icmOk;
      else
	return icmFailed;
    }

    case '(':{			/* A tuple of values ... */
      long arity;

      if((ret=icmExpectTuple(msg,x,&x,&arity))!=icmOk)
	return ret;
      
      while(arity-- && *fmt!=NULL && **fmt!='\0')
	if((ret=scan_msg(msg,x,&x,fmt,args))!=icmOk)
	  return ret;

      if(*fmt==NULL || strncmp(*fmt,ETPL_FMT,2)!=0)
	return icmFailed;
      else{
	if(nx!=NULL)
	  *nx = x;
	(*fmt)+=2;
	return icmOk;
      }
    }


    case '[':{			/* A list of objects ... */

      while(*fmt!=NULL && **fmt!='\0' && strncmp(*fmt,ELST_FMT,2)!=0){
	if((ret=icmExpectList(msg,x,&x))!=icmOk ||
	   (ret=scan_msg(msg,x,&x,fmt,args))!=icmOk)
	  return ret;
      }

      if(strncmp(*fmt,ELST_FMT,2)!=0)
	return icmError;
      else{
	(*fmt)+=2;		/* skip over the closing list brackets */
	return icmExpectNil(msg,x,nx);
      }
    }

    case 'R':{			/* A record structure ... */
      if((ret=icmExpectApply(msg,x,&x))==icmOk){
	char buff[MAXSTRING];
	char *bf = buff;
	char copy[MAXSTRING];
    
	while(**fmt && (**fmt!='%' || *((*fmt)+1)=='%')){
	  if(**fmt=='%' && *((*fmt)+1)=='%')
	    (*fmt)++;
	  *bf++ = *(*fmt)++;
	}
	
	*bf='\0';		/* terminate the symbol */

	if((ret=icmExpectSym(msg,x,&x,copy,MAXSTRING))!=icmOk)
	  return ret;
	else if(strcmp(buff,copy)!=0)
	  return icmFailed;
	else
	  return scan_msg(msg,x,nx,fmt,args);
      }
      else
	return ret;
    }

    case '{':{			/* A signed value */
      char sig[MAXSTRING];
      long sigLen=NumberOf(sig);

      if(icmExpectSigned(msg,x,&x,sig,&sigLen)!=icmOk)
	return icmFailed;
      else{
	if((ret=scan_msg(msg,x,nx,fmt,args))!=icmOk)
	  return ret;
	else if(strcmp(*fmt,ESIG_FMT)!=0)
	  return icmError;
	else{
	  (*fmt)+=2;		/* skip over the closing brace */
	  return icmOk;
	}
      }
    }

    case ',':
      return scan_msg(msg,x,nx,fmt,args);

    case '*':{			/* A repeated format -- list of items */
      switch(*(*fmt)++){
      case '*':{		/* we only count the number of entries */
	long *count = va_arg(*args, long *);
	icmMsgElSize(msg,x,count);
	return icmSkipElement(msg,x,nx);
      }

      case 'd':{
	long *count = va_arg(*args,long *);
	long *k = va_arg(*args,long *); /* pick up a integer array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  llong ii;

	  if((ret=icmExpectInt(msg,x,&x,&ii))!=icmOk)
	    return ret;
	  else if(i<*count){
	    *k++=ii;
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case 'i':{
	long *count = va_arg(*args,long *);
	llong *k = va_arg(*args,llong *); /* pick up a integer array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  llong ii;

	  if((ret=icmExpectInt(msg,x,&x,&ii))!=icmOk)
	    return ret;
	  else if(i<*count){
	    *k++=ii;
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case 'f':{
	long *count = va_arg(*args,long *);
	double *k = va_arg(*args,double *); /* pick up a float array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  double ff;

	  if((ret=icmExpectFlt(msg,x,&x,&ff))!=icmOk)
	    return ret;
	  else if(i<*count){
	    *k++=ff;
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case 's':{		/* expect a symbol */
	long *count = va_arg(*args,long *);
	char **syms = va_arg(*args,char **); /* pick up a string array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  char buff[MAXSTRING];

	  if((ret=icmExpectSym(msg,x,&x,buff,MAXSTRING))!=icmOk)
	    return ret;
	  else if(i<*count){
	    *syms++=icmDupStr(buff);
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case 'h':{
	long *count = va_arg(*args,long *);
	icmHandle *hdls = va_arg(*args,icmHandle *); /* pick up a handle array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  icmHandle h;

	  if((ret=icmExpectHandle(msg,x,&x,&h))!=icmOk)
	    return ret;
	  else if(i<*count){
	    *hdls++=h;
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case 'S':{
	long *count = va_arg(*args,long *);
	icmDataPo *data = va_arg(*args,icmDataPo *); /* pick up a data array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  icmDataPo p=*data;
	  long size;

	  icmMsgElSize(msg,x,&size);
	  
	  p->data=(Uchar*)malloc(sizeof(Uchar)*size);
	  p->size=size;

	  if((ret=icmExpectData(msg,x,&x,p->data,&size,p->size))!=icmOk)
	    return ret;
	  else if(i<*count){
	    data++;
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case 'C':{
	long *count = va_arg(*args,long *);
	icmDataPo *data = va_arg(*args,icmDataPo *); /* pick up a data array pointer */
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  icmDataPo p=*data;
	  long size;

	  icmMsgElSize(msg,x,&size);
	  
	  p->data=(Uchar*)malloc(sizeof(Uchar)*size);
	  p->size=size;

	  if((ret=icmExpectCode(msg,x,&x,p->data,&size,size))!=icmOk)
	    return ret;
	  else if(i<*count){
	    data++;
	    i++;
	  }
	  else
	    return icmError;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case '@':{		/* we simply return indices to the list elements */
	long *count = va_arg(*args,long *);
	long *marks = va_arg(*args,long *);
	long i=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if(i<*count){
	    *marks++=x;		/* upload the index to the message element */
	    i++;
	  }

	  if((ret=icmSkipElement(msg,x,&x))!=icmOk)
	    return ret;
	}

	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = i;
	  return icmOk;
	}
	else
	  return ret;
      }

      case '#':{		/* list of `terms' as an encoded block */
	long *count = va_arg(*args,long *);
	icmMsg *m = va_arg(*args,icmMsg*);
	icmMsg *mm=m;
	long cnt=0;
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  icmMsg sub;
	  long xx;

	  if((ret=icmNewMsg(&sub))!=icmOk || (ret=icmSkipElement(msg,x,&xx))!=icmOk)
	    return ret;
	  else if(cnt<*count){
	    long ix=x;
	    long len=xx-x;

	    while(len--)
	      putByte(getByte(msg,ix++),sub);

	    cnt++;
	    *m++=sub;
	    x = xx;
	  }
	  else{
	    while(cnt--)
	      icmReleaseMsg(mm[cnt]);
	    
	    return icmError;
	  }
	}
	if((ret=icmExpectNil(msg,x,nx))==icmOk){
	  *count = cnt;
	  return icmOk;
	}
	else
	  return ret;
      }
      default:
	return icmError;
      }
    }
    }
  }

  default:{
    char buff[2048];
    char *bf = buff;
    char copy[MAXSTRING];
    
    *bf++=c;		/* copy the expected string into the buffer */
    while(**fmt && (**fmt!='%' || *((*fmt)+1)=='%')){
      if(**fmt=='%' && *((*fmt)+1)=='%')
	(*fmt)++;
      *bf++ = *(*fmt)++;
    }
	
    *bf='\0';		/* terminate the symbol */

    if((ret=icmExpectSym(msg,x,nx,copy,MAXSTRING))!=icmOk)
      return ret;
    else if(strcmp(buff,copy)!=0)
      return icmFailed;
    else
      return icmOk;
  }
  }
}


/* we test a message to see if it has the right shape */
static icmStatus test_msg(icmMsg msg,long x,long *nx,char **fmt)
{
  char c=*(*fmt)++;
  icmStatus ret = icmOk;

#ifdef COMMTRACE_
  if(traceMsg){
    long X;
    outMsg(icmLogFile,"%d: test msg x=%d`",pthread_self(),x);
    displayMsg(icmLogFile,msg,10,icmYes,x,&X);
    outMsg(icmLogFile,"'\n");
    flushFile(icmLogFile);
  }
#endif

  switch(c){
  case '%':{
    switch((c=*(*fmt)++)){

    case 'd':
      return icmExpectInt(msg,x,nx,NULL);

    case 'i':
      return icmExpectInt(msg,x,nx,NULL);

    case 'f':
      return icmExpectFlt(msg,x,nx,NULL);

    case 's':			/* expect a symbol */
      return icmExpectSym(msg,x,nx,NULL,0);

    case 'h':
      return icmExpectHandle(msg,x,nx,NULL);

    case 'S':
      return icmExpectData(msg,x,nx,NULL,NULL,0);

    case 'C':
      return icmExpectCode(msg,x,nx,NULL,NULL,0);

    case 'x':
      if((ret=icmExpectOpaque(msg,x,&x))==icmOk &&
	 (ret=icmExpectSym(msg,x,&x,NULL,0))==icmOk)
	return icmExpectData(msg,x,nx,NULL,NULL,0);
      else
	return ret;

    case 'O':
    case '#':			/* one `term' as an encoded block */
    case '@':
      return icmSkipElement(msg,x,nx);

    case '"':{			/* Literal string in the format spec */
      Uchar data[MAXSTRING];
      long size;
      char strdata[MAXSTRING];
      char *s = strdata;
      fmt++;

      while(**fmt!='\0' && strncmp(*fmt,"%\"",2)!=0)
	*s++=*(*fmt)++;

      ret = icmExpectData(msg,x,nx,data,&size,MAXSTRING);

      if(ret!=icmOk)
	return ret;

      (*fmt)+=2;		/* move over the closing quote */

      if(memcmp(strdata,data,size)==0)
	return icmOk;
      else
	return icmFailed;
    }

    case '(':{			/* A tuple of values ... */
      long arity;

      if((ret=icmExpectTuple(msg,x,&x,&arity))!=icmOk)
	return ret;
      
      while(arity-- && *fmt!=NULL && **fmt!='\0')
	if((ret=test_msg(msg,x,&x,fmt))!=icmOk)
	  return ret;

      if(*fmt==NULL || strncmp(*fmt,ETPL_FMT,2)!=0)
	return icmFailed;
      else{
	*nx = x;
	(*fmt)+=2;
	return icmOk;
      }
    }

    case '{':{			/* A signed value */
      char sig[MAXSTRING];
      long sigLen=NumberOf(sig);

      if(icmExpectSigned(msg,x,&x,sig,&sigLen)!=icmOk)
	return icmFailed;
      else{
	if((ret=test_msg(msg,x,nx,fmt))!=icmOk)
	  return ret;
	else if(strcmp(*fmt,ESIG_FMT)!=0)
	  return icmError;
	else{
	  (*fmt)+=2;		/* skip over the closing brace */
	  return icmOk;
	}
      }
    }

    case '[':{			/* A list of objects ... */

      while(*fmt!=NULL && **fmt!='\0' && strncmp(*fmt,ELST_FMT,2)!=0){
	if((ret=icmExpectList(msg,x,&x))!=icmOk ||
	   (ret=test_msg(msg,x,&x,fmt))!=icmOk)
	  return ret;
      }

      if(*fmt==NULL || strncmp(*fmt,ELST_FMT,2)!=0)
	return icmError;
      else{
	(*fmt)+=2;		/* skip over the closing list brackets */
	return icmExpectNil(msg,x,nx);
      }
    }

    case ',':
      return test_msg(msg,x,nx,fmt);

    case 'R':{
      if((ret=icmExpectApply(msg,x,&x))==icmOk){
	char buff[MAXSTRING];
	char *bf = buff;
	char copy[MAXSTRING];
    
	while(**fmt && (**fmt!='%' || *((*fmt)+1)=='%')){
	  if(**fmt=='%' && *((*fmt)+1)=='%')
	    (*fmt)++;
	  *bf++ = *(*fmt)++;
	}
	
	*bf='\0';		/* terminate the symbol */

	if((ret=icmExpectSym(msg,x,&x,copy,MAXSTRING))!=icmOk)
	  return ret;
	else if(strcmp(buff,copy)!=0)
	  return icmFailed;
	else
	  return test_msg(msg,x,nx,fmt);
      }
      else
	return ret;
    }

    case '*':{			/* A repeated format -- list of items */
      switch(*(*fmt)++){
      case '*':			/* we only count the number of entries */
	return icmSkipElement(msg,x,nx);

      case 'd':{
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmExpectInt(msg,x,&x,NULL))!=icmOk)
	    return ret;
	}

	return icmExpectNil(msg,x,nx);
      }

      case 'f':{
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmExpectFlt(msg,x,&x,NULL))!=icmOk)
	    return ret;
	}

	return icmExpectNil(msg,x,nx);
      }

      case 's':{		/* expect a symbol */
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmExpectSym(msg,x,&x,NULL,0))!=icmOk)
	    return ret;
	}

	return icmExpectNil(msg,x,nx);
      }

      case 'h':{
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmExpectHandle(msg,x,&x,NULL))!=icmOk)
	    return ret;
	}

	return icmExpectNil(msg,x,nx);
      }

      case 'S':{
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmExpectData(msg,x,&x,NULL,NULL,0))!=icmOk)
	    return ret;
	}

	return icmExpectNil(msg,x,nx);
      }

      case 'C':{
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmExpectCode(msg,x,&x,NULL,NULL,0))!=icmOk)
	    return ret;
	}

	return icmExpectNil(msg,x,nx);
      }

      case '#':{		/* list of `terms' as an encoded block */
	icmStatus ret;

	while(icmExpectList(msg,x,&x)==icmOk){
	  if((ret=icmSkipElement(msg,x,&x))!=icmOk)
	    return ret;
	}
	return icmExpectNil(msg,x,nx);
      }
      default:
	return icmError;
      }
    }
    }
  }

  default:{
    char buff[2048];
    char *bf = buff;
    char copy[MAXSTRING];
    
    *bf++=c;		/* copy the expected string into the buffer */
    while(**fmt && (**fmt!='%' || *((*fmt)+1)=='%')){
      if(**fmt=='%' && *((*fmt)+1)=='%')
	(*fmt)++;
      *bf++ = *(*fmt)++;
    }
	
    *bf='\0';		/* terminate the symbol */

    if((ret=icmExpectSym(msg,x,nx,copy,MAXSTRING))!=icmOk)
      return ret;
    else if(strcmp(buff,copy)!=0)
      return icmFailed;
    else
      return icmOk;
  }
  }
}

icmStatus icmScanMsg(icmMsg msg, char *fmt, ...)
{
  long end;
  char *tfmt = fmt;

  icmStatus ret = test_msg(msg,0,&end,&tfmt);

  if(ret!=icmOk)
    return ret;
  else{
    va_list args;

    /* Now look at the format string to determine the expected message */
    va_start(args,fmt);		/* start accessing the arguments */

    ret=scan_msg(msg,0,&end,&fmt,&args);
    va_end(args);			/* exit */

    return ret;
  }
}

icmStatus icmXScanMsg(icmMsg msg,long start,long *exit,char *fmt, ...)
{
  long end;
  char *tfmt = fmt;
  icmStatus ret = test_msg(msg,start,&end,&tfmt);

  if(ret!=icmOk)
    return ret;
  else{
    va_list args;

    /* Now look at the format string to determine the expected message */
    va_start(args,fmt);		/* start accessing the arguments */

    ret=scan_msg(msg,start,exit,&fmt,&args);
    va_end(args);			/* exit */

    return ret;
  }
}

/**********************************************************************/
/*                     Display a message                              */
/**********************************************************************/

static icmStatus displayMsg(filePo f,icmMsg msg,int depth,icmTruth alt,long x,long *nx)
{
  int ch = getByte(msg,x);

  if(ch==EOF)
    return icmEndFile;

  switch(icm_tag(ch)){
  case icmInt:{
    llong i;
    icmStatus ret = icmExpectInt(msg,x,nx,&i);

    if(ret!=icmOk)
      return ret;
    else
      return outInt(f,i);
  }

  case icmFlt:
  case icmNegFlt:{
    double g;

    icmStatus ret = icmExpectFlt(msg,x,nx,&g);

    if(ret!=icmOk)
      return ret;
    else
      return outFloat(f,g);
  }

  case icmHdl:{
    int ch = getByte(msg,x++);

    if(ch==EOF)
      return icmEndFile;
    else if(icm_tag(ch)!=icmHdl)
      return icmFailed;
    else{
      char tgt[MAXSTRING];
      char name[MAXSTRING];
      char home[MAXSTRING];
      char *sep = "";
      icmStatus ret;

      /* We are doing this the hard way to avoid interacting with addresses */
      if(icmExpectSym(msg,x,&x,tgt,MAXSTRING)!=icmOk){
	if((ret=icmExpectNil(msg,x,&x))!=icmOk)
	  return ret;
      }
      else if(strlen(tgt)>0)
	outMsg(f,"%s:",tgt);

      if(icmExpectSym(msg,x,&x,name,MAXSTRING)!=icmOk){
	if((ret=icmExpectNil(msg,x,&x))!=icmOk)
	  return ret;
      }
      else
	outMsg(f,"%s",name);

      if(icmExpectSym(msg,x,&x,home,MAXSTRING)!=icmOk){
	if((ret=icmExpectNil(msg,x,&x))!=icmOk)
	  return ret;
      }
      else
	outMsg(f,"@%s",home);

      if(!alt){
	outMsg(f,"/[");

	while(icmExpectList(msg,x,&x)==icmOk){
	  char location[MAXSTRING];
	  
	  if((ret=icmExpectSym(msg,x,&x,location,MAXSTRING))!=icmOk)
	    return ret;
	  else{
	    outMsg(f,"%s%s",sep,location);
	    sep = ",";
	  }
	}
	if((ret=icmExpectNil(msg,x,nx))!=icmOk)
	  return ret;
	else
	  outStr(f,"]");
	
	return icmOk;
      }
      else
	return icmSkipElement(msg,x,nx);
    }
  }
	
  case icmSym:{
    char sym[MAXSTRING];
    icmStatus ret = icmExpectSym(msg,x,nx,sym,MAXSTRING);

    if(ret!=icmOk)
      return ret;
    else
      return outStr(f,sym);
  }

  case icmData:{
    long len;
    icmStatus ret = icmMsgElSize(msg,x,&len);

    if(ret!=icmOk)
      return ret;
    else{
      //alloca doesn't exist in windows
	  //Uchar *buffer = (Uchar*)alloca(sizeof(Uchar)*len);
	  Uchar *buffer = malloc(sizeof(Uchar)*len);

      icmExpectData(msg,x,nx,buffer,&len,len);

      outCh(f,'"');
      outBlock(f,buffer,len);
      outCh(f,'"');

	  //alloca doesn't exist in windows
	  free(buffer);

      return icmOk;
    }
  }

  case icmCode:{
    long len;
    icmMsgElSize(msg,x,&len);

    outMsg(f,"<<code%d>>",len);
    return icmSkipElement(msg,x,nx);
  }

  case icmOpaque:		/* An opaque value */
    outStr(f,"/*");
    displayMsg(f,msg,depth-1,alt,x+1,&x);
    outStr(f,"*/=");
    return displayMsg(f,msg,depth-1,alt,x,nx);

  case icmNil:
  case icmList:	
  case icmApply:		/* these map to the same point */
    if(ch==icmNil){
      *nx = x+1;
      return outStr(f,"[]");
    }
    else if(ch==icmList){
      char ch = '[';
      while(icmExpectList(msg,x,&x)==icmOk){
	icmStatus ret;

	outCh(f,ch);
	ch = ',';
	if((ret=displayMsg(f,msg,depth-1,alt,x,&x))!=icmOk)
	  return ret;
      }

      if(icmExpectNil(msg,x,nx)!=icmOk){
	icmStatus ret;

	outStr(f,",..");
	if((ret=displayMsg(f,msg,depth-1,alt,x,nx))!=icmOk)
	  return ret;
      }
      return outCh(f,']');
    }
    else{
      icmStatus ret=displayMsg(f,msg,depth-1,alt,x+1,&x);

      if(ret!=icmOk)
	return ret;
      if(icmExpectTuple(msg,x,NULL,NULL)!=icmOk)
	return displayMsg(f,msg,depth-1,alt,x,nx);
      else{
	outCh(f,'(');
	displayMsg(f,msg,depth-1,alt,x,nx);
	return outCh(f,')');
      }
    }
      
  case icmTuple:{
    long arity=0;
    char *sep = "";
    
    icmDecInt(msg,x,&x,&arity);
    outCh(f,'(');

    while(arity-->0){
      icmStatus ret;

      outStr(f,sep); sep = ",";
      
      if((ret=displayMsg(f,msg,depth-1,alt,x,&x))!=icmOk)
	return ret;
    }

    *nx = x;
    return outCh(f,')');
  }

  case icmSigned:{
    long len;
    icmStatus ret = icmMsgElSize(msg,x+1,&len);

    if(ret!=icmOk)
      return ret;
    else{
      //Uchar *buffer = (Uchar*)alloca(sizeof(Uchar)*len);
	  Uchar *buffer = malloc(sizeof(Uchar)*len);

      icmExpectData(msg,x+1,&x,buffer,&len,len);

      outSig(f,buffer);
      outStr(f,"=");

	  free(buffer);

      if((ret=displayMsg(f,msg,depth-1,alt,x,nx))!=icmOk)
	  {
		return ret;
	  } else {
		return icmOk;
	  }
    }
  }
  
  case icmTag:{
    long tag;
    icmDecInt(msg,x,&x,&tag);

    outMsg(f,BOLD_ESC_ON "L%d=" BOLD_ESC_OFF,tag);
    return displayMsg(f,msg,depth,alt,x,nx);
  }

  case icmRef:{
    long tag;
    icmDecInt(msg,x,nx,&tag);

    return outMsg(f,BOLD_ESC_ON "R%d" BOLD_ESC_OFF,tag);
  }

  case icmShort:{
    long tag;
    icmStatus ret;

    icmDecInt(msg,x,&x,&tag);

    outMsg(f,BOLD_ESC_ON "S%d=" BOLD_ESC_OFF,tag);

    if((ret=displayMsg(f,msg,depth-1,alt,x,&x))!=icmOk)
      return ret;

    return displayMsg(f,msg,depth,alt,x,nx);
  }
  default:
  return icmOk;
  }
}

/* Display a type signature properly */
static char *outSig(filePo f,char *sig)
{
  if(sig!=NULL){
    switch(*sig++){
    case number_sig:
      outStr(f,"number");
      return sig;
    case symbol_sig:
      outStr(f,"symbol");
      return sig;
    case string_sig:
      outStr(f,"string");
      return sig;
    case handle_sig: 
      outStr(f,"handle");
      return sig;
    case logical_sig: 
      outStr(f,"logical");
      return sig;
    case opaque_sig: 
      outStr(f,"opaque");
      return sig;
    case unknown_sig: 
      outStr(f,"any");
      return sig;
    case var_sig:{		/* type variable */
      outMsg(f,"%c%d",'%',*sig);
      return sig+1;
    }
    case forall_sig:{		/* Quantifier variable */
      outMsg(f,"%c%d:",'%',*sig);
      return outSig(f,sig+1);
    }
    case list_sig:		/* List closure type signature*/
      sig = outSig(f,sig);
      outStr(f,"[]");
      return sig;

    case tuple_sig:{
      int count = *sig++;	/* number of elements in the tuple */

      outCh(f,'(');
      while(sig!=NULL && count-->0){
	sig = outSig(f,sig);
	if(count>0)
	  outCh(f,',');
      }
      outCh(f,')');
      return sig;
    }

    case usr_sig:{
      char name[MAXSTRING], *n=name;
      char delim = *sig++;	/* pick up the delimiter character */

      /* look for the delimiter */
      while(*sig!=delim)
	*n++=*sig++;

      *n = '\0';

      outStr(f,name);
      return sig+1;
    }

    case poly_sig:{
      char name[MAXSTRING], *n=name;
      char delim = *sig++;	/* pick up the delimiter character */

      /* look for the delimiter */
      while(*sig!=delim)
	*n++=*sig++;

      sig++;
      *n = '\0';

      outStr(f,name);

      if(*sig==tuple_sig)
	return outSig(f,sig);
      else{
	outCh(f,'(');
	sig = outSig(f,sig);
	outCh(f,')');
	return sig;
      }
    }

    case query_sig:{
      char name[MAXSTRING], *n=name;
      char delim = *sig++;	/* pick up the delimiter character */

      /* look for the delimiter */
      while(*sig!=delim)
	*n++=*sig++;

      *n = '\0';

      sig = outSig(f,sig+1);

      outCh(f,'?');
      outStr(f,name);

      return sig;
    }

    case funct_sig:		/* function type signature */
      sig = outSig(f,sig);
      outStr(f,"=>");
      return outSig(f,sig);
      
    case proc_sig:		/* procedure type signature */
      sig = outSig(f,sig);
      outStr(f,"{}");
      return sig;

    default:
      outStr(f,"invalid type signature");
      return NULL;		/* We have an invalid type signature */
    }
  }
  else
    return NULL;
}



void disp_msg(icmMsg msg)
{
  long end;
  displayMsg(icmLogFile,msg,1000,icmYes,0,&end);
  flushFile(icmLogFile);
}

void d_msg(icmMsg msg,long from)
{
  displayMsg(icmLogFile,msg,1000,icmYes,from,&from);
  flushFile(icmLogFile);
}

static icmStatus dispMsg(filePo f,void *msg,int width,int precision,icmTruth alt)
{
  long end;
  return displayMsg(f,msg,width,alt,0,&end);
}

/*
 * Read a single message from a file, copying it into a message structure 
 */
static icmStatus readInt(filePo in,long *i,char tag,icmMsg msg);
static icmStatus readBlock(filePo in,long len,icmMsg msg);

/* Parse a coded message structure from a file into a message structure */
icmStatus icmReadMsg(filePo in,icmMsg msg)
{
  int ch = inCh(in);		/* Pick out the leading character */
  icmStatus res;

  if(ch==EOF)
    return icmEndFile;

  switch(icm_tag(ch)){
  case icmInt:
    return readInt(in,NULL,ch,msg);

  case icmFlt:
  case icmNegFlt:{
    unsigned short len = icm_val(ch);

    putByte(ch,msg);		/* Put the floating tag byte itself into the message */

    if((res=readInt(in,NULL,inCh(in),msg))!=icmOk || (res=readBlock(in,len,msg))!=icmOk)
      return res;

    return icmOk;
  }

  case icmSym:{
    long len;

    if((res=readInt(in,&len,ch,msg))!=icmOk)
      return res;

    return readBlock(in,len,msg);
  }

  case icmData:{
    long len;

    if((res=readInt(in,&len,ch,msg))!=icmOk)
      return res;

    return readBlock(in,len,msg);
  }

  case icmCode:{
    long len;

    if((res=readInt(in,&len,ch,msg))!=icmOk)
      return res;

    return readBlock(in,len,msg);
  }

  case icmHdl:{
    int i;

    putByte(ch,msg);		/* put the handle byte across */

    for(i=0;i<4;i++){
      if((res=icmReadMsg(in,msg))!=icmOk) /* consume an element of the handle */
	return res;
    }
    return icmOk;
  }

  case icmNil:
  case icmList:
  case icmApply:
    if(ch==icmNil)
      return putByte(ch,msg);
    else{
      putByte(ch,msg);

      if((res=icmReadMsg(in,msg))!=icmOk) /* consume the head of the list/apply */
	return res;

      return icmReadMsg(in,msg); /* consume the tail of the list/apply */
    }

  case icmTuple:{
    long arity=0;
    long i;

    if((res=readInt(in,&arity,ch,msg))!=icmOk)
     return res;

    for(i=0;i<arity;i++)
      if((res=icmReadMsg(in,msg))!=icmOk) /* consume a tuple element */
	return res;

    return icmOk;
  }

  case icmOpaque:		/* opaque values are similar to signed ones */
  case icmSigned:{		/* read a signed value */
    putByte(ch,msg);

    if((res=icmReadMsg(in,msg))!=icmOk)
      return res;
    else
      return icmReadMsg(in,msg); /* read in the value itself */
  }

  case icmTag:{
    long tag;

    if((res=readInt(in,&tag,ch,msg))!=icmOk)
      return res;

    return icmReadMsg(in,msg);
  }

  case icmRef:
    return readInt(in,NULL,ch,msg);

  case icmShort:{
    long tag;

    if((res=readInt(in,&tag,ch,msg))!=icmOk)
      return res;

    if((res=icmReadMsg(in,msg))!=icmOk)
      return res;

    return icmReadMsg(in,msg);
  }

  default:

#ifdef COMMTRACE
    icmLogMsg("Funny tag %c(0x%x) from %s",ch,ch,fileName(in));
#endif

    return icmError;
  }
}

static icmStatus readInt(filePo in,long *i,char tag,icmMsg msg)
{
  unsigned short len = icm_val(tag)==0?16:icm_val(tag);
  Uchar buff[16];		/* We get the block of data into here */
  icmStatus res = inBlock(in,buff,len);

  if(res!=icmOk)
    return res;

  putByte(tag,msg);		/* copy the tag byte across */
  icmPutBlock(buff,len,msg);	/* copy the block of data into the message */

  if(i!=NULL){
    long result = 0;
    int j;
    for(j=0;j<len;j++)
      result = (result<<8)|buff[j];
    *i = result;
  }
  return icmOk;
}

static icmStatus readBlock(filePo in,long len,icmMsg msg)
{
  char buff[512],*buffer=buff;
  icmStatus res;

  if(len>512)
    buffer = (char*)malloc(sizeof(char)*len);

  res=inBlock(in,buffer,len);	/* read in a block of bytes */

  icmPutBlock(buffer,len,msg);	/* copy the block into the message */

  if(len>512)
    free(buffer);		/* release the block */

  return res;
}

icmStatus icmWaitMsg(icmConn conn,icmHandle *to,icmHandle *sndr,
		     icmOption *opts,char *fmt,...)
{
  long seqNo = -LONG_MAX;
  icmMsg msg;
  icmStatus ret;

  while((ret=icmGetMsg(conn,to,sndr,opts,&msg,&seqNo))==icmOk){
    char *tfmt = fmt;
    long end;

    ret = test_msg(msg,0,&end,&tfmt);

    if(ret==icmOk){
      va_list args;

      /* Now look at the format string to determine the expected message */
      va_start(args,fmt);		/* start accessing the arguments */

      ret=scan_msg(msg,0,&end,&fmt,&args);
      va_end(args);			/* exit */
    }

    if(ret==icmOk){
      icmReleaseMsg(msg);
      return icmOk;
    }
    else
      icmReplaceMsg(conn,*to,*sndr,*opts,msg,seqNo);
  }
  return icmError;		/* Something went wrong here */
}

/*
 * options management
 */

icmOption icmNewOpt(icmOption next,icmOpt o,...)
{
  va_list args;
  icmOption opt = (icmOption)icmAllocPool(opool);

  va_start(args,o);		/* start accessing the arguments */

  opt->next = next;
  switch(opt->opt=o){
  case icmLeaseTime:
    opt->o.when = va_arg(args,time_t);
    return opt;
  case icmReplyto:
    opt->o.reply = va_arg(args,icmHandle);
    return opt;
  case icmReceiptRequest:
    opt->o.code = va_arg(args,llong);
    return opt;
  case icmAuditTrail:
    opt->o.trail = va_arg(args,icmHandle*);
    return opt;
  default:
    icmLogMsg("invalid option type");
    icmFreePool(opool,opt);
    return next;
  }

  return opt;
}

icmStatus icmReleaseOptions(icmOption opts)
{
  while(opts!=NULL){
    icmOption next = opts->next;
    if(opts->opt==icmAuditTrail && opts->o.trail!=NULL)
      free(opts->o.trail);

    icmFreePool(opool,opts);
    opts = next;
  }
  return icmOk;
}

icmStatus icmIsOption(icmOption opts,icmOpt o,...)
{
  va_list args;
  
  va_start(args,o);

  while(opts!=NULL){
    if(opts->opt==o){
      switch(o){
      case icmLeaseTime:{
	time_t *when = va_arg(args,time_t*);
	*when = opts->o.when;
	break;
      }

      case icmReplyto:{
	icmHandle *reply = va_arg(args,icmHandle*);
	*reply = opts->o.reply;
	break;
      }

      case icmReceiptRequest:{
	llong *key = va_arg(args,llong*);
	*key = opts->o.when;
	break;
      }

      case icmAuditTrail:{
	icmHandle **trail = va_arg(args,icmHandle**);
	*trail = opts->o.trail;
	break;
      }

      default:
	return icmError;
      }

      va_end(args);
      return icmOk;
    }
    else
      opts = opts->next;
  }

  va_end(args);
  return icmFailed;
}

static icmStatus dispOpts(filePo f,void *msg,int width,int precision,icmTruth alt)
{
  icmOption opts = (icmOption)msg;

  outStr(f,"{");
  while(opts!=NULL){
    switch(opts->opt){
    case icmLeaseTime:{
      time_t when = opts->o.when;
      outMsg(f,"lease=%d",when);
      break;
    }

    case icmReplyto:{
      outMsg(f,"replyto=%h",opts->o.reply);
      break;
    }

    case icmReceiptRequest:{
      outMsg(f,"receipt");
      break;
    }

    case icmAuditTrail:{
      icmHandle *trail = opts->o.trail;
      outMsg(f,"trail=[");
      while(*trail!=NULL){
	outMsg(f,"%h",*trail++);
	if(*trail!=NULL)
	  outStr(f,",");
      }
      outStr(f,"]");
      break;
    }

    default:
      return icmError;
    }
    opts = opts->next;
    if(opts!=NULL)
      outStr(f," ");
  }
  outStr(f,"}");
  return icmOk;
}


