/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * Handle incoming messages from the commserver
 * Part of the ICM message handling sub-system
 */

#include "config.h"		/* Invoke configuration header */
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include "icmP.h"		/* Private message handling functions */
#include "icmIo.h"		/* I/O interface */
#include "icmIoP.h"		/* I/O interface */
#include "icmSock.h"		/* ICM socket library interface */

static icmPoolBase mpool = NULL;
static icmPoolBase ppool = NULL;
static pthread_once_t poolInit = PTHREAD_ONCE_INIT;

static icmMsg aggregate(icmConn conn,icmHandle recip,long MsgNo,long count,long pNo,
			Uchar *block,long len);

/*
 *  Attempt to read a message from the comms server
 *  Returns Ok if there was a message
 *  Failed if there is no pending message and Err if the comms has closed down
 *  Automatically consumes messages which are associated with callbacks
 */

static void init_inmsg(void)
{
  mpool = icmInitPool(sizeof(icmMsgListRec),16);
  ppool = icmInitPool(sizeof(MsgPacket),16);
}

static icmStatus icmGtMsg(icmConn conn,icmMsg *msg,long *seqNo)
{
  long seq = (seqNo!=NULL?*seqNo:-LONG_MAX);

  icmLockMutex(&conn->mutex);

  {
    icmMsgListPo *p = &conn->pending;

    while((*p)!=NULL){
#ifdef COMMTRACE
      if(traceMsg)
	icmLogMsg("%d: check for outstanding messages",pthread_self());
#endif

      if((*p)->outstanding==0 && (*p)->seqNo>seq){ /* message waiting for us */
	icmMsgListPo pp = *p;

	*p = (*p)->next;
	*msg = pp->msg;		/* pick out the message */
	if(seqNo!=NULL)
	  *seqNo = pp->seqNo;

	icmFreePool(mpool,pp);
	icmUnlockMutex(&conn->mutex);
	return icmOk;		/* return the message */
      }
      else
	p = &(*p)->next;
    }
  }

  icmUnlockMutex(&conn->mutex);

  icmNewMsg(msg);		/* We grab ourselves a new message */

#ifdef COMMTRACE
  if(traceMsg)
    icmLogMsg("%d: try to read message from %s",pthread_self(),
	      fileName(conn->conn));
#endif

  /* Read the next message into the incoming buffer */
  switch(icmReadMsg(conn->conn,*msg)){
  case icmOk:
    if(seqNo!=NULL)
      *seqNo = conn->seqNo++;
    else
      conn->seqNo++;		/* increment anyway */
    return icmOk;
  case icmFailed:
    return icmFailed;		/* Should never happen */
  case icmEndFile:
    icmReleaseMsg(*msg);
    return icmEndFile;
  default:
#ifdef COMMTRACE
    icmLogMsg("%d: error in reading read msg from %s",pthread_self(),
	      fileName(conn->conn));
#endif
    icmReleaseMsg(*msg);
    return icmError;
  }
}

icmStatus icmGetMsg(icmConn conn,icmHandle *recip,icmHandle *sndr,
		    icmOption *opts,icmMsg *m,long *seqNo)
{
  icmStatus ret;
  icmMsg msg;

  while((ret=icmGtMsg(conn,&msg,seqNo))==icmOk){
#ifdef COMMTRACE
    if(traceMsg)
      icmLogMsg("%d: incoming message %M",pthread_self(),msg);
#endif

    /* We have at least one complete message in the buffer ... */
    if(icmScanMsg(msg,"%(%h%h%O%#%)",recip,sndr,opts,m)==icmOk){
      if(_CallCallBack_(conn,*recip,*sndr,*opts,*m)==icmOk){
	icmReleaseMsg(msg);
	icmReleaseMsg(*m);
	icmReleaseHandle(*recip);
	icmReleaseHandle(*sndr);
	icmReleaseOptions(*opts);
	continue;
      }
      else{
	icmReleaseMsg(msg);	/* release the outer level */
	return icmOk;
      }
    }
    else{
#ifdef COMMTRACE
      if(traceMsg)
	icmLogMsg("Msg %#M not in valid form",msg);
#endif
      return icmError;
    }
  }
  return ret;
}

icmStatus icmReplaceMsg(icmConn conn,icmHandle recip,icmHandle sndr,
			icmOption opts,icmMsg m,long seqNo)
{
  pthread_once(&poolInit,init_inmsg); /* initialize the pool just once */

  {
    icmMsg replace = icmFormatMsg(NULL,"%(%h%h%O%#%)",recip,sndr,opts,m);
    icmMsgListPo pp = (icmMsgListPo)icmAllocPool(mpool);

    pp->msg = replace;
    pp->outstanding = 0;
    pp->packets = NULL;
    pp->count = 1;
    pp->sndr = sndr;
    pp->seqNo = seqNo;

    if(icmIsOption(opts,icmLeaseTime,&pp->lease)!=icmOk)
      pp->lease = 0;

    icmReleaseHandle(recip);	/* We automatically release these handles */
    icmReleaseHandle(sndr);
    icmReleaseOptions(opts);

    icmLockMutex(&conn->mutex);

    {
      icmMsgListPo *p = &conn->pending;

      while((*p)!=NULL){
	if((*p)->seqNo>=seqNo){	/* Put the message in front of this one */
	  pp->next = *p;
	  *p = pp;		/* link in the new message into the queue */
	  icmUnlockMutex(&conn->mutex);
	  return icmOk;
	}
	else
	  p = &(*p)->next;
      }

      pp->next = NULL;
      *p = pp;
      icmUnlockMutex(&conn->mutex);
      return icmOk;
    }
  }
}

icmStatus icmMsgAvail(icmConn conn,long from)
{
  icmStatus stat = icmFailed;

  icmLockMutex(&conn->mutex);

  {
    icmMsgListPo *p = &conn->pending;

    while((*p)!=NULL){
      if((*p)->outstanding==0 && (*p)->seqNo>from){ /* message waiting for us */
	stat = icmOk;
	break;
      }
      else
	p = &(*p)->next;
    }
  }

  if(stat==icmFailed)
    stat = inReady(conn->conn);

  icmUnlockMutex(&conn->mutex);

  return stat;
}


static void linkPacket(icmMsgListPo pending,Uchar *block,long len,long pNo)
{
  packetPo pkt = pending->packets;
  packetPo *fp = &pending->packets;
  packetPo n = (packetPo)icmAllocPool(ppool);

  n->data = block;
  n->dlen = len;
  n->pNo = pNo;

  while(pkt!=NULL && pkt->pNo<pNo){
    fp = &pkt->next;
    pkt = pkt->next;
  }

  n->next = pkt;
  *fp = n;			/* link the packet into the list of packets */
  pending->outstanding--;
}

static icmMsg consolidateMsg(icmMsgListPo pending)
{
  icmMsg msg;
  packetPo pkt = pending->packets;

  icmNewMsg(&msg);

  while(pkt!=NULL){
    packetPo n = pkt->next;

    icmPutBlock(pkt->data,pkt->dlen,msg);
    free(pkt->data);	/* release the packet data */
    icmFreePool(ppool,pkt);
    pkt = n;
  }

  return msg;
}


