/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * Generic element pool handling
 * Implements pools of objects that can be allocated and freed
 */
#include "config.h"		/* Invoke configuration header */
#include <stdlib.h>		/* access malloc etc. */
#include <assert.h>

#include "icmP.h"		/* Access the private version of our interface */
#include "icmM.h"
#if DMALLOC
#include <dmalloc.h>		/* debugging malloc library */
#endif

typedef struct _pool *poolpo;

typedef struct _poolbase_{
  size_t elsize;		/* size of each element */
  poolpo free;			/* list of free elements */
  int incsize;			/* how many to add if we need more space */
  pthread_mutex_t mutex;	/* Mutex controls access to the pool table */
  long used;			/* How many have been allocated */
  long fcount;			/* How many in the free list */
  long alloced;			/* How many have been allocated */
} pbase;

typedef struct _item_{
  Ulong sign;			/* Pool signature */
} Pitem,*itemPo;

typedef struct _pool{
  Ulong sign;			/* Pool signature */
  poolpo next;			/* next element in the pool */
} ppl;

icmPoolBase icmInitPool(size_t elsize, int initial)
{
  char *alloc = calloc(initial,elsize=elsize+sizeof(Pitem));
  int i = 0;
  char *p = alloc;
  icmPoolBase base = (icmPoolBase)malloc(sizeof(pbase));

  if(base==NULL || alloc==NULL)
    return NULL;

  base->free=NULL;
  base->elsize=elsize;
  base->incsize=initial;
  base->used = 0;
  base->fcount = 0;
  base->alloced = 0;

  icmCreateMutex(&base->mutex);

  for(i=0;i<initial;i++){
    poolpo el = (poolpo)p;

    el->next = base->free;
    el->sign = (Ulong)base;
    base->free = el;
    base->fcount++;
    base->alloced++;
    p+=elsize;
  }

  assert(base->fcount+base->used==base->alloced);

  return base;
}
  
/* Allocate an element from the pool */  
void *icmAllocPool(icmPoolBase base)
{
  itemPo el = NULL;

  icmLockMutex(&base->mutex);

  assert(base->fcount+base->used==base->alloced);

  if(base->free==NULL){
    char *alloc = calloc(base->incsize,base->elsize);

    if(alloc!=NULL){
      int i = base->incsize; 

      while(i--){
	poolpo al = (poolpo)alloc;

	al->next = base->free;
	al->sign = (Ulong)base;
	base->free = al;
	base->fcount++;
	base->alloced++;

	alloc += base->elsize;
      }
    }
  }

  if(base->free!=NULL){
    poolpo p = base->free;

#ifdef TRACEPOOL
    assert(p->sign==(Ulong)base);
#endif

    base->free = p->next;
    base->used++;
    base->fcount--;
    p->sign = ~(Ulong)base;	/* set the sigature for a used entry*/
    el = (itemPo)p;
  }

  assert(base->fcount+base->used==base->alloced);
  icmUnlockMutex(&base->mutex);
  return (void*)(el+1);
}

void icmFreePool(icmPoolBase base, void *p)
{
  poolpo el = (poolpo)(((itemPo) p)-1); /* step back to the header structure */

  icmLockMutex(&base->mutex);

  assert(base->fcount+base->used==base->alloced && el->sign == ~(Ulong)base);

  el->next = base->free;
  el->sign = (Ulong)base;	/* set in the marker */
  base->free = el;
  base->used--;
  base->fcount++;

  assert(base->fcount+base->used==base->alloced);

  icmUnlockMutex(&base->mutex);
}

void icmVerifyPool(icmPoolBase base)
{
  poolpo el=base->free;

  outMsg(icmLogFile,"Verifying pool 0x%x %d bytes/element (%d used)\n",
	 base,base->elsize,base->used);

  while(el!=NULL){
    if(el->sign!=(Ulong)base)
      logMsg(icmLogFile,"Free list el. 0x%x in pool 0x%x corrupt",el,base);
    el = el->next;
  }
}

void icmverifyPoolElement(icmPoolBase base,void *p)
{
  poolpo el = ((poolpo)p)-1;	/* step back to the header structure */

  if(el->sign!=~(Ulong)base){
    icmLogMsg(BOLD_ESC_ON"Invalid entry in pool"BOLD_ESC_OFF);
    exit(1);
  }
}

