/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * A process abstraction as a special layer on top of the basic thread model
 * A process has a queue of messages as well as being a thread
 */
#include "config.h"		/* Invoke configuration header */
#include <signal.h>
#include "icmP.h"
#include "icmPP.h"

static icmPoolBase ppool = NULL;
static icmPoolBase mpool = NULL;

#ifdef PROCTRACE
icmTruth traceProc = icmNo;
#endif

typedef struct _msg_rec_ *icmMsgPo;
typedef struct _msg_rec_ {
  void *msg;
  icmConsumeMsg mp;
  icmProcessPo sender;
  long seqNo;
  icmMsgPo next;
} MsgRec;

typedef struct _process_rec_ {
  pthread_mutex_t mutex;
  pthread_cond_t avail;

  pthread_t process;
  icmProcProc proc;		/* not used v. much */
  void *cl;
  icmMsgPo first;		/* first msg in queue */
  icmMsgPo last;		/* last msg in queue */
  long seqNo;			/* current message number */

} ProcessRec;

static pthread_once_t procInit = PTHREAD_ONCE_INIT;

static void initProcesses(void)
{
  ppool = icmInitPool(sizeof(ProcessRec),16);
  mpool = icmInitPool(sizeof(MsgRec),4);
}

static void *procWorker(void *arg);

icmProcessPo icmCreateProcess(icmProcProc p,void *cl)
{
  pthread_once(&procInit,initProcesses);

  {
    icmProcessPo pp = (icmProcessPo)icmAllocPool(ppool);
    pthread_attr_t attr;
    
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

    if(pp!=NULL){
      pp->first = pp->last = NULL;
      pp->seqNo = 0;
      pp->proc = p;
      pp->cl = cl;

      icmCreateMutex(&pp->mutex);
      pthread_cond_init(&pp->avail,NULL);

      icmLockMutex(&pp->mutex);

      if(pthread_create(&pp->process,&attr,procWorker,pp)!=0){
	logMsg(icmLogFile,"Unable to create process thread");
	icmDestroyMutex(&pp->mutex);
	pthread_cond_destroy(&pp->avail);
	icmFreePool(ppool,pp);
	return NULL;
      }	
      pthread_attr_destroy(&attr);
      icmUnlockMutex(&pp->mutex);
    }

    return pp;
  }
}

static void *procWorker(void *arg)
{
  icmProcessPo p = (icmProcessPo)arg;
  
  p->proc(p,p->cl);
  return NULL;
}

icmStatus icmSendPMsg(icmProcessPo p,icmProcessPo sender,icmConsumeMsg mp,void *cmsg)
{
  icmMsgPo msg = (icmMsgPo)icmAllocPool(mpool);

  if(msg==NULL)
    return icmError;
  else{
    msg->sender = sender;
    msg->msg = cmsg;
    msg->mp = mp;
    msg->next = NULL;
    msg->seqNo = p->seqNo++;

    icmLockMutex(&p->mutex);

    if(p->first==NULL)
      p->first=p->last=msg;
    else{
      p->last->next=msg;
      p->last=msg;
    }
    pthread_cond_signal(&p->avail);
    icmUnlockMutex(&p->mutex);
    return icmOk;
  }
}

icmStatus icmReplacePMsg(icmProcessPo p,icmProcessPo sender,icmConsumeMsg mp,void *cmsg,
			 long seq)
{
  icmMsgPo msg = (icmMsgPo)icmAllocPool(mpool);

  if(msg==NULL)
    return icmError;
  else{
    msg->sender = sender;
    msg->msg = cmsg;
    msg->mp = mp;
    msg->next = NULL;
    msg->seqNo = seq;

    icmLockMutex(&p->mutex);

    if(p->first==NULL)
      p->first=p->last=msg;
    else{
      icmMsgPo *m = &p->first;

      while((*m)!=NULL && (*m)->seqNo<=seq)
	m=&(*m)->next;

      msg->next = *m;

      if((*m)==NULL){
	p->last->next=msg;
	p->last=msg;
      }
      else
	*m = msg;
    }
      
    pthread_cond_signal(&p->avail);
    icmUnlockMutex(&p->mutex);
    return icmOk;
  }
}

icmStatus icmNextPMsg(icmProcessPo p,icmProcessPo *sender,icmConsumeMsg *mp,void **msg,
		     long *seq)
{
  icmLockMutex(&p->mutex);

  while(p->first==NULL)
    if(pthread_cond_wait(&p->avail,&p->mutex)!=0){
      icmUnlockMutex(&p->mutex);
      return icmError;
    }

  if(p->first!=NULL){
    icmMsgPo *m = &p->first;

    while(*m!=NULL && (*m)->seqNo<=*seq)
      m = &(*m)->next;

    if(*m!=NULL){
      icmMsgPo mm = *m;

      *sender = mm->sender;
      *msg = mm->msg;
      *mp = mm->mp;
      *seq = mm->seqNo;

      if(mm == p->last)
	p->first=p->last=NULL;
      else
	*m=mm->next;

      icmFreePool(mpool,mm);
      icmUnlockMutex(&p->mutex);
      return icmOk;
    }
  }
  icmUnlockMutex(&p->mutex);
  return icmFailed;
}






