/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * Socket interface library for ICM
 *
 */
#include "config.h"		/* Invoke configuration header */
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <sys/types.h>
#ifdef WIN32
#include <signal.h>
#include <fcntl.h>
#include <winsock.h>
#include <windows.h>
#include <io.h>
#else
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h> 
#include <netinet/tcp.h> 
#include <unistd.h>
#endif

#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef SYSV
#include <stropts.h>
#endif
#include <errno.h>

#include "icmP.h"		/* TCP-level access functions */
#include "icmIoP.h"
#include "icmSock.h"		/* Socket library interface */

#ifndef SOCKET_ERROR
#define SOCKET_ERROR -1
#endif

#ifndef INVALID_SOCKET
#define INVALID_SOCKET -1
#endif

static void _initSockets(void) /* Sometimes we must initialize sockets */
{
}

static icmTruth initSockets(void)
{
  static pthread_once_t sockInit = PTHREAD_ONCE_INIT;
  int wsockErr;
  WSADATA wsaData;
 
  // Init windows sockets
  WORD wVersionRequested = MAKEWORD( 2, 2 );
  wsockErr = WSAStartup(wVersionRequested, &wsaData );

  pthread_once(&sockInit,_initSockets); /* initialize the sockets just once */
  return icmYes;		/* This is a dummy */
}

/* Socket reading and writing functions */
static icmStatus _refillSocket(filePo in,int sock)
{
again:
  if(in->in_pos==in->in_len){
    int nBytes = recv(sock,(char*)in->in_line,MAXLINE,0);
#ifndef WIN32
	int locErr = errno;
#endif

#ifdef COMMTRACE_
    if(traceMsg)
      icmLogMsg("%d: %d bytes read (errno=%d) from socket %s",
                pthread_self(),nBytes,locErr,fileName(in));
#endif

    if(nBytes>0){
      in->in_pos = 0;
      in->in_len = nBytes;
      return in->status = icmOk;
    }
    else if(nBytes==0){		/* End of file */
      in->in_pos=in->in_len=0;
      return in->status=icmEndFile;
    }
    else{
#ifdef WIN32
	  switch(WSAGetLastError()){
#else
      switch(locErr){
#endif
      case WSAEINTR:		/* Interrupted */
	goto again;

      case 0:			 /*this shouldnt happen ... but it does */
      case WSAEWOULDBLOCK:		/* Would have blocked */
	in->in_pos=in->in_len = 0;
	in->status = icmOk;
	return icmFailed;
      default:
	in->in_pos=in->in_len=0;
	in->status=icmEndFile;
	return icmError;	/* Something unspecific */
      }
    }
  }
  else
    return icmOk;
}

static icmStatus _flushSocket(filePo out,int sock)
{
  if(out->out_pos>0){
    long nBytes = send(sock,out->out_line,out->out_pos,0);

    if(nBytes!=SOCKET_ERROR){
      long left = out->out_pos-nBytes;
      long i;

      for(i=0;i<left;i++)	/* maybe not everything was written */
	out->out_line[i]=out->out_line[i+nBytes];
      out->out_pos = left;

      if(left==0)		/* Anything left to write? */
	return icmOk;
      else
	return icmFailed;		/* We couldnt write everything */
    }
    else{
      switch(errno){
      case WSAEWOULDBLOCK:
      case WSAEINTR:
	return icmFailed;
      default:
	return icmError;
      }
    }
  }
  else
    return icmOk;			/* Nothing to write ... */
}

static icmStatus _closeSocket(filePo sk,int sock)
{
  if(close(sock)==-1)
    return icmError;
  else
    return icmOk;
}

static icmStatus _configureSocket(filePo f, icmIoConfigOpt mode,int sock)
{
  switch(mode){
  case turnOffBlocking:{
    int SokOpt = icmYes;
#ifdef WIN32
	int fd_flags = ioctlsocket(sock,FIONBIO,0);
#else
	int fd_flags=fcntl(sock,F_GETFL,0);
#endif
    if(fd_flags==-1)
      return icmError;

#ifndef WIN32
    fd_flags |= O_NDELAY;
    if(fcntl(sock,F_SETFL,fd_flags)==-1)
      return icmError;
#endif
    /* Set the socket to non-delay mode to avoid buffering problems */
    setsockopt(sock,IPPROTO_TCP,TCP_NODELAY,(char*)&SokOpt,sizeof(SokOpt));
    return icmOk;
  }

  case turnOnBlocking:{
#ifdef WIN32
    if  (ioctlsocket(sock,FIONBIO,0))
	{
		return icmOk;
	} else {
		return icmError;
	}
#else
	int fd_flags=fcntl(sock,F_GETFL,0);

    if(fd_flags==-1)
      return icmError;

    fd_flags &= ~O_NDELAY;
    if(fcntl(sock,F_SETFL,fd_flags)==-1)
      return icmError;
    return icmOk;
#endif
  }

  case enableAsynch:{		/* Enable interrupt driven I/O */
#ifdef SYSV
    if(ioctl(sock,I_SETSIG,S_INPUT|S_OUTPUT)>=0)
      return icmOk;
    else
      return icmError;
#else
#ifdef WIN32
    if(ioctlsocket(sock,FIONBIO,1)<0)
#else
	int fd_flags=fcntl(sock,F_GETFL,0);

    if(fd_flags==-1)
      return icmError;

    if(fcntl(sock,F_SETOWN,getpid())<0)
      icmLogMsg("Warning: couldnt set ownership of socket %d",sock);
	if(fcntl(sock,F_SETFL,fd_flags|O_NONBLOCK)<0)
#endif
      return icmError;
    else
      return icmOk;
#endif
  }

  case disableAsynch:{		/* We no longer want to receive interrupts */

#ifndef WIN32
#ifdef SYSV
    if(ioctl(sock,I_SETSIG,0)>=0)
      return icmOk;
    else
      return icmError;
#else
    int fd_flags=fcntl(sock,F_GETFL,0);

    if(fd_flags==-1)
      return icmError;

    if(fcntl(sock,F_SETFL,fd_flags&~O_NONBLOCK)<0)
      return icmError;
    else
      return icmOk;
#endif
#endif

#ifdef WIN32
	if(ioctlsocket(sock,FIONBIO,0)<0)
      return icmError;
    else
      return icmOk;
#endif

  }

  default:
    return icmError;			/* Unknown option */
  }
}

filePo listeningPort(char *name,int port)
{
  icmTruth inited = initSockets();
  int sock = socket(AF_INET,SOCK_STREAM,0);

  if(!inited || sock==INVALID_SOCKET)
    return NULL;
  else{
    /* Set the socket to reuse addresses */
    int one=1, len=sizeof(int);

    if(setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char*)&one,len)!=0){
      close(sock);
      return NULL;
    }
    else{
      struct sockaddr_in addr;	/* Internet-domain protocol */

      memset((char *)&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = htonl(INADDR_ANY);
      addr.sin_port = htons((unsigned short)port);
  
      if(bind(sock,(struct sockaddr*)&addr,sizeof(addr))!=0){
	close(sock);
	return NULL;
      }

      listen(sock,5);

      return _new_io(name,ioFile,ioREAD,sock,
		     _refillSocket,0,_closeSocket,
		     _fileInReady,_fileOutReady,_configureSocket);
    }
  }
}

/* acceptConnection allows a connection from a connect socket and returns a filePo
   which can read/write to the remote host
 */
filePo acceptConnection(filePo listen,icmStatus *res)
{
  icmTruth inited = initSockets();
  struct sockaddr_in cli_addr;
  int clilen = sizeof(cli_addr);
  int lSock = (int)fileData(listen);
  int cliSock;

  memset((char *)&cli_addr, 0, sizeof(cli_addr));

  cliSock = accept(lSock,(struct sockaddr*)&cli_addr, &clilen);

  if(!inited || cliSock<0){
#ifdef WIN32
	switch(WSAGetLastError()){
#else
    switch(errno){
#endif
    case WSAEWOULDBLOCK:
      *res = icmFailed;
      return NULL;
    case WSAEINTR:			/* Interrupted */
      *res = icmEndFile;
      return NULL;
    default:
      icmLogMsg("Problem in accept: %s",strerror(errno));
      *res=icmError;
      return NULL;
    }
  }
  else{
    struct hostent *client=gethostbyaddr((char*)&cli_addr.sin_addr.s_addr,4,AF_INET);
    char cl_name[MAXSTRING];
    
    if(client!=NULL)
      icmLowerStr(cl_name,client->h_name,MAXSTRING);
    else{
      struct in_addr addr = *(struct in_addr*)&cli_addr;
      icmIPToStr(&addr,cl_name,NumberOf(cl_name));
    }

    *res=icmOk;

    return _new_io(cl_name,ioFile,ioREAD|ioWRITE,cliSock,
		   _refillSocket,_flushSocket,_closeSocket,
		   _fileInReady,_fileOutReady,_configureSocket);
  }
}

/* return the peername of a connection */
char *peerName(filePo stream,int *port)
{
  icmTruth inited = initSockets();
  struct sockaddr_in cli_addr;
  int cli_len = sizeof(cli_addr);
  int sock = (int)fileData(stream);
  int stat = getpeername(sock,(struct sockaddr*)&cli_addr,&cli_len);

  if(inited && stat==0){
    if(port!=NULL)
      *port = ntohs(cli_addr.sin_port);

    return fileName(stream);
  }
  else
    return NULL;
}

/* return the peername IP of a connection */
char *peerIP(filePo stream,int *port)
{
  icmTruth inited = initSockets();
  struct sockaddr_in cli_addr;
  int cli_len = sizeof(cli_addr);
  int sock = (int)fileData(stream);
  int stat = getpeername(sock,(struct sockaddr*)&cli_addr,&cli_len);

  if(inited && stat==0){
    if(port!=NULL)
      *port = ntohs(cli_addr.sin_port);

    return inet_ntoa(cli_addr.sin_addr);
  }
  else{
    icmLogMsg("Failure to get IP address of peer, errno=%d, problem=%s",errno,
	      strerror(errno));
    return NULL;
  }
}


icmIpPort icmPeerAddr(filePo stream)
{
  initSockets();

  {
    struct sockaddr_in cli_addr;
    int cli_len = sizeof(cli_addr);
    int sock = (int)fileData(stream);
    icmIpPort addr;

    getpeername(sock,(struct sockaddr*)&cli_addr,&cli_len);

    addr.ip = cli_addr.sin_addr;
    addr.port = ntohs(cli_addr.sin_port);

    return addr;
  }
}

/* Attempt a connection with a server 
   specified as a pair: hostname(or ip address)/port
*/

extern struct in_addr *icmGetHostIP(char *name);

filePo connectRemote(char *where,int port,icmStatus *ret)
{
  char msg[MAXLINE];
  filePo remote = icmConnectRemote(where,port,ret,msg,NumberOf(msg));

  if(remote==NULL)
    icmLogMsg("%s",msg);

  return remote;
}

filePo icmConnectRemote(char *where,int port,icmStatus *ret,
			char *retmsg,long retlen)
{
  icmTruth inited = initSockets();
  int sock;
  struct sockaddr_in serv_addr;
  char *host = icmGetHostname(where);
  struct in_addr *addr = host!=NULL?icmGetHostIP(host):NULL;

  if(inited && addr!=NULL){
    /* Attempt to establish links to the server */
    memset((char *)&serv_addr, 0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr = *addr;
    serv_addr.sin_port = htons((u_short)port);
  
    /* Create the socket ... */
    if((sock=socket(AF_INET,SOCK_STREAM,0)) == INVALID_SOCKET){
      *ret = icmError;
      if(retmsg!=NULL)
	icmStrMsg(retmsg,retlen,"invalid socket");
      return NULL;
    }
    else{
      while(connect(sock,(struct sockaddr *)&serv_addr,sizeof(serv_addr))!=0){
#ifdef WIN32
	switch(WSAGetLastError()){
#else
	switch(errno){		/* Unix version */
#endif
	case WSAEACCES:
	case WSAEADDRNOTAVAIL:
	  if(retmsg!=NULL)
	    icmStrMsg(retmsg,retlen,"Address %s not available",host);
	  icmMarkHostUnavail(host);
	  close(sock);
	  *ret = icmError;
	  return NULL;
	case WSAECONNREFUSED:
	  if(retmsg!=NULL)
	    icmStrMsg(retmsg,retlen,"Connection to %s refused",host);
	  icmMarkHostUnavail(host);
	  close(sock);
	  *ret = icmError;
	  return NULL;
	case WSAETIMEDOUT:
	  if(retmsg!=NULL)
	    icmStrMsg(retmsg,retlen,"Connection to %s timed out",host);
	  icmMarkHostUnavail(host);
	  close(sock);
	  *ret = icmError;
	  return NULL;
	case WSAENETUNREACH:
	  if(retmsg!=NULL)
	    icmStrMsg(retmsg,retlen,"Network down or %s unreachable",host);
	  icmMarkHostUnavail(host);
	  close(sock);
	  *ret = icmError;
	  return NULL;
	case WSAEALREADY:
	case WSAEINTR:
	case WSAEWOULDBLOCK:
	case WSAEINPROGRESS:
	  continue;		/* Interrupted or busy, try again*/
	default:
	  if(retmsg!=NULL)
	    icmStrMsg(retmsg,retlen,"Connection to %s refused",host);
	  icmMarkHostUnavail(host);
	  close(sock);
	  *ret = icmError;
	  return NULL;
	}
      }

      *ret = icmOk;
      return _new_io(host,ioFile,ioREAD|ioWRITE,sock,
		     _refillSocket,_flushSocket,_closeSocket,
		     _fileInReady,_fileOutReady,_configureSocket);
    }
  }
  else{
    if(retmsg!=NULL)
      icmStrMsg(retmsg,retlen,"ICM library not initialized properly");
    *ret = icmError;
    return NULL;
  }
}


