/* 
  UTF-8 parsing functions
  (c) 2000 F.G. McCabe

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Contact: Francis McCabe <fgm@fla.fujitsu.com>

*/ 

#include "config.h"		/* pick up standard configuration header */
#include "icmP.h"
#include <assert.h>
/*#include "utf.h"*/

/* The various ranges in the UTF-8 encoding are as follows:
  
  0x00000000 - 0x0000007F:
  0xxxxxxx

  0x00000080 - 0x000007FF:
  110xxxxx 10xxxxxx
*/
#define U80 ((0x6)<<5)
#define M80 ((0x7)<<5)
#define UC80(x) (((x)&M80)==U80)
#define UX80(x) ((x)&(~M80))

#define UR ((0x2)<<6)
#define MR ((0x3)<<6)
#define UCR(x) (((x)&MR)==UR)
#define UXR(x) ((x)&(~MR))

/*

  0x00000800 - 0x0000FFFF:
  1110xxxx 10xxxxxx 10xxxxxx

*/
#define U800 ((0xe)<<4)
#define M800 ((0xf)<<4)
#define UC800(x) (((x)&M800)==U800)
#define UX800(x) ((x)&(~M800))

/*
  0x00010000 - 0x001FFFFF:
  11110xxx 10xxxxxx 10xxxxxx 10xxxxxx

*/
#define U10000 ((0x1e)<<3)
#define M10000 ((0x1f)<<3)
#define UC10000(x) (((x)&M10000)==U10000)
#define UX10000(x) ((x)&(~M10000))

/*
  0x00200000 - 0x03FFFFFF:
  111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx

*/
#define U200000 ((0x3e)<<2)
#define M200000 ((0x3f)<<2)
#define UC200000(x) (((x)&M200000)==U200000)
#define UX200000(x) ((x)&(~M200000))

/*
  0x04000000 - 0x7FFFFFFF:
  1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx

*/
#define U4000000 ((0x7e)<<1)
#define M4000000 ((0x7f)<<1)
#define UC4000000(x) (((x)&M4000000)==U4000000)
#define UX4000000(x) ((x)&(~M4000000))

/*
  The  xxx  bit  positions  are  filled with the bits of the
  character code number in binary representation.  Only  the
  shortest  possible  multibyte sequence which can represent
  the code number of the character can be used.
*/


uniChar nChar(filePo f)
{
  int ch = inCh(f);
  if(ch==EOF)
    return (uniChar)EOF;
  else if(ch>=0 && ch<=0x7f)
    return (uniChar)ch;
  else if(UC80(ch))
    return UX80(ch)<<6|UXR(inCh(f));
  else if(UC800(ch)){
    uniChar code = UX800(ch)<<12;
    code |= UXR(inCh(f))<<6;
    code |= UXR(inCh(f));
    return code;
  }
  else if(UC10000(ch)){
    uniChar code = UX10000(ch)<<18;
    code |= UXR(inCh(f))<<12;
    code |= UXR(inCh(f))<<6;
    code |= UXR(inCh(f));
    return code;
  }
  else if(UC200000(ch)){
    uniChar code = UX200000(ch)<<24;
    code |= UXR(inCh(f))<<12;
    code |= UXR(inCh(f))<<6;
    code |= UXR(inCh(f));
    return code;
  }
  else if(UC4000000(ch)){
    uniChar code = UX4000000(ch)<<30;
    code |= UXR(inCh(f))<<24;
    code |= UXR(inCh(f))<<12;
    code |= UXR(inCh(f))<<6;
    code |= UXR(inCh(f));
    return code;
  }
  else
    return (uniChar)ch;
}

int utf2uni(const char *str,uniChar *buff,long len)
{
  int pos = 0;

  while(*str!='\0' && pos<len-1){
    int ch = *str++;

    if(ch>=0 && ch<=0x7f)
      buff[pos++]=ch;
    else if(UC80(ch))
      buff[pos++]= UX80(ch)<<6|UXR(*str++);
    else if(UC800(ch)){
      uniChar code = UX800(ch)<<12;
      code |=UXR(*str++)<<6;
      code |=UXR(*str++);
      buff[pos++]=code;
    }
    else if(UC10000(ch)){
      uniChar code = UX10000(ch)<<18;
      code |=UXR(*str++)<<12;
      code |=UXR(*str++)<<6;
      code |=UXR(*str++);
      buff[pos++]=code;
    }
    else if(UC200000(ch)){
      uniChar code = UX200000(ch)<<24;
      code |=UXR(*str++)<<18;
      code |=UXR(*str++)<<12;
      code |=UXR(*str++)<<6;
      code |=UXR(*str++);
      buff[pos++]=code;
    }
    else if(UC4000000(ch)){
      uniChar code = UX4000000(ch)<<30;
      code |=UXR(*str++)<<24;
      code |=UXR(*str++)<<18;
      code |=UXR(*str++)<<12;
      code |=UXR(*str++)<<6;
      code |=UXR(*str++);
      buff[pos++]=code;
    }
    else
      return -1;		/* invalid utf-8 character? */
  }
  if(pos<len){
    buff[pos]=0;		/* terminate the unicode string */
    return pos;
  }
  else
    return -1;			/* we couldnt perform the mapping */
}


icmStatus pChar(filePo f,uniChar c)
{
  if(0<=c && c<=0x7f)
    return outCh(f,(int)c);
  else if(0x80<=c && c<=0x7FF){
    icmStatus ret = outCh(f,M80|UX80(c>>6));
    if(ret!=icmOk)
      return ret;
    return outCh(f,MR|UXR(c));
  }
  else if(0x800<=c && c<=0x0000FFFF){
    icmStatus ret = outCh(f,M800|UX800(c>>12));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>6));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c));
    return ret;
  }
  else if(0x00010000<=c && c<=0x001FFFFF){
    icmStatus ret = outCh(f,M10000|UX10000(c>>18));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>12));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>6));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c));
    return ret;
  }
  else if(0x00200000<=c && c<=0x03FFFFFF){
    icmStatus ret = outCh(f,M200000|UX200000(c>>24));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>18));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>12));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>6));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c));
    return ret;
  }
  else{
    icmStatus ret = outCh(f,M4000000|UX4000000(c>>24));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>24));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>18));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>12));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c>>6));
    if(ret==icmOk)
      ret=outCh(f,MR|UXR(c));
    return ret;
  }
}

icmStatus outUni(filePo f,uniChar *s)
{
  while(*s!=0){
    icmStatus ret = pChar(f,*s++);
    if(ret!=icmOk)
      return ret;
  };
  return icmOk;
}

icmStatus uniMsg(filePo f,void *data,int width,int precision,icmTruth alt)
{
  return outUni(f,(uniChar*)data);
}


char *uni2utf(uniChar *s,char *buff,long len)
{
  long pos = 0;

  while(*s!=0 && pos<len-1){
    uniChar c = *s++;
    if(0<=c && c<=0x7f)
      buff[pos++]=c;
    else if(0x80<=c && c<=0x7FF){
      buff[pos++]=M80|UX80(c>>6);
      buff[pos++]=MR|UXR(c);
    }
    else if(0x800<=c && c<=0x0000FFFF){
      buff[pos++]=M800|UX800(c>>12);
      buff[pos++]=M80|UX80(c>>6);
      buff[pos++]=MR|UXR(c);
    }
    else if(0x00010000<=c && c<=0x001FFFFF){
      buff[pos++]=M10000|UX10000(c>>18);
      buff[pos++]=M800|UX800(c>>12);
      buff[pos++]=M80|UX80(c>>6);
      buff[pos++]=MR|UXR(c);
    }
    else if(0x00200000<=c && c<=0x03FFFFFF){
      buff[pos++]=M200000|UX200000(c>>24);
      buff[pos++]=M10000|UX10000(c>>18);
      buff[pos++]=M800|UX800(c>>12);
      buff[pos++]=M80|UX80(c>>6);
      buff[pos++]=MR|UXR(c);
    }
    else{
      buff[pos++]=M4000000|UX4000000(c>>24);
      buff[pos++]=M200000|UX200000(c>>24);
      buff[pos++]=M10000|UX10000(c>>18);
      buff[pos++]=M800|UX800(c>>12);
      buff[pos++]=M80|UX80(c>>6);
      buff[pos++]=MR|UXR(c);
    }
  }
  buff[pos]='\0';

  assert(pos<len);
  return buff;
}

int uniStrLen(uniChar *s)
{
  int len = 0;

  assert(s!=NULL);

  while(*s++!=0)
    len++;
  return len;
}

uniChar *uniCat(uniChar *dest,long len,const uniChar *src)
{
  int pos = 0;
  while(pos<len-1 && dest[pos]!=0)
    pos++;

  while(pos<len-1 && *src!=0)
    dest[pos++]=*src++;
  dest[pos]=0;
  return dest;
}

uniChar *uniCpy(uniChar *dest,long len,const uniChar *src)
{
  int pos = 0;

  while(pos<len-1 && *src!=0)
    dest[pos++]=*src++;
  dest[pos]=0;
  return dest;
}
