/***********************************************************************
(c) 1994-1999 Imperial College, F.G. McCabe  and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

/*
 * Manage a queue of worker threads 
 * each thread picks up a work item and `runs' with it, depending on the 
 * kind of workers being set up
 */

#include "config.h"		/* Invoke configuration header */
#include <signal.h>
#ifdef WIN32
#include <io.h>
#include <windows.h>
#else
#include <unistd.h>
#endif
#include "icmP.h"
#include "icmW.h"

static icmPoolBase wpool = NULL;
static icmPoolBase qpool = NULL;
static void blockInterrupts(void);

/*
 * A separate thread is established for each attempt to make a remote connection
 * If successful, the result is that a remote commserver is registered locally, and
 * we can start forwarding messages to it.
 */

#ifdef WKRTRACE
icmTruth traceWkr = icmNo;
#endif

typedef struct _workItem_ *workPo;

typedef struct _workItem_ {
  icmWorkProc proc;
  void *cl;
  workPo next;
} WorkItemRec;

typedef struct _queue_rec_ {
  pthread_mutex_t mutex;
  pthread_cond_t avail;
  pthread_cond_t empty;
  pthread_cond_t idle;
  workPo Q,endQ;
  int nthreads;
  int active;
  icmTruth autoGrow;
  pthread_t grower;
  pthread_t shrinker;
} QueueRec;

static void *queueWorker(void *arg);
static void *queueGrower(void *arg);
static void *queueShrinker(void *arg);

static pthread_once_t wkrInit = PTHREAD_ONCE_INIT;

static void initWorkers(void)
{
  wpool = icmInitPool(sizeof(WorkItemRec),16);
  qpool = icmInitPool(sizeof(QueueRec),4);
}
  
icmQueuePo icmCreateWorkers(int count,icmTruth autoGrow)
{
  pthread_once(&wkrInit,initWorkers);

  {
    icmQueuePo queue = (icmQueuePo)icmAllocPool(qpool);
    int i;
    pthread_attr_t attr;
    
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

    queue->Q = queue->endQ = 0;
    queue->nthreads = count;
    queue->active = 0;
    queue->autoGrow = autoGrow;

    icmCreateMutex(&queue->mutex);
    icmLockMutex(&queue->mutex);
    pthread_cond_init(&queue->avail,NULL);
    pthread_cond_init(&queue->empty,NULL);
    pthread_cond_init(&queue->idle,NULL);

    for(i=0;i<count;i++){
      pthread_t worker;
      if(pthread_create(&worker,&attr,queueWorker,queue)!=0)
	logMsg(icmLogFile,"Unable to create worker thread %d",i);
    }

    if(autoGrow){
      if(pthread_create(&queue->grower,&attr,queueGrower,queue)!=0)
	logMsg(icmLogFile,"Unable to create grower thread %d",i);
#ifndef WKRTRACE
      if(pthread_create(&queue->shrinker,&attr,queueShrinker,queue)!=0)
	logMsg(icmLogFile,"Unable to create grower thread %d",i);
#endif
    }
      
    pthread_attr_destroy(&attr);
    icmUnlockMutex(&queue->mutex);
    return queue;
  }
}

/* This is the main thread worker function that processes items from the queue as
 * they become available 
 */

static void *queueWorker(void *arg)
{
  icmQueuePo queue = (icmQueuePo)arg;
  workPo item;

  blockInterrupts();		/* These are not interruptable... */

  icmLockMutex(&queue->mutex);

  while(1){

    while(queue->Q==NULL){
#ifdef WKRTRACE
      if(traceWkr)
	icmLogMsg("%d: worker on 0x%x waiting for work",pthread_self(),
		  (Ulong)queue);
#endif
      icmWaitCond(&queue->avail,&queue->mutex);
    }

    queue->active++;

    if(queue->autoGrow){
      if(queue->active==queue->nthreads)
	pthread_cond_signal(&queue->empty);
      else if(queue->active<queue->nthreads/4)
	pthread_cond_signal(&queue->idle);
    }

#ifdef WKRTRACE
    if(traceWkr)
      logMsg(icmLogFile,"%d: dequeue item on 0x%x",pthread_self(),
	     (Ulong)queue);
#endif

    item = queue->Q;
    if(queue->Q==queue->endQ)
      queue->endQ = queue->Q = NULL;
    else
      queue->Q = queue->Q->next;

    if(queue->Q!=NULL)
      pthread_cond_broadcast(&queue->avail); /* queue is still not empty */
    
    icmUnlockMutex(&queue->mutex);

    item->proc(item->cl);

    icmFreePool(wpool,item);

    icmLockMutex(&queue->mutex);
    queue->active--;
  }
}

/*
 * This thread function is triggered whenever the worker queue becomes empty
 * I.e., all workers are busy
 */

static void *queueGrower(void* arg)
{
  icmQueuePo queue = (icmQueuePo)arg;
  pthread_attr_t attr;
    
  blockInterrupts();		/* Grower is not interruptable... */

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

  icmLockMutex(&queue->mutex);

  while(1){
    pthread_t newWorker;

    while(queue->active<queue->nthreads)
      pthread_cond_wait(&queue->empty,&queue->mutex);

    if(pthread_create(&newWorker,&attr,queueWorker,queue)!=0)
      logMsg(icmLogFile,"Unable to create new worker thread");
    else
      queue->nthreads++;
  }
  return NULL;			/* Will never get here */
}

/*
 * This thread function is triggered whenever there are too many idle workers
 */

static void shrinkOne(void *arg)
{
  icmQueuePo queue = (icmQueuePo)arg;

  icmLockMutex(&queue->mutex);

  queue->nthreads--;

  icmUnlockMutex(&queue->mutex);

  pthread_exit(NULL);		/* kill this thread */
}

static void *queueShrinker(void* arg)
{
  icmQueuePo queue = (icmQueuePo)arg;
  pthread_attr_t attr;
    
  blockInterrupts();		/* Shrinker is not interruptable... */

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

  while(1){
#ifdef WIN32
    Sleep(60*1000);			/* We run this at most once/minute */
#else
    sleep(60);
#endif

    icmLockMutex(&queue->mutex);

    while(queue->active>=queue->nthreads/4)
      pthread_cond_wait(&queue->idle,&queue->mutex);

    icmUnlockMutex(&queue->mutex);

    icmEnqueueItem(queue,shrinkOne,queue);
  }
  return NULL;			/* Will never get here */
}

icmStatus icmEnqueueItem(icmQueuePo queue,icmWorkProc proc,void *cl)
{
  workPo item = icmAllocPool(wpool);

#ifdef WKRTRACE
  if(traceWkr)
    logMsg(icmLogFile,"%d: enqueue item on 0x%x",pthread_self(),(Ulong)queue);
#endif

  if(item!=NULL){
    item->cl = cl;
    item->next = NULL;
    item->proc = proc;

    icmLockMutex(&queue->mutex);
    if(queue->Q == NULL)
      queue->Q = queue->endQ = item;
    else
      queue->endQ = queue->endQ->next = item;
    pthread_cond_signal(&queue->avail);
    icmUnlockMutex(&queue->mutex);

    return icmOk;
  }
  else
    return icmError;
}

static void blockInterrupts(void)
{
#ifdef HAVE_PTHREAD_SIGMASK
  sigset_t signals;

#ifdef WKRTRACE
  return;
#endif

  sigemptyset(&signals);

  sigaddset(&signals,SIGINT);
  sigaddset(&signals,SIGTERM);

  pthread_sigmask(SIG_BLOCK,&signals,NULL); /* block SIGINT and SIGTERM */
#endif
}

