/***********************************************************************

Post box model for the ICM communications server
(c) 1994-1998 Imperial College, F.G. McCabe and Fujitsu Labs

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

Contact: fgm@fla.fujitsu.com
***********************************************************************/

#include "config.h"		/* Invoke configuration header */
#include "daemon.h"		/* Comm server's functions */
#include "office.h"		/* Private post office structures */

#ifdef COMMTRACE
extern icmTruth traceMsg;
#endif

static icmHashPo boxes = NULL;

static icmPoolBase msgpool = NULL;
static icmPoolBase boxpool = NULL;

static icmQueuePo boxWorkers = NULL;

static void TermPostBox(PostPo box);
static void readMessages(void *arg);
static void cleanUpBox(void *arg);
static void writeMessages(void *arg);
static void unDeliverMessage(MsgPo msg);
static void undeliveredBox(void);
static void *redeliverMessages(void *arg);

static PostPo undelivered = NULL;
static PostPo CreatePostBox(icmIpPort *addr,icmConn conn,char *host,time_t timeout);

static long cmp_box(void *a,void *b)
{
  icmIpPort *A = (icmIpPort *)a;
  icmIpPort *B = (icmIpPort *)b;

  if(A->ip.s_addr==B->ip.s_addr && A->port == B->port)
    return 0;
  else
    return 1;
}

static unsigned long hash_box(void *a)
{
  icmIpPort *A = (icmIpPort *)a;
  return (long)((A->ip.s_addr<<8)+A->port);
}

static icmStatus dispBox(filePo f,void *h,int width,int precision,icmTruth alt)
{
  PostPo box = (PostPo)h;

  if(box!=NULL) {
    char ip[MAXSTRING];
    icmIPToStr(&box->addr.ip,ip,NumberOf(ip));
    return outMsg(f,"{%s/%d}",ip,box->addr.port);
  }
  else
    return outMsg(f,"{null-box}");
}

status initBoxes(int count)
{
  msgpool = icmInitPool(sizeof(MsgRec),32);

  boxes = icmNewHash(16,hash_box,cmp_box,NULL);
  boxpool = icmInitPool(sizeof(PostBoxRec),16);

  _installMsgProc('B',dispBox);	/* extend outMsg to cope with boxes */

  undeliveredBox();

  boxWorkers = icmCreateWorkers(count,icmYes);

  if(msgpool!=NULL)
    return initForwarding(4);	/* We allow up to 4 outstanding remote sessions*/
  else
    return MakeError("initBoxes: initialization"); /* Problem in initialization */
}

static icmStatus _closeBox(void *n,void *r,void *c)
{
  PostPo box = (PostPo)r;

  TermPostBox(box);		/* Phase I of box termination */
  return icmOk;
}

void closeBoxes(void)
{
  icmProcessTable(_closeBox,boxes,NULL);
}

static PostPo CreatePostBox(icmIpPort *addr,icmConn conn,char *host,time_t timeout)
{
  PostPo box = (PostPo)icmAllocPool(boxpool);

  if(addr!=NULL)
    box->addr = *addr;
  else{
    box->addr.ip.s_addr = htonl(0);
    box->addr.port = 0;
  }

  strcpy(box->host,host);
  box->conn = conn;
  box->queue = NULL;
  box->timeout = timeout;
  box->timed_out = icmNo;

  box->hasWtr = icmNo;
  box->hasRdr = icmNo;
  box->terminating = icmNo;
  box->refCount = 1;

  icmCreateMutex(&box->mutex);
  pthread_cond_init(&box->enqueued,NULL);

  pthread_cond_init(&box->dead,NULL);

  if(timeout>0)
    box->out_time = time(NULL)+timeout;
  else
    box->out_time = 0;		/* Signal no timeout */

  return box;
}

void refBox(PostPo box)
{
  if(box!=NULL){
    icmLockMutex(&box->mutex);
    box->refCount++;
    icmUnlockMutex(&box->mutex);
  }
}

void unRefBox(PostPo box)
{
  if(box!=NULL){
    icmLockMutex(&box->mutex);
    box->refCount--;
    icmUnlockMutex(&box->mutex);
  }
}

static void dumpBox(PostPo box)
{
  MsgPo msg=box->queue;

  outMsg(icmLogFile,"Dump of box %B:%s",box,box->terminating?"*":"");

  if(box->hasRdr && box->hasWtr)
    outMsg(icmLogFile," reader=%d, writer=%d\n",box->reader,box->writer);
  else if(box->hasRdr)
    outMsg(icmLogFile," reader=%d\n",box->reader);
  else if(box->hasWtr)
    outMsg(icmLogFile," writer=%d\n",box->writer);

  while(msg!=NULL){
    outMsg(icmLogFile,"Waiting message %M for %h\n",msg->msg,msg->recip);
    msg = msg->next;
  }
}

static icmStatus dumpBx(void *n,void *r,void *c)
{
  dumpBox((PostPo)r);
  return icmOk;
}

void dumpBoxes(void)
{
  icmProcessTable(dumpBx,boxes,NULL);
}

/*
 * Phase I of terminate a post box 
 * We must put any outstanding messages into the undelivered Q
 */
static void TermPostBox(PostPo box)
{
#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: Closing down box %B",pthread_self(),box);
#endif

  box->terminating = icmYes;	/* we are now closing this one down */
  deregisterAllOnBox(box);
  
  /* We must be able to lock the box to terminate it */
  icmLockMutex(&box->mutex);

  box->refCount--;

  {
    MsgPo msg = box->queue;

    while(msg!=NULL){
      MsgPo next = msg->next;
      
      unDeliverMessage(msg);

      msg = next;
    }
  }

  if(box->hasWtr)		/* The writer will initiate Phase II */
    icmSignalCond(&box->enqueued);
  else
    icmEnqueueItem(boxWorkers,cleanUpBox,(void*)box); 
  icmUnlockMutex(&box->mutex);
}

/* Phase II of box termination */
/* Free resources used by a box */
static void cleanUpBox(void *arg)
{
  PostPo box = (PostPo)arg;

  icmUninstall((void*)&box->addr,boxes);

  icmLockMutex(&box->mutex);

  while(box->refCount>0 && (box->hasRdr==icmYes||box->hasWtr==icmYes))
    icmWaitCond(&box->dead,&box->mutex);

  icmCloseComms(box->conn);

  pthread_cond_destroy(&box->enqueued);
  icmFreePool(boxpool,box);
  icmDestroyMutex(&box->mutex);
}

void clearOSMessages(agentPo agent,PostPo box)
{
  icmLockMutex(&box->mutex); 

  {
    MsgPo *msg = &box->queue;

    while(*msg!=NULL){
      MsgPo next = (*msg)->next;

      if(agentOf((*msg)->recip)==agent){
	unDeliverMessage(*msg);
	*msg = next;
      }
      else
	msg = &(*msg)->next;
    }
  }

  icmUnlockMutex(&box->mutex); 
}

PostPo OpenPostBox(icmIpPort addr,icmConn conn,char *host,time_t timeout)
{
  PostPo box = icmSearch((void*)&addr,boxes);

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: creating new postbox",pthread_self());
#endif

  if(box!=NULL)
    return box;
  else{
    box = CreatePostBox(&addr,conn,host,timeout);

    if(icmEnqueueItem(boxWorkers,readMessages,(void*)box)!=icmOk ||
       icmEnqueueItem(boxWorkers,writeMessages,(void*)box)!=icmOk){
      icmLogMsg("Unable to schedule creation of reader/writer threads");
      box->conn=NULL;
      icmFreePool(boxpool,box);
      return NULL;		/* couldnt create the box */
    }

    icmInstall((void*)&box->addr,box,boxes);
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: new postbox %B",pthread_self(),box);
#endif
    return box;
  }
}

/*
 * This is a core function to move a message from a source Q to a destination Q
 * The actual sending is done asynchronously
 */
status PostMsg(PostPo box,icmHandle recip,icmHandle sender,icmOption opts,icmMsg m)
{
  MsgPo msg = icmAllocPool(msgpool);

  if(box==NULL){
    icmLogMsg(BOLD_ESC_ON "%B: undelivered message: %M for %h from %h" BOLD_ESC_OFF,
	      box,m,recip,sender);
    box = undelivered;
  }
#ifdef COMMTRACE
  else if(traceMsg)
    icmLogMsg("%d: posting msg %M for %h on to %B",pthread_self(),m,recip,box);
#endif

  msg->msg = m;
  msg->recip = recip;
  msg->snder = sender;
  msg->opts = opts;

  msg->sending = icmNo;		/* We havent started sending this message yet */
  msg->next = NULL;

  icmLockMutex(&box->mutex); 

  if(box->queue == NULL)
    box->queue = msg;
  else{
    MsgPo f = box->queue;	/* attach message to the end of the Q */
    
    while(f->next!=NULL)
      f = f->next;
    f->next = msg;
  }

  icmSignalCond(&box->enqueued); /* box has at least one msg in it */
  icmUnlockMutex(&box->mutex); 

  return okay;			/* We have managed to store the message */
}

static void unDeliverMessage(MsgPo msg)
{
#ifdef COMMTRACE
  if(traceMsg)
    icmLogMsg(BOLD_ESC_ON "undelivering message %M" BOLD_ESC_OFF,msg->msg);
#endif
  
  icmLockMutex(&undelivered->mutex); 

  msg->next = NULL;
  if(undelivered->queue == NULL){
    undelivered->queue = msg;
  }
  else{
    MsgPo f = undelivered->queue;	/* attach message to the end of the Q */
    
    while(f->next!=NULL)
      f = f->next;
    f->next = msg;
  }

  icmSignalCond(&undelivered->enqueued);
  icmUnlockMutex(&undelivered->mutex); 
}

static icmTruth inLocations(char *src,char **locs,int count)
{
  int i;

  for(i=0;i<count;i++){
    if(strncmp(src,locs[i],MAXSTRING)==0)
      return icmYes;
  }

  return icmNo;
}

icmHandle removeLocation(icmHandle h,char *duff)
{
  int count,Ocount,ncount;
  char *target,*name,*home,**locs,*Ohome,**Olocs;
  char *nLocs[20];

  icmAnalyseHandle(ourHandle,NULL,NULL,&Ohome,&Olocs,&Ocount);

  if(icmAnalyseHandle(h,&target,&name,&home,&locs,&count)==icmOk){
    int i;

    ncount=0;
    for(i=0;i<count;i++){
      if((duff==NULL || strcmp(locs[i],duff)!=0) &&
	 !inLocations(locs[i],Olocs,Ocount)){
	nLocs[ncount++]=locs[i];
      }
    }

    return icmMakeHandle(target,name,home,ncount,nLocs);
  }
  else
    return NULL;		/* something went wrong here */
}

icmHandle mergeLocations(icmHandle h1,icmHandle h2)
{
  int h1Count,h2Count,ncount=0;
  char *tgt,*name,*home,**h1Locs,**h2Locs;
  char *nLocs[20];
  int i;

  icmAnalyseHandle(h1,&tgt,&name,&home,&h1Locs,&h1Count);
  icmAnalyseHandle(h2,NULL,NULL,NULL,&h2Locs,&h2Count);

  for(i=0;i<h1Count;i++){
    if(!inLocations(h1Locs[i],h2Locs,h2Count))
      nLocs[ncount++]=h1Locs[i];
  }

  for(i=0;i<h2Count;i++)
    nLocs[ncount++]=h2Locs[i];

  return icmMakeHandle(tgt,name,home,ncount,nLocs);
}

/*
 * Thread function that processes the messages in a post box Q 
 */
static void writeMessages(void *arg)
{
  PostPo box = (PostPo)arg;

  blockSignals();

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: deliver messages to %B",pthread_self(),box);
#endif

  box->hasWtr = icmYes;
  box->writer = pthread_self();

  icmLockMutex(&box->mutex);

  while(!box->terminating){
    while(!box->terminating && box->queue==NULL)
      icmWaitCond(&box->enqueued,&box->mutex);

    if(!box->terminating && box->queue!=NULL){
      MsgPo msg = box->queue;	/* Dequeue the first message */
      icmHandle recip = removeLocation(msg->recip,NULL);

      box->queue = msg->next;
      icmUnlockMutex(&box->mutex); /* allow others to add messages */

      {
	icmMsg chunk=icmFormatMsg(NULL,"%(%h%h%O%#%)",
				  recip,msg->snder,msg->opts,msg->msg);

	assert(box->conn!=NULL && box->conn->conn!=NULL);

	if(outBlock(box->conn->conn,(char*)chunk->data,chunk->used)!=icmOk){
	  box->hasWtr = icmNo;
	  TermPostBox(box);	/* put everything on the undelivered list */
	  return;		/* We must return 'cos box no longer valid */
	}

	flushFile(box->conn->conn);

	icmReleaseMsg(msg->msg); /* we have delivered the message */
	icmReleaseMsg(chunk);
	icmReleaseHandle(recip); /* We can finally get rid of the handles... */
	icmReleaseHandle(msg->recip);
	icmReleaseHandle(msg->snder);
	icmReleaseOptions(msg->opts);
	icmFreePool(msgpool,msg);
      }
      icmLockMutex(&box->mutex);
    }
  }

  box->hasWtr = icmNo;
  icmSignalCond(&box->dead);	/* box can be collected */
  icmEnqueueItem(boxWorkers,cleanUpBox,(void*)box); 
  icmUnlockMutex(&box->mutex);	/* make sure the mutex is unlocked */
}

static void readMessages(void *arg)
{
  PostPo box = (PostPo)arg;

  blockSignals();

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: read all messages from %B",pthread_self(),box);
#endif

  box->reader = pthread_self();
  box->hasRdr = icmYes;

  while(!box->terminating){
    icmHandle recip,sender;
    icmOption opts;
    icmMsg msg;

    switch(icmGetMsg(box->conn,&recip,&sender,&opts,&msg,NULL)){
    case icmOk:{
      PostPo rBox;
      icmHandle rh = agentAddress(recip);

      if(rh!=NULL && recip!=rh){
	icmKeepHandle(rh);
	icmReleaseHandle(recip);
	recip = rh;		/* We have a forwarding request in place */
      }

      /* Add our name to the audittrail of this message */
      {
	icmOption o = opts;
	
	while(o!=NULL){
	  if(o->opt==icmAuditTrail){
	    long count;
	    icmHandle *tr = o->o.trail;
	    
	    for(count=0;*tr!=NULL;count++,tr++)
	      ;
	    o->o.trail = realloc(o->o.trail,sizeof(icmHandle)*(count+1));
	    o->o.trail[count++] = ourHandle;
	    o->o.trail[count] = NULL;
	    break;
	  }
	  o = o->next;
	}
      }

      if(isOurHandle(recip)==okay)
	ServerMsg(recip,sender,opts,msg,box);
      else if((rBox=agentBox(recip))!=NULL)
	PostMsg(rBox,recip,sender,opts,msg); /* deliver the message */
      else
	forwardMessage(recip,sender,opts,msg); /* forward the message */

      continue;
    }
    case icmFailed:
      continue;
    case icmError:
      icmLogMsg("Problem with connection to %B",box);
    case icmEndFile:{
      box->hasRdr = icmNo;	/* Were outta here */
      TermPostBox(box);
      return;			/* box is no longer valid */
    }
    }
  }
  box->hasRdr = icmNo;
  icmSignalCond(&box->dead);	/* box can be collected */
}

/*
 * thread routine that tries to deal with undelivered messages
 * It is woken up whenever an agent registers with the commserver
 */

static void *redeliverMessages(void *arg)
{
  PostPo box = (PostPo)arg;

  blockSignals();

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: thread to re-deliver messages",pthread_self());
#endif

  assert(box==undelivered);

  icmLockMutex(&box->mutex);

  while(icmYes){
    MsgPo *msg = &box->queue;

    icmWaitCond(&box->enqueued,&box->mutex);

    while(*msg!=NULL){
      PostPo nbox = fwdBox((*msg)->recip);

      if(nbox!=NULL){
	MsgPo next = (*msg)->next;
	icmHandle recip = (*msg)->recip;
	icmHandle snder = (*msg)->snder;
	icmOption opts = (*msg)->opts;

#ifdef COMMTRACE
	if(traceMsg)
	  icmLogMsg("%d: able to re-deliver message %M to %h",pthread_self(),
		    (*msg)->msg,recip);
#endif

	PostMsg(nbox,recip,snder,opts,(*msg)->msg);
	icmFreePool(msgpool,*msg);
	*msg = next;
      }
      else
	msg = &(*msg)->next;
    }
  }
  return NULL;			/* will never actually return */
}

void tryRedelivery(void)
{
  icmLockMutex(&undelivered->mutex);
  icmSignalCond(&undelivered->enqueued); 
  icmUnlockMutex(&undelivered->mutex); 
}

static void undeliveredBox(void)
{
  PostPo box = undelivered = CreatePostBox(NULL,NULL,"",-1);

  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

  if(pthread_create(&box->writer,&attr,redeliverMessages,(void*)box)!=0){
    icmLogMsg("Unable to create thread for message redelivery");
    box->hasWtr = icmNo;
  }
  else
    pthread_attr_destroy(&attr);
}


