/*
 * Handle the registration, detachment and address forwarding for ogents
 * Part of the ICM communications server
 * (c) 1994-1998 Imperial College, F.G.McCabe and Fujitsu labs
 * All rights reserved
 */
#include "config.h"		/* Invoke configuration header */
#include "daemon.h"		/* Our data structures etc. */
#include "office.h"		/* Private post office structures */

static icmHashPo postoffice = NULL;
static icmPoolBase agpool = NULL;
static icmPoolBase mtpool = NULL;

static icmStatus clsAgent(void *n,void *r,void *c);
static agentPo findAgent(icmHandle handle);
static void triggerMonitors(char *event,agentPo agent);
static void removeMonitors(agentPo client);

static pthread_mutex_t agentMutex = PTHREAD_MUTEX_INITIALIZER;

typedef struct _agent_name_{
  char name[MAXNAME];
  char home[MAXNAME];
}AgentNameRec;

typedef struct _ag_monitor_ *monitorPo;

typedef struct _agent_rec_ {
  AgentNameRec name;
  icmHandle handle;		/* What is this agent's current handle? */
  icmHandle forward;		/* Where to forward messages to */
  time_t lease;			/* When does this registration run out? */
  PostPo box;			/* If this agent is attached and registered */
  monitorPo monitors;
  long tag;			/* A tag used to automatically reestablish connections */
} AgentRec;

typedef struct _ag_monitor_ {
  icmHandle client;		/* Who is interested in this agent? */
  monitorPo next;
} MonitorRec;

/* Name hashing functions */
unsigned long hash_name(void *n)
{
  register char *name = n;
  register long hash = 0;

  if(name!=NULL)
    while(*name){
      hash = (hash<<4)+(hash^*name);
      name++;
    }
  return hash;
}

long cmp_name(void *a, void *b)
{
  return strncmp((char*)a,(char*)b,MAXNAME);
}

static unsigned long hashAgent(void *n)
{
  agentNamePo name = (agentNamePo)n;

  return hash_name(name->name)+hash_name(name->home);
}

static long cmpAgent(void *a, void *b)
{
  agentNamePo A = (agentNamePo)a;
  agentNamePo B = (agentNamePo)b;
  int cmp = strncmp(A->name,B->name,MAXNAME);

  if(cmp==0)
    cmp = strncmp(A->home,B->home,MAXNAME);

  return cmp;
}

static icmStatus dispAg(filePo f,void *h,int width,int precision,icmTruth alt)
{
  agentPo a = (agentPo)h;

  if(a==NULL)
    return outMsg(f,"<null agent>");
  else
    return outMsg(f,"<%s@%s%s>",a->name.name,a->name.home,a->lease!=0?ctime(&a->lease):"");
}

status initPostOffice(icmHandle ourHandle)
{
  postoffice = icmNewHash(32,hashAgent,cmpAgent,NULL);
  agpool = icmInitPool(sizeof(AgentRec),16);
  mtpool = icmInitPool(sizeof(MonitorRec),16);

  _installMsgProc('A',dispAg);	/* extend outMsg to cope with agent structures */

  initServer();			/* Initialize server functiona */

  if(agpool!=NULL && mtpool!=NULL && postoffice!=NULL)
    return initBoxes(4);
  else{
    error_code = "PostOffice: initialization"; /* Problem in initialization */
    return error;
  }
}

static pthread_once_t closeOnce = PTHREAD_ONCE_INIT;

static void closePO(void)
{
  if(postoffice!=NULL){
#ifdef SERVERTRACE
    if(traceSvr){
      icmLogMsg("Closing down post office");
      dumpPostOffice();
    }
#endif

    icmLockMutex(&agentMutex);
    icmProcessTable(clsAgent,postoffice,postoffice);
    icmUnlockMutex(&agentMutex);
    closeBoxes();
  }
}

void closePostOffice(void)
{
  pthread_once(&closeOnce,closePO);
}
  
static icmStatus showAgent(void *n,void *r,void *c)
{
  agentPo a = (agentPo)r;
  monitorPo mon = a->monitors;

  outMsg(icmLogFile,"%A",a);

  while(mon!=NULL){
    outMsg(icmLogFile," ?-%h",mon->client);
    mon=mon->next;
  }

  if(a->lease!=0)
    outMsg(icmLogFile,"%s",ctime(&a->lease));

  if(a->forward!=NULL)
    outMsg(icmLogFile," --> %h\n",a->forward);
  else if(a->handle!=NULL)
    outMsg(icmLogFile," = %h\n",a->handle);
  else
    outMsg(icmLogFile," detached\n");
  return icmOk;
}

void dumpPostOffice(void)
{
  icmLogMsg("All registered agents");
  icmProcessTable(showAgent,postoffice,NULL);
  flushFile(icmLogFile);
}

status registerAgent(icmHandle handle,icmHandle *real,time_t lease,PostPo box)
{
  status ret = okay;

  icmLockMutex(&agentMutex);

  {
    agentPo agent = findAgent(handle);

    if(agent==NULL){
      char *name,*home;

#ifdef SERVERTRACE
      if(traceSvr)
	icmLogMsg("%d: registering agent %h at %B",pthread_self(),handle,box);
#endif

      agent = (agentPo)icmAllocPool(agpool); /* create a new agent structure */

      icmAnalyseHandle(handle,NULL,&name,&home,NULL,NULL);

      icmStrNCpy(agent->name.name,name,MAXNAME);
      icmStrNCpy(agent->name.home,home,MAXNAME);

      agent->box = box;
      refBox(box);
      agent->handle = mergeLocations(handle,ourHandle);
      agent->forward = NULL;	/* This too probably */
      agent->monitors = NULL;
      agent->lease = lease;

      if(real!=NULL)
	*real = agent->handle;

      icmInstall(&agent->name,agent,postoffice);

#ifdef SERVERTRACE
      if(traceSvr)
	icmLogMsg("%d: agent registered %A at %B",pthread_self(),agent,box);
#endif
    }
    else{
      if(agent->box == NULL && agent->handle==NULL && agent->forward==NULL){
	agent->box = box;
	refBox(box);
	agent->handle = mergeLocations(handle,ourHandle);
	agent->forward = NULL;
	agent->lease = lease;

	if(real!=NULL)
	  *real = agent->handle;
	triggerMonitors("icmRegisterAgent",agent);
      }
      else{
	ret = MakeError("duplicate agent"); /* Cant register more than once */
#ifdef SERVERTRACE
	if(traceSvr){
	  icmLogMsg("%d: duplicate agent %A ",pthread_self(),agent);
	  exit(100);
	}
#endif

      }
    }
  }

  icmUnlockMutex(&agentMutex);
  tryRedelivery();
  return ret;
}

/* must be invoked when agentMutex is locked */
static void closeAgent(agentPo agent)
{
#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: deregistering agent %A",pthread_self(),agent);
#endif

  if(agent->box!=NULL)		/* Clear all outstanding messages for this agent */
    clearOSMessages(agent,agent->box);

  removeMonitors(agent);	/* remove all monitors owned by this agent */

  if(agent->monitors==NULL){	/* There are no monitors on this agent */
    icmHandle h;

    icmUninstall(&agent->name,postoffice);

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: agent deregistered %A",pthread_self(),agent);
#endif
    h = agent->handle;
    agent->handle = NULL;
    icmReleaseHandle(h);
    h = agent->forward;
    agent->forward = NULL;
    icmReleaseHandle(h);
    unRefBox(agent->box);
    icmFreePool(agpool,agent);
  }
  else{
    icmHandle h;

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: agent %A deregistered, but keeping for monitors",
		pthread_self(),agent);
#endif
    h = agent->handle;
    agent->handle = NULL;
    icmReleaseHandle(h);

    h = agent->forward;
    agent->forward = NULL;
    icmReleaseHandle(h);

    unRefBox(agent->box);
    agent->box = NULL;
  }
}

status deregisterAgent(icmHandle handle,PostPo box)
{
  status ret = okay;

  icmLockMutex(&agentMutex);

  {
    agentPo agent = findAgent(handle);

    if(agent!=NULL){
      if(box!=agent->box)
	ret = MakeError("agent not registered on this connection");
      else{
	triggerMonitors("icmDeregisterAgent",agent);
	closeAgent(agent);
      }
    }
    else
      ret = MakeError("agent not registered");
  }

  icmUnlockMutex(&agentMutex);
  return ret;
}

typedef struct {
  PostPo box;
  time_t now;
} deregInfoRec;

static icmStatus dereg(void *n,void *r,void *c)
{
  deregInfoRec *info = (deregInfoRec*)c;
  agentPo agent = (agentPo)r;

  /* This agent is attached to this box*/
  if(agent->box==info->box){

    if(info->now>agent->lease){
#ifdef SERVERTRACE
      if(traceSvr)
	icmLogMsg("%d: forcibly deregister %A from %B",pthread_self(),
		  agent,info->box);
#endif

      triggerMonitors("icmDeregisterAgent",agent);
      closeAgent(agent);
    }
    else{
#ifdef SERVERTRACE
      if(traceSvr)
	icmLogMsg("%d: forcibly detach %A from %B",pthread_self(),
		  agent,info->box);
#endif
      detachAgent(agent->handle,info->box);
    }
  }
  return icmOk;
}

void deregisterAllOnBox(PostPo box)
{
  deregInfoRec info = {box, time(NULL)};

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("Deregistering all agents on box %B",box);
#endif

  icmLockMutex(&agentMutex);
  icmProcessTable(dereg,postoffice,&info);
  icmUnlockMutex(&agentMutex);
}

typedef struct {
  PostPo box;
  long tag;
} AttachinfoRec;

static icmStatus detch(void *n,void *r,void *c)
{
  AttachinfoRec *info = (AttachinfoRec*)c;
  agentPo agent = (agentPo)r;

  if(agent->box==info->box){	/* This agent is attached to this box*/
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: forcibly detach %A from %B",pthread_self(),agent,info->box);
#endif

    detachAgent(agent->handle,info->box);
    agent->tag = info->tag;	/* We mark the agent for later re-attachment */
  }
  return icmOk;
}

long detachAllOnBox(PostPo box)
{
  AttachinfoRec info = {box,rand()};

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("Detaching all agents on box %B",box);
#endif

  icmLockMutex(&agentMutex);
  icmProcessTable(detch,postoffice,&info);
  icmUnlockMutex(&agentMutex);
  return info.tag;
}

static icmStatus retch(void *n,void *r,void *c)
{
  AttachinfoRec *info = (AttachinfoRec*)c;
  agentPo agent = (agentPo)r;

  if(agent->tag==info->tag){	/* This agent was attached with this tag */
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: reattach  %A to %B",pthread_self(),agent,info->box);
#endif

    attachAgent(agent->handle,(time_t)0,info->box);
    agent->tag = 0;
  }
  return icmOk;
}

void reattachAllOnBox(PostPo box,long tag)
{
  AttachinfoRec info = {box,tag};

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("Reattaching all agents to box %B",box);
#endif

  icmLockMutex(&agentMutex);
  icmProcessTable(retch,postoffice,&info);
  icmUnlockMutex(&agentMutex);
}


typedef struct {
  icmHandle *table;
  long pos;
} ListAgentRec;

static icmStatus countAgs(void *n,void *r,void *c)
{
  long *count = (long*)c;
  agentPo agent = (agentPo)r;

  if(agent->handle!=NULL){
    icmKeepHandle(agent->handle);
    (*count)++;
  }
  return icmOk;
}
  
static icmStatus listAgs(void *n,void *r,void *c)
{
  agentPo agent = (agentPo)r;
  ListAgentRec *info = (ListAgentRec*)c;

  if(agent->handle!=NULL)
    info->table[info->pos++]=agent->handle;

  return icmOk;
}

icmHandle *listAllAgents(long *count)
{
  ListAgentRec info = {NULL,0};

  icmLockMutex(&agentMutex);

  icmProcessTable(countAgs,postoffice,&info.pos);

  *count = info.pos;

  info.table = (icmHandle*)icmMalloc(sizeof(icmHandle)*info.pos);
  info.pos = 0;

  icmProcessTable(listAgs,postoffice,&info);

  assert(*count==info.pos);

  icmUnlockMutex(&agentMutex);

  return info.table;
}

void freeAgentList(icmHandle *list,long count)
{
  /* int i;

     for(i=0;i<count;i++)
     icmReleaseHandle(list[0]);
  */
  icmFree(list);		/* Release the space occupied by the list */
}

static icmStatus clsAgent(void *n,void *r,void *c)
{
  agentPo agent = (agentPo)r;

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: close agent %A",pthread_self(),agent);
#endif

  triggerMonitors("icmDeregisterAgent",agent);
  closeAgent(agent);

  return icmOk;
}

status attachAgent(icmHandle handle,time_t lease,PostPo box)
{
  agentPo agent;
  status ret = okay;

  icmLockMutex(&agentMutex);

  agent = findAgent(handle);

  if(agent!=NULL){

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: attach agent %A to %B",pthread_self(),agent,box);
#endif

    if(agent->box!=NULL)
      ret = MakeError("agent already attached");
    else{
      unRefBox(agent->box);
      agent->box = box;
      refBox(box);
      if(agent->handle!=handle){
	icmHandle h = agent->handle;

	icmKeepHandle(handle);
	agent->handle = handle;
	agent->tag = 0;
	agent->lease = lease;

	icmReleaseHandle(h);
	triggerMonitors("icmAttachAgent",agent);
      }
    }
  }
  else
    ret = MakeError("agent not registered"); /* Can't attach if not registered */

  icmUnlockMutex(&agentMutex);
  tryRedelivery();
  return ret;
}

status detachAgent(icmHandle handle,PostPo box)
{
  agentPo agent;
  status ret = okay;

  icmLockMutex(&agentMutex);

  agent = findAgent(handle);

  if(agent!=NULL){

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: detach agent %A from %B",pthread_self(),agent,box);
#endif

    if(agent->box==NULL)
      ret = MakeError("agent not attached");
    else if(agent->box!=box)
      ret = MakeError("agent not attached here");
    else{
      triggerMonitors("icmDetachAgent",agent);

      unRefBox(agent->box);
      agent->box = NULL;
    }
  }
  else
    ret = MakeError("agent not registered"); /* Can't attach if not registered */

  icmUnlockMutex(&agentMutex);
  return ret;
}

status requestForwarding(icmHandle handle,icmHandle forward)
{
  agentPo agent;
  status ret = okay;

  icmLockMutex(&agentMutex);

  agent = findAgent(handle);

  if(agent!=NULL){

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: request forwarding for %A to %h",pthread_self(),agent,forward);
#endif

    if(agent->forward!=forward){
      icmHandle ward = agent->forward;

      icmKeepHandle(forward);
      agent->forward = forward;

      icmReleaseHandle(ward);
    }
    tryRedelivery();		/* We might be able to deliver messages now */
  }
  else
    ret = MakeError("agent not registered"); /* Can't attach if not registered */
  icmUnlockMutex(&agentMutex);
  return ret;
}

status cancelForwarding(icmHandle handle,icmHandle forward)
{
  agentPo agent;
  status ret = okay;

  icmLockMutex(&agentMutex);

  agent = findAgent(handle);

  if(agent!=NULL){
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: cancel forwarding for %A to %h",pthread_self(),agent,forward);
#endif

    if(agent->forward==forward){
      if(agent->forward!=NULL){
	icmHandle h = agent->forward;

	agent->forward = NULL;
	icmReleaseHandle(h);
      }

    }
    tryRedelivery();		/* We might be able to deliver messages now */
  }
  else
    ret = MakeError("agent not registered"); /* Can't attach if not registered */
  icmUnlockMutex(&agentMutex);
  return ret;
}

PostPo agentBox(icmHandle handle)
{
  agentPo agent = agentOf(handle);

  if(agent==NULL)
    return NULL;
  else
    return agent->box;
}


icmHandle agentAddress(icmHandle handle)
{
  agentPo agent;
  icmHandle real=NULL;

  icmLockMutex(&agentMutex);

  agent = findAgent(handle);

  if(agent!=NULL){
    if(agent->forward !=NULL)
      real = agent->forward;
    else
      real = agent->handle;
  }

  if(real!=NULL && real!=handle){
    char *tgt,*name,*home,**locs;
    int count;
    if(icmAnalyseHandle(real,NULL,&name,&home,&locs,&count)==icmOk &&
       icmAnalyseHandle(handle,&tgt,NULL,NULL,NULL,NULL)==icmOk)
      handle = icmMakeHandle(tgt,name,home,count,locs);
  }

  icmUnlockMutex(&agentMutex);
  return handle;
}

status registeredAgent(icmHandle handle,icmHandle *real)
{
  agentPo agent;
  status ret = okay;

  icmLockMutex(&agentMutex);

  agent = findAgent(handle);

  if(agent!=NULL){
    if(agent->forward !=NULL)
      *real = agent->forward;
    else
      *real = agent->handle;
  }
  else
    ret = failed;

  icmUnlockMutex(&agentMutex);
  return ret;
}

agentPo agentOf(icmHandle handle)
{
  agentPo agent;

  /*  icmLockMutex(&agentMutex);*/

  agent = findAgent(handle);

  /*  icmUnlockMutex(&agentMutex);*/

  return agent;
}

static agentPo findAgent(icmHandle handle)
{
  char *name,*home,**locs;
  int count;
  if(handle!=NULL && icmAnalyseHandle(handle,NULL,&name,&home,&locs,&count)==icmOk){
    AgentNameRec agname;

    icmStrNCpy(agname.name,name,MAXNAME);
    icmStrNCpy(agname.home,home,MAXNAME);

    return (agentPo)icmSearch(&agname,postoffice);
  }
  else
    return NULL;
}

status monitorAgent(icmHandle ag,icmHandle client)
{
  agentPo agent;
  monitorPo mon;

  icmLockMutex(&agentMutex);

  agent = findAgent(ag);

  if(agent==NULL){
    char *name,*home;

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: setting up agent %h",pthread_self(),ag);
#endif

    agent = (agentPo)icmAllocPool(agpool); /* create a new agent structure */

    icmAnalyseHandle(ag,NULL,&name,&home,NULL,NULL);

    icmStrNCpy(agent->name.name,name,MAXNAME);
    icmStrNCpy(agent->name.home,home,MAXNAME);

    agent->box = NULL;
    agent->handle = NULL;
    agent->forward = NULL;
    agent->monitors = NULL;
    agent->lease = (time_t)0;

    icmInstall(&agent->name,agent,postoffice);
  }

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("Attaching monitor to %A for %h",agent,client);
#endif

  mon = agent->monitors;

  while(mon!=NULL){
    if(icmSameAgentHandle(client,mon->client)==icmOk){
      icmUnlockMutex(&agentMutex);

      icmLogMsg("%A already monitored by %h",agent,mon->client);
      return okay;		/* cant put a monitor on twice */
    }
    else
      mon=mon->next;
  }

  mon = (monitorPo)icmAllocPool(mtpool);

  icmKeepHandle(client);
  mon->client = client;
  mon->next = agent->monitors;
  agent->monitors = mon;

  icmUnlockMutex(&agentMutex);
  return okay;
}

status unmonitorAgent(icmHandle ag,icmHandle client)
{
  agentPo agent;
  monitorPo *mon;
  status stat = okay;

  icmLockMutex(&agentMutex);

  agent = findAgent(ag);

  if(agent!=NULL){

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("detaching monitor to %A for %h",agent,client);
#endif

    mon = &agent->monitors;

    while((*mon)!=NULL){
      if(icmSameAgentHandle(client,(*mon)->client)){
	monitorPo mn = *mon;

	icmReleaseHandle(mn->client);
	*mon = mn->next;
	icmFreePool(mtpool,mn);
      }
      else
	mon = &(*mon)->next;
    }

    if(agent->monitors==NULL && agent->handle == NULL && agent->forward==NULL){
      icmUninstall(&agent->name,postoffice);

#ifdef SERVERTRACE
      if(traceSvr)
	icmLogMsg("%d: agent removed %A",pthread_self(),agent);
#endif
      icmFreePool(agpool,agent);
    }
  }
  else
    stat = failed;
  icmUnlockMutex(&agentMutex);
  return stat;
}

void reportAgentStatus(agentPo agent,icmHandle client,PostPo box)
{

  icmLockMutex(&agentMutex);

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("Informing %h of current status of %A",client,agent);
#endif

  if(agent->box!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%{%(icmMonitorAgent%s%h%)%}","icmRegisterAgent",
			      agent->handle);
    icmKeepHandle(client);
    icmKeepHandle(ourHandle);
    PostMsg(box,client,ourHandle,NULL,msg);
  }
  else if(agent->handle!=NULL){
    icmMsg msg = icmFormatMsg(NULL,"%{%(icmMonitorAgent%s%h%)%}","icmRegisterAgent",
			      agent->handle);
    icmKeepHandle(client);
    icmKeepHandle(ourHandle);
    PostMsg(box,client,ourHandle,NULL,msg);

    msg = icmFormatMsg(NULL,"%{%(icmMonitorAgent%s%h%)%}","icmDetachAgent",agent->handle);
    icmKeepHandle(client);
    icmKeepHandle(ourHandle);
    PostMsg(box,client,ourHandle,NULL,msg);
  }
  icmUnlockMutex(&agentMutex);
}

static icmStatus remMonitor(void *n,void *r,void *c)
{
  agentPo client = (agentPo)c;
  agentPo agent = (agentPo)r;
  monitorPo *mon;

  mon = &agent->monitors;

  while((*mon)!=NULL){
    monitorPo mn = *mon;

    if(icmSameAgentHandle(client->handle,mn->client)==icmOk){
      icmReleaseHandle(mn->client);
      *mon = mn->next;
      icmFreePool(mtpool,mn);
    }
    else
      mon = &mn->next;
  }

  return icmOk;
}

static void removeMonitors(agentPo client)
{
  icmProcessTable(remMonitor,postoffice,client);
}

/* let everyone know we are going */
/* must only be invoked if the locks are set ok */

static void triggerMonitors(char *event,agentPo agent)
{
  icmHandle handle = agent->handle;
  monitorPo mon = agent->monitors;

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: informing clients that %A of %s event",pthread_self(),agent,event);
#endif

  if(handle!=NULL)
    while(mon!=NULL){
      icmMsg msg;
      agentPo clAgent = findAgent(mon->client);

#ifdef SERVERTRACE
      if(traceSvr)
	icmLogMsg("%d: informing %h that %A of %s event",pthread_self(),
		  mon->client,agent,event);
#endif

      msg = icmFormatMsg(NULL,"%{%(icmMonitorAgent%s%h%)%}",event,handle);

      if(clAgent->box!=NULL){
	icmKeepHandle(mon->client);
	icmKeepHandle(ourHandle);
	PostMsg(clAgent->box,mon->client,ourHandle,
		icmNewOpt(NULL,icmReplyto,handle),msg);
      }
      else
	forwardMessage(mon->client,ourHandle,
		       icmNewOpt(NULL,icmReplyto,handle),msg);

      mon = mon->next;
    }
}

#if 0
/* This stuff to be done later */
static status timeoutRequest(char *client,char *sender,Uchar *msg,PostPo box)
{
  Uchar msg_buffer[2048];

  if(scan_april_msg(msg,"<timeoutRequest>")){
    if(canWeTimeout(box))
      fmt_april_msg(msg_buffer,"<timeoutOk>");
    else
      fmt_april_msg(msg_buffer,"<timeoutDenied>");

    return PostMsg(box,client,our_name,our_name,msg_buffer);
  }
  else
    return error;
}
    
static status timeoutOk(char *client,char *sender,Uchar *msg,PostPo box)
{
  if(scan_april_msg(msg,"<timeoutOk>")){
    if(canWeTimeout(box)){
#ifdef SERVERTRACE
    if(traceSvr)
      april_log_msg("remote connection at %s/%d timed out",MailHost(box),MailPort(box));
#endif
      TermPostBox(box);
    }
    return okay;
  }
  else
    return error;
}

static status timeoutDenied(char *client,char *sender,Uchar *msg,PostPo box)
{
  if(scan_april_msg(msg,"<timeoutDenied>"))
    return okay;
  else
    return error;
}

#endif

