/*
 * Set up and maintain connections to remote commservers
 * Part of the ICM communications server
 * (c) 1994-1998 Imperial College, F.G.McCabe and Fujitsu labs
 * All rights reserved
 */
#include "config.h"		/* Invoke configuration header */
#include "daemon.h"		/* Our data structures etc. */

/*
 * A separate thread is established for each attempt to make a remote 
 * connection. If successful, the result is that a remote commserver is
 * registered locally, and we can start forwarding messages to it.
 */

static void fwdMsg(void *cl);
static char *protocolName(char *location,char *buffer,char **remhost);

static icmQueuePo forwarders = NULL;
static icmPoolBase ipool = NULL;

typedef struct _msg_fwd_ {
  icmHandle recip;
  icmHandle snder;
  icmOption opts;
  void *cl;			/* miscellaneous user field */
  icmMsg msg;
} msgItemRec, *msgItemPo;  

status initForwarding(long count)
{
#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: starting %d forwarding servers",pthread_self(),count);
#endif

  forwarders = icmCreateWorkers(count,icmNo);

  ipool = icmInitPool(sizeof(msgItemRec),64);

  if(forwarders!=NULL && ipool!=NULL)
    return okay;
  else
    return error;
}

void forwardMessage(icmHandle recip,icmHandle snder,icmOption opts,icmMsg msg)
{
  msgItemPo item = (msgItemPo)icmAllocPool(ipool);

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: request forwarding of msg to non-registered agent %h",
	      pthread_self(),recip);
#endif

  item->recip = recip;
  item->snder = snder;
  item->opts = opts;
  item->cl = NULL;
  item->msg = msg;

  icmKeepHandle(recip);
  icmKeepHandle(snder);

  icmEnqueueItem(forwarders,fwdMsg,(void*)item);
}

void handleFwdFailed(icmDataPo badLoc,icmHandle recip,icmHandle snder,
		     icmOption opts,icmMsg msg)
{
  msgItemPo item = (msgItemPo)icmAllocPool(ipool);

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: request forwarding of msg to non-registered agent %h",
	      pthread_self(),recip);
#endif

  item->recip = recip;
  item->snder = snder;
  item->opts = opts;
  item->cl = (void*)badLoc;
  item->msg = msg;

  icmKeepHandle(recip);
  icmKeepHandle(snder);

  icmEnqueueItem(forwarders,fwdMsg,(void*)item);
}


icmConn connectToRemote(char *loc)
{
  icmConn remote;
  char *remhost = icmGetHostname(loc);

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: try to connect with %s",pthread_self(),remhost);
#endif
	
  if(remhost!=NULL && icmInitComms(ourPort,remhost,&remote)==icmOk){
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: connected to remote commserver %Q",pthread_self(),remote);
#endif
    if(icmRegisterAgent(remote,ourHandle,NULL,NULL)==icmOk){
      icmHandle rem = icmCommServerHandle(remhost);

      registerAgent(rem,NULL,(time_t)0,
		    OpenPostBox(icmPeerAddr(remote->conn),remote,remhost,
				STAYUPTIME));
      return remote;
    }
    else
      return NULL;
  }
  else
    return NULL;
}

static void fwdMsg(void *cl)
{
  msgItemPo item = (msgItemPo)cl;
  char **locs;
  int count;
  char badLoc[MAXSTRING]="";

  if(item->cl!=NULL){
    icmCstring((icmDataPo)item->cl,badLoc,MAXSTRING);
    free(((icmDataPo)item->cl)->data);
  }

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: search for a connection to %h",pthread_self(),item->recip);
#endif

  if(icmAnalyseHandle(item->recip,NULL,NULL,NULL,&locs,&count)==icmOk){
    int i;

    for(i=0;i<count;i++){	/* We try each of the locations in turn */
      char *remhost;
      char pr[MAXSTRING];
      char *proto = protocolName(locs[i],pr,&remhost);
      icmHandle rem;
      PostPo box=NULL;

      if(strcmp(proto,"icm")==0){ /* forward to another icm server */
	remhost = icmGetHostname(remhost);

	if(remhost!=NULL)
	  rem = icmCommServerHandle(remhost);
	else
	  rem = NULL;

	if(rem!=NULL)
	  box = agentBox(rem); /* check for an existing connection */

	if(box!=NULL){
	  PostMsg(box,item->recip,item->snder,item->opts,item->msg);
	  icmFreePool(ipool,item);

	  return;		/* were done here */
	}
	else if(remhost!=NULL && !isOurAddress(remhost)){ /* need to establish contact with remote icm */
	  icmConn remote;

#ifdef SERVERTRACE
	  if(traceSvr)
	    icmLogMsg("%d: try to connect with %h",pthread_self(),rem);
#endif
	  
	  switch(icmInitComms(ourPort,remhost,&remote)){
	  case icmOk:{
	    if(icmRegisterAgent(remote,ourHandle,NULL,NULL)==icmOk){
	      box = OpenPostBox(icmPeerAddr(remote->conn),remote,remhost,
				STAYUPTIME);
	      registerAgent(rem,NULL,(time_t)0,box); /* register remote with us too */
	    
	      PostMsg(box,item->recip,item->snder,item->opts,item->msg);
	      icmFreePool(ipool,item);

	      return;		/* We're done */
	    }
	  }
	  default:
#ifdef SERVERTRACE
	    if(traceSvr)
	      icmLogMsg("%d: failed to open postbox to remote commserver %C",
			pthread_self(),remote);
#endif
	    continue;
	  }
	}
      }
      else{			/* non-icm protocol handler required */
	rem = icmParseHandle(proto);

	if(rem!=NULL)
	  box = agentBox(rem); /* is the handler registered? */

	if(box!=NULL){
	  icmDataRec loc = {strlen(remhost),remhost};
	  icmHandle recip = removeLocation(item->recip,locs[i]);
	  icmMsg fwd = icmFormatMsg(NULL,"%{%(icmFwdMsg%S%(%h%h%O%#%)%)%}",
				    &loc,recip,item->snder,item->opts,
				    item->msg);

	  icmReleaseHandle(recip);


#ifdef SERVERTRACE
	  if(traceSvr)
	    icmLogMsg("%d: forwarding via registered handler %h",
		      pthread_self(),rem);
#endif
	  PostMsg(box,rem,ourHandle,NULL,fwd);
	  icmFreePool(ipool,item);
	  return;
	}
      }
    }

#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: failed to find a forwarding location for %h",pthread_self(),
		item->recip);
#endif
    PostMsg(NULL,item->recip,item->snder,item->opts,item->msg);
    icmFreePool(ipool,item);
  }
}

/*
 * Analyse a location string to extract the protocol and location data of a location
 */
static char *protocolName(char *location,char *buffer,char **remhost)
{
  char *p = strchr(location,':');
  char *b = buffer;

  if(p!=NULL){
    long len = p-location;

    while(len--)
      *b++=*location++;
    *b = '\0';

    *remhost = p+1;
    return buffer;
  }
  else{
    *remhost = location;
    return "icm";
  }
}

PostPo fwdBox(icmHandle recip)
{
  char **locs;
  int count;
  PostPo box;

  if((box=agentBox(recip))!=NULL)
    return box;

  if(icmAnalyseHandle(recip,NULL,NULL,NULL,&locs,&count)==icmOk){
    int i;

    for(i=0;box==NULL && i<count;i++){	/* We try each of the locations in turn */
      char *remhost;
      char pr[MAXSTRING];
      char *proto = protocolName(locs[i],pr,&remhost);
      icmHandle rem;

      if(strcmp(proto,"icm")==0){
	if(remhost!=NULL)
	  remhost = icmGetHostname(remhost);

	if(remhost!=NULL)
	  rem = icmCommServerHandle(remhost);
	else
	  rem = NULL;
      }
      else
	rem = icmParseHandle(proto);

      if(rem!=NULL)
	box = agentBox(rem);
    }

    return box;
  }
  else
    return NULL;
}

