/*
 * Command fan-out for the ICM server
 * (c) 1994-1998 Imperial College, F.G.McCabe and Fujitsu labs
 * All rights reserved
 */

#include "config.h"
#include "daemon.h"		/* Our data structures etc. */
#include "office.h"		/* Private post office structures */

#ifdef SERVERTRACE
icmTruth traceSvr = icmNo;
#endif

static icmQueuePo servers = NULL;
static icmPoolBase ipool = NULL;

typedef struct _srv_msg_ {
  icmHandle snder;
  icmOption opts;
  icmMsg msg;
  PostPo box;
} svrItemRec, *svrItemPo;  

static status registerAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status deregisterAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status detachAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status attachAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status unmonitorAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status monitorAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status addressForwardingCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status cancelForwardingCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status listAgentsCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status pingAgentCmd(icmHandle sndr,icmOption opts,icmMsg msg,PostPo box);
static status disconnectAgentsCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status reconnectAgentsCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status dumpCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);
static status fwdFailedCmd(icmMsg msg,icmOption opts,PostPo box);
static status errorCase(icmHandle snder,icmOption opts,icmMsg msg,PostPo box);

static void doServerMsg(void *cl);

void initServer(void)
{
  servers = icmCreateWorkers(1,icmNo);
  ipool = icmInitPool(sizeof(svrItemRec),8);
}

static char *checkOperation(icmMsg msg,long x,char *key,long klen)
{
  long index;
  icmElTag tag;

  if(icmMsgElType(msg,x,&tag)==icmOk)
    switch(tag){
    case icmSym:
      icmExpectSym(msg,x,&index,key,klen);
      return key;

    case icmApply:{
      icmExpectApply(msg,x,&index);
      if(icmMsgElType(msg,index,&tag)!=icmOk || tag!=icmSym)
	return NULL;
      else{
	icmExpectSym(msg,index,&index,key,MAXNAME);

	return key;
      }
    }
    case icmTuple:{
      long arity;
      icmExpectTuple(msg,x,&index,&arity);

      if(arity==0 || icmMsgElType(msg,index,&tag)!=icmOk || tag!=icmSym)
	return NULL;
      else{
	icmExpectSym(msg,index,&index,key,MAXNAME);

	return key;
      }
    }
    case icmSigned:{		/* look for a signed value */
      char sig[MAXSTRING];
      long len = NumberOf(sig);

      if(icmExpectSigned(msg,x,&x,sig,&len)==icmOk)
	return checkOperation(msg,x,key,klen);
      else
	return NULL;
    }

    default:
      return NULL;
    }
  else
    return NULL;
}

void ServerMsg(icmHandle recip,icmHandle sndr,icmOption opts,icmMsg msg,
	       PostPo box)
{
  svrItemPo item = (svrItemPo)icmAllocPool(ipool);

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: request server handling of msg %M",pthread_self(),msg);
#endif

  item->snder = sndr;
  item->msg = msg;
  item->opts = opts;
  item->box = box;

  icmKeepHandle(sndr);

  icmEnqueueItem(servers,doServerMsg,item); /* We have a server thread that does this */
}

static void doServerMsg(void *cl)
{
  svrItemPo item = (svrItemPo)cl;
  char key[MAXNAME];

  if(checkOperation(item->msg,0,key,NumberOf(key))!=NULL){
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: server handling %s",pthread_self(),key);
#endif

    if(strcmp(key,"icmRegisterAgent")==0)
      registerAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmDeregisterAgent")==0)
      deregisterAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmAttachAgent")==0)
      attachAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmDetachAgent")==0)
      detachAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmMonitorAgent")==0)
      monitorAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmUnMonitorAgent")==0)
      unmonitorAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmListAgents")==0)
      listAgentsCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmPingAgent")==0)
      pingAgentCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmRequestForwarding")==0)
      addressForwardingCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmCancelForwarding")==0)
      cancelForwardingCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmDisconnect")==0)
      disconnectAgentsCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmReconnect")==0)
      reconnectAgentsCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"icmFwdFailed")==0)
      fwdFailedCmd(item->msg,item->opts,item->box);
    else if(strcmp(key,"dump")==0)
      dumpCmd(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"error")==0)
      errorCase(item->snder,item->opts,item->msg,item->box);
    else if(strcmp(key,"quit")==0)
      exit_daemon(0);
#ifdef SERVERTRACE
    if(traceSvr)
      icmLogMsg("%d: server handled %s",pthread_self(),key);
#endif

  }
  else
    icmLogMsg("Unknown server command from %h: %M",item->snder,item->msg);

  icmReleaseMsg(item->msg);
  icmReleaseHandle(item->snder);
  icmFreePool(ipool,item);
}

static status errorCase(icmHandle recip,icmOption opts,icmMsg msg,PostPo box)
{
  icmDataRec str;

  if(icmScanMsg(msg,"%(error%S%#%)",&str,NULL)==icmOk){
    icmLogMsg("Error %S detected",&str);
    free(str.data);
  }
  else if(icmScanMsg(msg,"%{%(error%S%#%)%}",&str,NULL)==icmOk){
    icmLogMsg("Error %S detected",&str);
    free(str.data);
  }
  else if(icmScanMsg(msg,"%{%Rerror%(%S%#%)%}",&str,NULL)==icmOk){
    icmLogMsg("Error %S detected",&str);
    free(str.data);
  }
  return okay;
}

/*
 * Each protocol is implemented using a separate function
 * which makes it easier to extend.
 */

static status registerAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box)
{
  icmHandle agent,real;
  icmOption ropt;
  status ret = okay;

  if(icmScanMsg(msg,"%(icmRegisterAgent%h%O%)",&agent,&ropt)==icmOk){
    icmHandle replyto;
    time_t lease = 0;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;
    icmIsOption(ropt,icmLeaseTime,&lease);
    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(registerAgent(agent,&real,lease,box)){
    case okay:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmRegisterAgent%,icmOk%h%)",real);
      icmLogMsg("Agent %h registered at %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }

    case failed:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmRegisterAgent%,icmFailed%h%)",
				  agent);
      icmLogMsg("Unable to register agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      ret = failed;
      break;
    }
      
    default:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmRegisterAgent%,icmFailed%h%)",
				  agent);
      icmLogMsg("Duplicate agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      ret = error;
      break;
    }
    }

    icmReleaseHandle(agent);
    icmReleaseOptions(ropt);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmRegisterAgent%h%O%)%}",&agent,&ropt)==icmOk){
    icmHandle replyto;
    time_t lease = 0;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;
    icmIsOption(ropt,icmLeaseTime,&lease);
    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(registerAgent(agent,&real,lease,box)){
    case okay:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmRegisterAgent%,icmOk%h%)%}",real);
      icmLogMsg("Agent %h registered at %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }

    case failed:{
      icmMsg reply = icmFormatMsg(NULL,"%{%(icmRegisterAgent%,icmFailed%h%)%}",
				  agent);
      icmLogMsg("Unable to register agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      ret = failed;
      break;
    }
      
    default:{
      icmMsg reply = icmFormatMsg(NULL,"%{%(icmRegisterAgent%,icmFailed%h%)%}",
				  agent);
      icmLogMsg("Duplicate agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      ret = error;
      break;
    }
    }

    icmReleaseHandle(agent);
    icmReleaseOptions(ropt);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status deregisterAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box)
{
  icmHandle agent;

#ifdef SERVERTRACE
  if(traceSvr)
    icmLogMsg("%d: deregister agent command",pthread_self());
#endif

  if(icmScanMsg(msg,"%(icmDeregisterAgent%h%)",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(deregisterAgent(agent,box)){
    case okay:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmDeregisterAgent%,icmOk%h%)",agent);
      icmLogMsg("Agent %h deregistered from %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
      
    default:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmDeregisterAgent%,icmFailed%h%)",agent);
      icmLogMsg("Duplicate agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmDeregisterAgent%h%)%}",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(deregisterAgent(agent,box)){
    case okay:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmDeregisterAgent%,icmOk%h%)%}",
				agent);
      icmLogMsg("Agent %h deregistered from %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
      
    default:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmDeregisterAgent%,icmFailed%h%)%}",
				agent);
      icmLogMsg("Duplicate agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status detachAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,
			     PostPo box)
{
  icmHandle agent;
  icmOption aopts;

  if(icmScanMsg(msg,"%(icmDetachAgent%h%O%)",&agent,&aopts)==icmOk){
    icmHandle replyto;
    time_t lease = 0;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;
    icmIsOption(aopts,icmLeaseTime,&lease);

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(detachAgent(agent,box)){
    case okay:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmDetachAgent%,icmOk%h%)",agent);
      icmLogMsg("Agent %h detached from %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    default:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmDetachAgent%,icmFailed%h%)",agent);
      icmLogMsg("Unable to detach agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    }
    icmReleaseHandle(agent);
    icmReleaseOptions(aopts);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmDetachAgent%h%O%)%}",&agent,&aopts)==icmOk){
    icmHandle replyto;
    time_t lease = 0;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;
    icmIsOption(aopts,icmLeaseTime,&lease);

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(detachAgent(agent,box)){
    case okay:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmDetachAgent%,icmOk%h%)%}",agent);
      icmLogMsg("Agent %h detached from %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    default:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmDetachAgent%,icmFailed%h%)%}",
				agent);
      icmLogMsg("Unable to detach agent %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    }
    icmReleaseHandle(agent);
    icmReleaseOptions(aopts);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status attachAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,
			     PostPo box)
{
  icmHandle agent;
  icmOption aopts;

  if(icmScanMsg(msg,"%(icmAttachAgent%h%O%)",&agent,&aopts)==icmOk){
    icmHandle replyto;
    time_t lease = 0;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;
    icmIsOption(aopts,icmLeaseTime,&lease);

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(attachAgent(agent,lease,box)){
    case okay:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmAttachAgent%,icmOk%h%)",agent);
      icmLogMsg("Agent %h attached to %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    default:{
      icmMsg reply = icmFormatMsg(NULL,"%(icmAttachAgent%,icmFailed%h%)",agent);
      icmLogMsg("Unable to attach %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    }
    icmReleaseHandle(agent);
    icmReleaseOptions(aopts);
    /*    icmReleaseOptions(opts);*/
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmAttachAgent%h%O%)%}",&agent,&aopts)==icmOk){
    icmHandle replyto;
    time_t lease = 0;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = agent;
    icmIsOption(aopts,icmLeaseTime,&lease);

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);

    switch(attachAgent(agent,lease,box)){
    case okay:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmAttachAgent%,icmOk%h%)%}",agent);
      icmLogMsg("Agent %h attached to %B",agent,box);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    default:{
      icmMsg reply=icmFormatMsg(NULL,"%{%(icmAttachAgent%,icmFailed%h%)%}",
				agent);
      icmLogMsg("Unable to attach %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,reply);
      break;
    }
    }
    icmReleaseHandle(agent);
    icmReleaseOptions(aopts);
    /*    icmReleaseOptions(opts);*/
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status monitorAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,
			      PostPo box)
{
  icmHandle agent;

  if(icmScanMsg(msg,"%(icmMonitorAgent%h%)",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    switch(monitorAgent(agent,snder)){
    case okay:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmMonitorAgent%,icmOk%h%)",agent);
      icmLogMsg("Agent %h monitored by %h",agent,replyto);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      reportAgentStatus(agentOf(agent),replyto,box);
      break;
    }
    default:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmMonitorAgent%,icmFailed%h%)",agent);
      icmLogMsg("Unable to monitor %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmMonitorAgent%h%)%}",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    switch(monitorAgent(agent,snder)){
    case okay:{
      icmMsg replyM=icmFormatMsg(NULL,"%{%(icmMonitorAgent%,icmOk%h%)%}",agent);
      icmLogMsg("Agent %h monitored by %h",agent,replyto);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      reportAgentStatus(agentOf(agent),replyto,box);
      break;
    }
    default:{
      icmMsg replyM=icmFormatMsg(NULL,"%{%(icmMonitorAgent%,icmFailed%h%)%}",
				 agent);
      icmLogMsg("Unable to monitor %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status unmonitorAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,
				PostPo box)
{
  icmHandle agent;

  if(icmScanMsg(msg,"%(icmUnMonitorAgent%h%)",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    switch(unmonitorAgent(agent,snder)){
    case okay:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmUnMonitorAgent%,icmOk%h%)",agent);
      icmLogMsg("Agent %h no longer monitored by %h",agent,snder);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    default:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmMonitorAgent%,icmFailed%h%)",agent);
      icmLogMsg("Unable to remove monitor for %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmUnMonitorAgent%h%)%}",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    switch(unmonitorAgent(agent,snder)){
    case okay:{
      icmMsg replyM = icmFormatMsg(NULL,"%{%(icmUnMonitorAgent%,icmOk%h%)%}",
				   agent);
      icmLogMsg("Agent %h no longer monitored by %h",agent,snder);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    default:{
      icmMsg replyM = icmFormatMsg(NULL,"%{%(icmMonitorAgent%,icmFailed%h%)%}",
				   agent);
      icmLogMsg("Unable to remove monitor for %h",agent);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status listAgentsCmd(icmHandle snder,icmOption opts,icmMsg msg,
			    PostPo box)
{
  if(icmScanMsg(msg,"icmListAgents")==icmOk){
    long count;
    icmHandle *agents=listAllAgents(&count);
    icmMsg reply = icmFormatMsg(NULL,"%(icmListAgents%*h%)",count,agents);
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);
    PostMsg(box,replyto,ourHandle,NULL,reply);

    freeAgentList(agents,count);
    return okay;
  }
  else if(icmScanMsg(msg,"%{icmListAgents%}")==icmOk){
    long count;
    icmHandle *agents=listAllAgents(&count);
    icmMsg reply = icmFormatMsg(NULL,"%{%(icmListAgents%*h%)%}",count,agents);
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);
    PostMsg(box,replyto,ourHandle,NULL,reply);

    freeAgentList(agents,count);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status pingAgentCmd(icmHandle snder,icmOption opts,icmMsg msg,
			   PostPo box)
{
  icmHandle agent,real;

  if(icmScanMsg(msg,"%(icmPingAgent%h%)",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    if(registeredAgent(agent,&real)==okay){
      icmMsg replyM = icmFormatMsg(NULL,"%(icmPingAgent%,icmOk%h%)",real);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
    }
    else{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmPingAgent%,icmFailed%h%)",agent);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmPingAgent%h%)%}",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    if(registeredAgent(agent,&real)==okay){
      icmMsg replyM = icmFormatMsg(NULL,"%{%(icmPingAgent%,icmOk%h%)%}",real);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
    }
    else{
      icmMsg replyM = icmFormatMsg(NULL,"%{%(icmPingAgent%,icmFailed%h%)%}",
				   agent);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status addressForwardingCmd(icmHandle sndr,icmOption opts,icmMsg msg,
				   PostPo box)
{
  icmHandle agent;

  if(icmScanMsg(msg,"%(icmRequestForwarding%h%)",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = sndr;

    switch(requestForwarding(agent,sndr)){
    case okay:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmRequestForwarding%,icmOk%h%)",agent);
      icmLogMsg("forwarding messages for %h to %h",agent,sndr);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    default:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmRequestForwarding%,icmError%h%)",agent);
      icmLogMsg("forwarding messages for %h to %h not allowed",agent,sndr);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmRequestForwarding%h%)%}",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = sndr;

    switch(requestForwarding(agent,sndr)){
    case okay:{
      icmMsg replyM=icmFormatMsg(NULL,"%{%(icmRequestForwarding%,icmOk%h%)%}",
				 agent);
      icmLogMsg("forwarding messages for %h to %h",agent,sndr);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    default:{
      icmMsg replyM=icmFormatMsg(NULL,"%{%(icmRequestForwarding%,icmError%h%)%}",
				 agent);
      icmLogMsg("forwarding messages for %h to %h not allowed",agent,sndr);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else
    return MakeError("invalid request");
}

static status cancelForwardingCmd(icmHandle snder,icmOption opts,icmMsg msg,
				  PostPo box)
{
  icmHandle agent;

  if(icmScanMsg(msg,"%(icmCancelForwarding%h%)",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    switch(cancelForwarding(agent,snder)){
    case okay:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmCancelForwarding%,icmOk%h%)",agent);
      icmLogMsg("cancelling forwarding messages for %h to %h",agent,snder);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    default:{
      icmMsg replyM = icmFormatMsg(NULL,"%(icmCancelForwarding%,icmError%h%)",agent);
      icmLogMsg("cant cancel forwarding messages for %h to %h not allowed",agent,snder);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmCancelForwarding%h%)%}",&agent)==icmOk){
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    switch(cancelForwarding(agent,snder)){
    case okay:{
      icmMsg replyM = icmFormatMsg(NULL,"%{%(icmCancelForwarding%,icmOk%h%)%}",
				   agent);
      icmLogMsg("cancelling forwarding messages for %h to %h",agent,snder);
      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    default:{
      icmMsg replyM=icmFormatMsg(NULL,"%{%(icmCancelForwarding%,icmError%h%)%}",
				   agent);
      icmLogMsg("cant cancel forwarding messages for %h to %h not allowed",agent,snder);

      icmKeepHandle(replyto);
      icmKeepHandle(ourHandle);
      PostMsg(box,replyto,ourHandle,NULL,replyM);
      break;
    }
    }
    icmReleaseHandle(agent);
    return okay;
  }
  else
    return MakeError("invalid request");
}


static status fwdFailedCmd(icmMsg msg,icmOption opts,PostPo box)
{
  icmHandle recip,sendr;
  icmDataRec str;
  icmMsg subMsg;
  icmOption ropts;

  if(icmScanMsg(msg,"%(icmFwdFailed%S%h%h%O%#%)",&str,&recip,&sendr,&ropts,
		&subMsg)==icmOk){
    handleFwdFailed(&str,recip,sendr,ropts,subMsg);
    icmReleaseHandle(recip);
    icmReleaseHandle(sendr);
    /*    icmReleaseOptions(ropts);*/
    free(str.data);
    return icmOk;
  }
  else if(icmScanMsg(msg,"%{%(icmFwdFailed%S%h%h%O%#%)%}",&str,&recip,&sendr,
		     &ropts,&subMsg)==icmOk){
    handleFwdFailed(&str,recip,sendr,ropts,subMsg);
    icmReleaseHandle(recip);
    icmReleaseHandle(sendr);
    /*    icmReleaseOptions(ropts);*/
    free(str.data);
    return icmOk;
  }
  else
    return MakeError("invalid request");
}

static status disconnectAgentsCmd(icmHandle snder,icmOption opts,icmMsg msg,
				  PostPo box)
{
  if(icmScanMsg(msg,"icmDisconnect")==icmOk){
    long tag = detachAllOnBox(box);
    icmMsg replyM = icmFormatMsg(NULL,"%(icmDisconnect%,icmOk%d%)",tag);
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);
    PostMsg(box,replyto,ourHandle,NULL,replyM);
    return okay;
  }
  else if(icmScanMsg(msg,"%{icmDisconnect%}")==icmOk){
    long tag = detachAllOnBox(box);
    icmMsg replyM = icmFormatMsg(NULL,"%{%(icmDisconnect%,icmOk%d%)%}",tag);
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);
    PostMsg(box,replyto,ourHandle,NULL,replyM);
    return okay;
  }
  else
    return MakeError("invalid request");
}


static status reconnectAgentsCmd(icmHandle snder,icmOption opts,icmMsg msg,PostPo box)
{
  long tag;
  if(icmScanMsg(msg,"%(icmReconnect%d%)",&tag)==icmOk){
    icmMsg replyM = icmFormatMsg(NULL,"%(icmReconnect%,icmOk%)");
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    reattachAllOnBox(box,tag);
    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);
    PostMsg(box,replyto,ourHandle,NULL,replyM);
    return okay;
  }
  else if(icmScanMsg(msg,"%{%(icmReconnect%d%)%}",&tag)==icmOk){
    icmMsg replyM = icmFormatMsg(NULL,"%{%(icmReconnect%,icmOk%)%}");
    icmHandle replyto;

    if(icmIsOption(opts,icmReplyto,&replyto)!=icmOk)
      replyto = snder;

    reattachAllOnBox(box,tag);
    icmKeepHandle(replyto);
    icmKeepHandle(ourHandle);
    PostMsg(box,replyto,ourHandle,NULL,replyM);
    return okay;
  }
  else
    return MakeError("invalid request");
}


static status dumpCmd(icmHandle recip,icmOption opts,icmMsg msg,PostPo box)
{
  if(icmScanMsg(msg,"dump")==icmOk){
    dumpPostOffice();
    dumpBoxes();
    icmDumpAddresses();

    return okay;
  }
  else if(icmScanMsg(msg,"%{dump%}")==icmOk){
    dumpPostOffice();
    dumpBoxes();
    icmDumpAddresses();

    return okay;
  }
  else
    return MakeError("invalid request");
}

